/*-module-*****************************************************************
 *
 * Modul     : Agent (FUNKTIONSMODUL)
 *
 * Beschreib.:
 *
 * Autor     : <autor>
 *
 * Datum     : <datum>
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/

#include "Agent.h"

#include "AgentPrivate.h"
#include "AgentList.h"

#include "../program/AgentPrgFunc.h"
#include "../misc/Messages.h"
#include "../misc/Memory.h"
#include "../dump/Dump.h"
#include <stdlib.h>
#include <stdio.h>


static void 
freeProgram (ProgramType * Program)
{
  NodePtrType root;
  unsigned child;

  if ((root = AgentPrgFunc_Program2NodePtr (Program)) != NULL)
    {
      for (child = 0; child < 3; child++)
	AgentPrgFunc_Del (root, child);

      free (root);
    }
  free (Program);
}

static ProgramType *
copyProgram (ProgramType * Program)
{
  ProgramType *copy;
  NodePtrType root;
  CountType length;

  if (Program)
    {
      root = AgentPrgFunc_CopyTree (AgentPrgFunc_Program2NodePtr (Program));
      length = AgentPrgFunc_NodeCount (root);
      copy = AgentPrgFunc_NodePtr2Program (root, length);

      return copy;
    }
  else
    return NULL;
}


AgentType *
Agent_CreateAgent ()
{
  return Callocate (sizeof (AgentType), 1);
}

void 
Agent_DestroyAgent (AgentType * agent)
{
  freeProgram (agent->Program);
  if (agent->FriendList)
    AgentList_DestroyList (agent->FriendList);
  Free (agent);
}


void 
Agent_CopyAgent (AgentType * dest, const AgentType * src)
{
  CopyMemory (dest, src, sizeof (AgentType));
  dest->FriendList = AgentList_CopyList (NULL, src->FriendList);
  dest->Program = copyProgram (src->Program);
}

void 
Agent_SetId (AgentType * PtrAgent, AgentIdType AgentId)
{
  PtrAgent->Id = AgentId;
}

AgentIdType 
Agent_GetId (AgentType * PtrAgent)
{
  return PtrAgent->Id;
}

void 
Agent_SetType (AgentType * PtrAgent, AgentIdType TypeId)
{
  PtrAgent->Type = TypeId;
}

AgentIdType 
Agent_GetType (AgentType * PtrAgent)
{
  return PtrAgent->Type;
}

void 
Agent_SetEnergy (AgentType * PtrAgent, EnergyType Energy)
{
  if (Energy > PtrAgent->MaxEnergy)
    Energy = PtrAgent->MaxEnergy;
  else if (Energy < 0)
    Energy = 0;

  PtrAgent->Energy = Energy;

  DebugLevel (2, "Energy set to %d at %d.\n", Energy, PtrAgent->Id);
}

EnergyType 
Agent_GetEnergy (AgentType * PtrAgent)
{
  return PtrAgent->Energy;
}

void 
Agent_SetAge (AgentType * PtrAgent, CountType Age)
{
  PtrAgent->Age = Age;
}

CountType 
Agent_GetAge (AgentType * PtrAgent)
{
  return PtrAgent->Age;
}

void 
Agent_SetAttacks (AgentType * PtrAgent, CountType Attacks)
{
  /* if (Attacks < 0)
    Warning ("Agent_SetAttacks", "Agent Attacks below 0.\n");
  else */
    PtrAgent->Attacks = Attacks;
}

CountType 
Agent_GetAttacks (AgentType * PtrAgent)
{
  return PtrAgent->Attacks;
}

void 
Agent_SetIdle (AgentType * PtrAgent, CountType Idle)
{
  PtrAgent->Idle = Idle;
}

CountType 
Agent_GetIdle (AgentType * PtrAgent)
{
  return PtrAgent->Idle;
}

void 
Agent_SetMemory (AgentType * PtrAgent, MemoryType Memory)
{
  PtrAgent->Memory = Memory;
}

MemoryType 
Agent_GetMemory (AgentType * PtrAgent)
{
  return PtrAgent->Memory;
}

void 
Agent_SetXPosition (AgentType * PtrAgent, int Position)
{
  PtrAgent->Position.x = Position;
}

void 
Agent_SetYPosition (AgentType * PtrAgent, int Position)
{
  PtrAgent->Position.y = Position;
}

void 
Agent_SetPosition (AgentType * PtrAgent, PositionType Position)
{
  PtrAgent->Position = Position;
}

PositionType 
Agent_GetPosition (AgentType * PtrAgent)
{
  return PtrAgent->Position;
}

ProgramType *
Agent_GetProgram (AgentType * PtrAgent)
{
  return PtrAgent->Program;
}

void 
Agent_SetProgram (AgentType * PtrAgent, ProgramType * Program)
{
  if (PtrAgent->Program)
    freeProgram (PtrAgent->Program);

/*      if (Program)
   PtrAgent->Program = copyProgram(Program);
   else */
  PtrAgent->Program = Program;
}

CountType 
Agent_GetProgramLength (AgentType * PtrAgent)
{
  return AgentPrgFunc_GetProgramLength (PtrAgent->Program);
}

void 
Agent_SetProgramLength (AgentType * PtrAgent, CountType Length)
{
  AgentPrgFunc_SetProgramLength (PtrAgent->Program, Length);
}

void 
Agent_SetOrientation (AgentType * PtrAgent, OrientationType Orientation)
{
/*   if ((Orientation % 45) == 0)
   {
   if (Orientation < 0)
   PtrAgent->Orientation = 360 - (Orientation % 360);
   else */
  PtrAgent->Orientation = Orientation % 360;
/*      }
   else
   Warning("Agent_SetOrientation", "Orientation not multiple of 45.\n"); */
}

OrientationType 
Agent_GetOrientation (AgentType * PtrAgent)
{
  return PtrAgent->Orientation;
}

void 
Agent_SetScreams (AgentType * PtrAgent, CountType Screams)
{
  PtrAgent->Screams = Screams;
}

void 
Agent_SetScreamed (AgentType * PtrAgent, bool Screamed)
{
  Agent_SetScreams (PtrAgent, (CountType) Screamed);
}

CountType 
Agent_GetScreams (AgentType * PtrAgent)
{
  return PtrAgent->Screams;
}

bool 
Agent_GetScreamed (AgentType * PtrAgent)
{
  return (bool) Agent_GetScreams (PtrAgent);
}

void 
Agent_SetMaxEnergy (AgentType * PtrAgent, EnergyType MaxEnergy)
{
  if (MaxEnergy > 0)
    PtrAgent->MaxEnergy = MaxEnergy;
  else
    Warning ("Agent_SetMaxEnergy", "MaxEnergy below 0.\n");
}

EnergyType 
Agent_GetMaxEnergy (AgentType * PtrAgent)
{
  return PtrAgent->MaxEnergy;
}

void 
Agent_SetMaxFriends (AgentType * PtrAgent, CountType MaxFriends)
{
  PtrAgent->MaxFriends = MaxFriends;
}

CountType 
Agent_GetMaxFriends (AgentType * PtrAgent)
{
  return PtrAgent->MaxFriends;
}

void 
Agent_SetRangeOfVision (AgentType * PtrAgent, CountType RangeOfVision)
{
  PtrAgent->RangeOfVision = RangeOfVision;
}

CountType 
Agent_GetRangeOfVision (AgentType * PtrAgent)
{
  return PtrAgent->RangeOfVision;
}

void 
Agent_SetRangeOfHearing (AgentType * PtrAgent, CountType RangeOfHearing)
{
  PtrAgent->RangeOfHearing = RangeOfHearing;
}

CountType 
Agent_GetRangeOfHearing (AgentType * PtrAgent)
{
  return PtrAgent->RangeOfHearing;
}

void 
Agent_SetStrength (AgentType * PtrAgent, StrengthType Strength)
{
  PtrAgent->Strength = Strength;
}

StrengthType 
Agent_GetStrength (AgentType * PtrAgent)
{
  return PtrAgent->Strength;
}

void 
Agent_SetMinEnergyFactor (AgentType * PtrAgent, FactorType MinEnergyFactor)
{
  PtrAgent->MinEnergyFactor = MinEnergyFactor;
}

FactorType 
Agent_GetMinEnergyFactor (AgentType * PtrAgent)
{
  return PtrAgent->MinEnergyFactor;
}

void 
Agent_SetHearingRange (AgentType * PtrAgent, CountType HearingRange)
{
  PtrAgent->RangeOfHearing = HearingRange;
}

CountType 
Agent_GetHearingRange (AgentType * PtrAgent)
{
  return PtrAgent->RangeOfHearing;
}


void 
Agent_SetHungerFactor (AgentType * PtrAgent, FactorType HungerFactor)
{
  PtrAgent->HungerFactor = HungerFactor;
}

FactorType 
Agent_GetHungerFactor (AgentType * PtrAgent)
{
  return PtrAgent->HungerFactor;
}

void 
Agent_SetAttackFactor (AgentType * PtrAgent, FactorType AttackFactor)
{
  PtrAgent->AttackFactor = AttackFactor;
}

FactorType 
Agent_GetAttackFactor (AgentType * PtrAgent)
{
  return PtrAgent->AttackFactor;
}

void 
Agent_SetDefenseFactor (AgentType * PtrAgent, FactorType DefenseFactor)
{
  PtrAgent->DefenseFactor = DefenseFactor;
}

FactorType 
Agent_GetDefenseFactor (AgentType * PtrAgent)
{
  return PtrAgent->DefenseFactor;
}

void 
Agent_SetCostOfLiving (AgentType * PtrAgent, EnergyType CostOfLiving)
{
  PtrAgent->CostOfLiving = CostOfLiving;
}

EnergyType 
Agent_GetCostOfLiving (AgentType * PtrAgent)
{
  return PtrAgent->CostOfLiving;
}

void 
Agent_SetNumberOfChildren (AgentType * PtrAgent, CountType NumberOfChildren)
{
  PtrAgent->NumberOfChildren = NumberOfChildren;
}

CountType 
Agent_GetNumberOfChildren (AgentType * PtrAgent)
{
  return PtrAgent->NumberOfChildren;
}

void 
Agent_SetMaxAge (AgentType * PtrAgent, CountType MaxAge)
{
  PtrAgent->MaxAge = MaxAge;
}

CountType 
Agent_GetMaxAge (AgentType * PtrAgent)
{
  return PtrAgent->MaxAge;
}

void 
Agent_SetMinMatingAge (AgentType * PtrAgent, CountType MinMatingAge)
{
  PtrAgent->MinMatingAge = MinMatingAge;
}

CountType 
Agent_GetMinMatingAge (AgentType * PtrAgent)
{
  return PtrAgent->MinMatingAge;
}

/* added by Martin */
void 
Agent_SetLastMate (AgentType * PtrAgent, CountType LastMate)
{
  PtrAgent->LastMate = LastMate;
}

/* added by Martin */
CountType 
Agent_GetLastMate (AgentType * PtrAgent)
{
  return PtrAgent->LastMate;
}

inline CountType 
Agent_GetMatePause (AgentType * PtrAgent)
{
  return PtrAgent->MatePause;
}

inline void 
Agent_SetMatePause (AgentType * PtrAgent, CountType MatePause)
{
  PtrAgent->MatePause = MatePause;
}

inline void 
Agent_IncreaseAge (AgentType * PtrAgent, CountType Timesteps)
{
  PtrAgent->Age += Timesteps;
}

inline bool 
Agent_Age (AgentType * PtrAgent)
{
  return PtrAgent->Age >= PtrAgent->MaxAge;
}

bool 
Agent_ConsumeEnergy (AgentType * PtrAgent, EnergyType Energy)
{
  EnergyType energy, maxEnergy, minEnergy;
  float minEnergyFactor;
  bool dead = 0;

  energy = Agent_GetEnergy (PtrAgent);
  maxEnergy = Agent_GetMaxEnergy (PtrAgent);
  minEnergyFactor = Agent_GetMinEnergyFactor (PtrAgent);

  if (Energy >= 0)
    {
      energy -= Energy;
      minEnergy = maxEnergy * minEnergyFactor;
      if (energy < minEnergy)
	{
	  energy = minEnergy;
	  dead = 1;
	}

      Agent_SetEnergy (PtrAgent, energy);
    }
  else if (Energy < 0)
    {
      Warning ("Agent_ConsumeEnergy",
	       "Agent %d: Consuming negative Energy %d\n",
	       Agent_GetId (PtrAgent), Energy);

      energy -= Energy;
      Agent_SetEnergy (PtrAgent, energy);
    }
  return dead;
}

int 
Agent_Timestep (AgentType * PtrAgent)
{
  char buffer[256];
  int state = AGENT_ALIVE;
  AgentIdType agent = Agent_GetId (PtrAgent);
  AgentIdType agenttype = Agent_GetType (PtrAgent);
  FactorType MinEnergyFactor = Agent_GetMinEnergyFactor (PtrAgent);
  EnergyType MaxEnergy = Agent_GetMaxEnergy (PtrAgent);

  if (Agent_Age (PtrAgent))
    {
      if (Dump_WantDumpBirthAndDeath ())
	{
	  sprintf (buffer, "#! Death = %06d %c : Age = %06d\n",
		   agent, agenttype, Agent_GetAge (PtrAgent));
	  Dump_DumpBirthAndDeath (buffer);
	}

      DebugInfo ("Agent %d died - his age was %4d timesteps\n",
		 agent, Agent_GetAge (PtrAgent));
      state = AGENT_FATAL_ALZHEIMER;
    }
  else if ((Agent_GetEnergy (PtrAgent) <= (EnergyType) (MinEnergyFactor * MaxEnergy)) || Agent_GetAttacks (PtrAgent) == -1)
    {
      if (Dump_WantDumpBirthAndDeath ())
	{
	  sprintf (buffer, "#! Death = %06d %c : Energy = %06d\n",
		   agent, agenttype, Agent_GetEnergy (PtrAgent));
	  Dump_DumpBirthAndDeath (buffer);
	}

      DebugInfo ("Agent %d died - his energy was %4d\n",
		 agent, Agent_GetEnergy (PtrAgent));
      state = AGENT_SCHLAPP;
    }
  else
    {
      DebugInfo ("Agent %d still alive.\n", agent);
      state = AGENT_ALIVE;
    }

  return state;
}

/* ----------------------------------------------------------------- */

IterStateType 
Agent_InitFriendIteration (AgentType * PtrAgent)
{
  if (PtrAgent->FriendList)
    return AgentList_InitIteration (PtrAgent->FriendList);
  else
    return NULL;
}

AgentIdType 
Agent_GetNextFriend (AgentType * PtrAgent, IterStateType * PtrState)
{
  if (PtrState && *PtrState)
    return AgentList_Next (PtrAgent->FriendList, PtrState);
  else
    return 0;
}

AgentIdType 
Agent_LooseFriend (AgentType * PtrAgent)
{
  AgentIdType Friend;

  if (PtrAgent->NumberOfFriends > 0)
    {
      Friend = AgentList_RemoveFront (PtrAgent->FriendList);
      PtrAgent->NumberOfFriends--;
      return Friend;
    }
  else
    return 0;
}

bool 
Agent_AddFriend (AgentType * PtrAgent, AgentIdType Friend)
{
  if (PtrAgent->NumberOfFriends >= PtrAgent->MaxFriends)
    Agent_LooseFriend (PtrAgent);

  if (!PtrAgent->FriendList)
    {
      PtrAgent->FriendList = AgentList_CreateList ();
      PtrAgent->NumberOfFriends = 0;
    }
  AgentList_Append (PtrAgent->FriendList, Friend);
  PtrAgent->NumberOfFriends++;
  return 1;
}

bool 
Agent_IsFriendOf (AgentType * PtrAgent, AgentIdType Friend)
{
  if (PtrAgent->FriendList)
    return AgentList_IsMember (PtrAgent->FriendList, Friend);
  else
    return 0;
}

bool 
Agent_MeetFriend (AgentType * PtrAgent, AgentIdType Friend)
{
  if (Agent_IsFriendOf (PtrAgent, Friend))
    {
      AgentList_Remove (PtrAgent->FriendList, Friend);
      AgentList_Append (PtrAgent->FriendList, Friend);
      return 1;
    }
  else
    return 0;
}

bool 
Agent_MakeFriend (AgentType * PtrAgent, AgentIdType Friend)
{
  if (Agent_IsFriendOf (PtrAgent, Friend))
    return Agent_MeetFriend (PtrAgent, Friend);
  else
    return Agent_AddFriend (PtrAgent, Friend);
}

/* ----------------------------------------------------------------- */


bool 
Agent_SetParameterValueForAgent (AgentType * PtrAgent,
				 const string Parameter,
				 const string Value)
{
  if (0 == strcmp (Parameter, "Energy"))
    Agent_SetEnergy (PtrAgent, (EnergyType) atoi (Value));
  else if (0 == strcmp (Parameter, "Age"))
    Agent_SetAge (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "Attacks"))
    Agent_SetAttacks (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "Idle"))
    Agent_SetIdle (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "Memory"))
    Agent_SetMemory (PtrAgent, (MemoryType) atoi (Value));
  else if (0 == strcmp (Parameter, "XPosition"))
    Agent_SetXPosition (PtrAgent, atoi (Value));
  else if (0 == strcmp (Parameter, "YPosition"))
    Agent_SetYPosition (PtrAgent, atoi (Value));
  else if (0 == strcmp (Parameter, "Orientation"))
    Agent_SetOrientation (PtrAgent, (OrientationType) atoi (Value));
  else if (0 == strcmp (Parameter, "Screams"))
    Agent_SetScreamed (PtrAgent, (bool) atoi (Value));
  else if (0 == strcmp (Parameter, "MaxEnergy"))
    Agent_SetMaxEnergy (PtrAgent, (EnergyType) atoi (Value));
  else if (0 == strcmp (Parameter, "MaxFriends"))
    Agent_SetMaxFriends (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "RangeOfHearing"))
    Agent_SetRangeOfHearing (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "RangeOfVision"))
    Agent_SetRangeOfVision (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "Strength"))
    Agent_SetStrength (PtrAgent, (StrengthType) atoi (Value));
  else if (0 == strcmp (Parameter, "MinEnergyFactor"))
    Agent_SetMinEnergyFactor (PtrAgent, (FactorType) atof (Value));
  else if (0 == strcmp (Parameter, "HungerFactor"))
    Agent_SetHungerFactor (PtrAgent, (FactorType) atof (Value));
  else if (0 == strcmp (Parameter, "AttackFactor"))
    Agent_SetAttackFactor (PtrAgent, (FactorType) atof (Value));
  else if (0 == strcmp (Parameter, "DefenseFactor"))
    Agent_SetDefenseFactor (PtrAgent, (FactorType) atof (Value));
  else if (0 == strcmp (Parameter, "CostOfLiving"))
    Agent_SetCostOfLiving (PtrAgent, (EnergyType) atoi (Value));
  else if (0 == strcmp (Parameter, "NrOfChildren"))
    Agent_SetNumberOfChildren (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "MaxAge"))
    Agent_SetMaxAge (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "MinMatingAge"))
    Agent_SetMinMatingAge (PtrAgent, (CountType) atoi (Value));
  else if (0 == strcmp (Parameter, "Friends"))
    Agent_AddFriend (PtrAgent, (AgentIdType) atoi (Value));
  else if (0 == strcmp (Parameter, "MatePause"))
    Agent_SetMatePause (PtrAgent, atoi (Value));
  else if (0 == strcmp (Parameter, "LastMate"))
    Agent_SetLastMate (PtrAgent, atoi (Value));
  else
    return 0;

  return 1;
}


Constraint 
Agent_GetFriendConstraint (AgentType * PtrAgent,
			   AgentIdType FriendTypeId)
{
  if (PtrAgent->friendConstraints)
    return PtrAgent->friendConstraints[FriendTypeId];

  return POSSIBLE;
}

void 
Agent_SetFriendConstraint (AgentType * PtrAgent,
			   AgentIdType FriendTypeId, Constraint constraint)
{
  if (!PtrAgent->friendConstraints)
    {
      int i;

      if (constraint == POSSIBLE)
	return;

      PtrAgent->friendConstraints = Allocate (sizeof (Constraint) * 256);

      for (i = 0; i < 256; i++)
	PtrAgent->friendConstraints[i] = POSSIBLE;
    }

  PtrAgent->friendConstraints[FriendTypeId] = constraint;
}


Constraint 
Agent_GetMateConstraint (AgentType * PtrAgent,
			 AgentIdType MateTypeId)
{
  if (PtrAgent->mateConstraints)
    return PtrAgent->mateConstraints[MateTypeId];

  return POSSIBLE;
}

void 
Agent_SetMateConstraint (AgentType * PtrAgent,
			 AgentIdType MateTypeId, Constraint constraint)
{
  if (!PtrAgent->mateConstraints)
    {
      int i;

      if (constraint == POSSIBLE)
	return;

      PtrAgent->mateConstraints = Allocate (sizeof (Constraint) * 256);

      for (i = 0; i < 256; i++)
	PtrAgent->mateConstraints[i] = POSSIBLE;
    }

  PtrAgent->mateConstraints[MateTypeId] = constraint;
}


Constraint 
Agent_GetAttackConstraint (AgentType * PtrAgent,
			   AgentIdType AttackTypeId)
{
  if (PtrAgent->attackConstraints)
    return PtrAgent->attackConstraints[AttackTypeId];

  return POSSIBLE;
}

void 
Agent_SetAttackConstraint (AgentType * PtrAgent,
			   AgentIdType AttackTypeId, Constraint constraint)
{
  if (!PtrAgent->attackConstraints)
    {
      int i;

      if (constraint == POSSIBLE)
	return;

      PtrAgent->attackConstraints = Allocate (sizeof (Constraint) * 256);

      for (i = 0; i < 256; i++)
	PtrAgent->attackConstraints[i] = POSSIBLE;
    }

  PtrAgent->attackConstraints[AttackTypeId] = constraint;
}

Constraint 
Agent_GetEatConstraint (AgentType * PtrAgent, AgentIdType Food)
{
  if (PtrAgent->eatConstraints)
    return PtrAgent->eatConstraints[Food];

  return POSSIBLE;
}

void 
Agent_SetEatConstraint (AgentType * PtrAgent,
			AgentIdType Food, Constraint constraint)
{
  if (!PtrAgent->eatConstraints)
    {
      int i;

      if (constraint == POSSIBLE)
	return;

      PtrAgent->eatConstraints = Allocate (sizeof (Constraint) * 256);

      for (i = 0; i < 256; i++)
	PtrAgent->eatConstraints[i] = POSSIBLE;
    }

  PtrAgent->eatConstraints[Food] = constraint;
}

bool 
Agent_CanAttack (AgentType * PtrAgent, AgentIdType DefenderTypeId)
{
/*      Constraint cs;

   cs = Agent_GetAttackConstraint(PtrAgent, DefenderTypeId);
   if (cs == IMPOSSIBLE)
   fprintf(stdout, "Agent-Type %d cannot attack Agent-Type %d\n", 
   PtrAgent->Type, DefenderTypeId); */

  return Agent_GetAttackConstraint (PtrAgent, DefenderTypeId) != IMPOSSIBLE;
}

bool 
Agent_CanMate (AgentType * PtrAgent, AgentIdType MateTypeId)
{
/*      Constraint cs;

   cs = Agent_GetMateConstraint(PtrAgent, MateTypeId);
   if (cs == IMPOSSIBLE)
   fprintf(stdout, "Agent-Type %d cannot mate Agent-Type %d\n", 
   PtrAgent->Type, MateTypeId); */

  return Agent_GetMateConstraint (PtrAgent, MateTypeId) != IMPOSSIBLE;
}

bool 
Agent_CanBeFriend (AgentType * PtrAgent, AgentIdType FriendTypeId)
{
/*      Constraint cs;

   cs = Agent_GetFriendConstraint(PtrAgent, FriendTypeId);
   if (cs == IMPOSSIBLE)
   fprintf(stdout, "Agent-Type %d cannot be a friend of Agent-Type %d\n", 
   FriendTypeId, PtrAgent->Type); */

  return Agent_GetFriendConstraint (PtrAgent, FriendTypeId) != IMPOSSIBLE;
}

bool 
Agent_CanEat (AgentType * PtrAgent, AgentIdType Food)
{
  return Agent_GetEatConstraint (PtrAgent, Food) != IMPOSSIBLE;
}
