#ifndef AGENT_H
#define AGENT_H

/*-module-*****************************************************************
 * 
 * Modul     : Agent(Funktionsmodul-Definitionen)
 *
 * Beschreib.: Diese Header-Datei dient zur Bereitstellung der
 *             Exportschnittstelle der gleichnamigen C-Datei.
 *             Sie muss in den Modulen includet werden, in denen
 *             die beschriebenen Ressourcen importiert werden sollen.
 *
 * Autor     : <autor>
 *
 * Datum     : <datum>
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/


/*-includes-*/

#include "AgentTypeDefs.h"
#include "AgentList.h"

/*--*/

/*-defines-*/

/*-import-*/
/*--*/

/*-typedefs-*/

typedef enum
  {
    IMPOSSIBLE = 0, POSSIBLE = 1, AUTOMATIC = 2,
    INVALID
  }
Constraint;


/*typedef void  AgentType; */
typedef struct _AgentType
  {
    AgentIdType Id;
    AgentIdType Type;

    /* Status */
    EnergyType Energy;
    CountType Age;
    MemoryType Memory;
    PositionType Position;
    OrientationType Orientation;

    CountType Attacks;
    CountType Screams;
    CountType LastMate;		/* added by Martin */
    CountType Idle;

    FriendListType *FriendList;
    CountType NumberOfFriends;

    /* Attributes */
    EnergyType MaxEnergy;
    CountType MaxFriends;
    CountType RangeOfVision;
    CountType RangeOfHearing;
    StrengthType Strength;
    CountType MatePause;

    Constraint *friendConstraints;
    Constraint *mateConstraints;
    Constraint *attackConstraints;
    Constraint *eatConstraints;
/*      Constraint*                     functionConstraints; */

    FactorType MinEnergyFactor;
    FactorType HungerFactor;
    FactorType AttackFactor;
    FactorType DefenseFactor;
    EnergyType CostOfLiving;
    CountType NumberOfChildren;
    CountType MaxAge;
    CountType MinMatingAge;

    /* Program */
    ProgramType *Program;
  }
AgentType;


/*-export-*/

AgentType *Agent_CreateAgent ();
void Agent_DestroyAgent (AgentType * agent);
void Agent_CopyAgent (AgentType * dest, const AgentType * src);

bool Agent_SetParameterValueForAgent (AgentType * PtrAgent,
				      const string Parameter,
				      const string Value);

void Agent_SetId (AgentType * PtrAgent, AgentIdType AgentId);
AgentIdType Agent_GetId (AgentType * PtrAgent);

void Agent_SetType (AgentType * PtrAgent, AgentIdType TypeId);
AgentIdType Agent_GetType (AgentType * PtrAgent);

void Agent_SetEnergy (AgentType * PtrAgent, EnergyType Energy);
EnergyType Agent_GetEnergy (AgentType * PtrAgent);

void Agent_SetAge (AgentType * PtrAgent, CountType Age);
CountType Agent_GetAge (AgentType * PtrAgent);

void Agent_SetAttacks (AgentType * PtrAgent, CountType Attacks);
CountType Agent_GetAttacks (AgentType * PtrAgent);

void Agent_SetIdle (AgentType * PtrAgent, CountType Idle);
CountType Agent_GetIdle (AgentType * PtrAgent);

void Agent_SetMemory (AgentType * PtrAgent, MemoryType Memory);
MemoryType Agent_GetMemory (AgentType * PtrAgent);

void Agent_SetPosition (AgentType * PtrAgent, PositionType Position);
PositionType Agent_GetPosition (AgentType * PtrAgent);

void Agent_SetXPosition (AgentType * PtrAgent, int Position);
void Agent_SetYPosition (AgentType * PtrAgent, int Position);

ProgramType *Agent_GetProgram (AgentType * PtrAgent);
void Agent_SetProgram (AgentType * PtrAgent, ProgramType * Program);

CountType Agent_GetProgramLength (AgentType * PtrAgent);
void Agent_SetProgramLength (AgentType * PtrAgent, CountType Length);

void Agent_SetOrientation (AgentType * PtrAgent,
			   OrientationType Orientation);
OrientationType Agent_GetOrientation (AgentType * PtrAgent);

void Agent_SetScreams (AgentType * PtrAgent, CountType Screams);
CountType Agent_GetScreams (AgentType * PtrAgent);
void Agent_SetScreamed (AgentType * PtrAgent, bool Screamed);
bool Agent_GetScreamed (AgentType * PtrAgent);

void Agent_SetMaxEnergy (AgentType * PtrAgent, EnergyType MaxEnergy);
EnergyType Agent_GetMaxEnergy (AgentType * PtrAgent);

void Agent_SetMaxFriends (AgentType * PtrAgent, CountType MaxFriends);
CountType Agent_GetMaxFriends (AgentType * PtrAgent);

void Agent_SetRangeOfVision (AgentType * PtrAgent,
			     CountType RangeOfVision);
CountType Agent_GetRangeOfVision (AgentType * PtrAgent);

void Agent_SetRangeOfHearing (AgentType * PtrAgent,
			      CountType RangeOfHearing);
CountType Agent_GetRangeOfHearing (AgentType * PtrAgent);

void Agent_SetScreams (AgentType * PtrAgent, CountType Screams);
CountType Agent_GetScreams (AgentType * PtrAgent);

void Agent_SetStrength (AgentType * PtrAgent, StrengthType Strength);
StrengthType Agent_GetStrength (AgentType * PtrAgent);

void Agent_SetMinEnergyFactor (AgentType * PtrAgent,
			       FactorType MinEnergyFactor);
FactorType Agent_GetMinEnergyFactor (AgentType * PtrAgent);

void Agent_SetHungerFactor (AgentType * PtrAgent,
			    FactorType HungerFactor);
FactorType Agent_GetHungerFactor (AgentType * PtrAgent);

void Agent_SetAttackFactor (AgentType * PtrAgent,
			    FactorType AttackFactor);
FactorType Agent_GetAttackFactor (AgentType * PtrAgent);

void Agent_SetDefenseFactor (AgentType * PtrAgent,
			     FactorType DefenseFactor);
FactorType Agent_GetDefenseFactor (AgentType * PtrAgent);

void Agent_SetCostOfLiving (AgentType * PtrAgent,
			    EnergyType CostOfLiving);
EnergyType Agent_GetCostOfLiving (AgentType * PtrAgent);

void Agent_SetNumberOfChildren (AgentType * PtrAgent,
				CountType NumberOfChildren);
CountType Agent_GetNumberOfChildren (AgentType * PtrAgent);

void Agent_SetMaxAge (AgentType * PtrAgent, CountType MaxAge);
CountType Agent_GetMaxAge (AgentType * PtrAgent);

void Agent_SetMinMatingAge (AgentType * PtrAgent,
			    CountType MinMatingAge);
CountType Agent_GetMinMatingAge (AgentType * PtrAgent);

/* added by Martin */
void Agent_SetLastMate (AgentType * PtrAgent, CountType LastMate);
CountType Agent_GetLastMate (AgentType * PtrAgent);

inline CountType Agent_GetMatePause (AgentType * PtrAgent);
inline void Agent_SetMatePause (AgentType * PtrAgent,
				CountType MatePause);

AgentIdType Agent_LooseFriend (AgentType * PtrAgent);
bool Agent_IsFriendOf (AgentType * PtrAgent, AgentIdType Friend);
bool Agent_MeetFriend (AgentType * PtrAgent, AgentIdType Friend);
bool Agent_MakeFriend (AgentType * PtrAgent, AgentIdType Friend);

bool Agent_AddFriend (AgentType * PtrAgent, AgentIdType Friend);

IterStateType Agent_InitFriendIteration (AgentType * PtrAgent);
AgentIdType Agent_GetNextFriend (AgentType * PtrAgent,
				 IterStateType * PtrState);

int Agent_Timestep (AgentType * PtrAgent);

void Agent_PrintAgent (AgentType * PtrAgent, FILE * FilePointer);

Constraint Agent_GetFriendConstraint (AgentType * PtrAgent,
				      AgentIdType FriendTypeId);
void Agent_SetFriendConstraint (AgentType * PtrAgent,
			   AgentIdType FriendTypeId, Constraint constraint);

Constraint Agent_GetMateConstraint (AgentType * PtrAgent,
				    AgentIdType FriendTypeId);
void Agent_SetMateConstraint (AgentType * PtrAgent,
			   AgentIdType FriendTypeId, Constraint constraint);

Constraint Agent_GetAttackConstraint (AgentType * PtrAgent,
				      AgentIdType FriendTypeId);
void Agent_SetAttackConstraint (AgentType * PtrAgent,
			   AgentIdType FriendTypeId, Constraint constraint);

Constraint Agent_GetEatConstraint (AgentType * PtrAgent, AgentIdType Food);
void Agent_SetEatConstraint (AgentType * PtrAgent,
			     AgentIdType Food, Constraint constraint);


bool Agent_CanAttack (AgentType * PtrAgent,
		      AgentIdType DefenderTypeId);
bool Agent_CanMate (AgentType * PtrAgent, AgentIdType MateTypeId);
bool Agent_CanBeFriend (AgentType * PtrAgent,
			AgentIdType FriendTypeId);
bool Agent_CanEat (AgentType * PtrAgent, AgentIdType Food);

inline void Agent_IncreaseAge (AgentType * PtrAgent, CountType Timesteps);
#endif
