


#include <stdlib.h>
#include <varargs.h>
#include "AgentSet.h"
#include "Agent.h"
#include "../program/AgentPrgFunc.h"
#include "../interpreter/Interpreter.h"
#include "../dump/Dump.h"
#include "../misc/DynaString.h"
#include "AgentTypeDefs.h"
#include "AgentList.h"

#define TRUE    1

#define INITIAL_BUFFER_SIZE        1024

static int DL = 0;

/* -------------------------------------------------------------------------- */

void 
AddIntTo (DynaStringType * BufferPtr, const string PtrPrefix, int Integer)
{
  if (PtrPrefix)
    DynaString_Append (BufferPtr, PtrPrefix);

  if (strlen (PtrPrefix) < 9)
    {
      DynaString_Append (BufferPtr, "\t\t");
    }
  else
    {
      DynaString_Append (BufferPtr, "\t");
    }

  DynaString_AppendInt (BufferPtr, Integer);
  DynaString_Append (BufferPtr, "\n");
}

/* -------------------------------------------------------------------------- */

void 
AddFloatTo (DynaStringType * BufferPtr, const string PtrPrefix, float Float)
{
  if (PtrPrefix)
    DynaString_Append (BufferPtr, PtrPrefix);

  if (strlen (PtrPrefix) < 9)
    {
      DynaString_Append (BufferPtr, "\t\t");
    }
  else
    {
      DynaString_Append (BufferPtr, "\t");
    }

  DynaString_AppendFloat (BufferPtr, Float);
  DynaString_Append (BufferPtr, "\n");
}

/* -------------------------------------------------------------------------- */

void 
AgentSet_FillBufferWithAttributesOfAgent (DynaStringType * BufferPtr,
					  AgentIdType Agent)
{
  DynaString_Append (BufferPtr, "\tATTRIBUTES\n");

  AddIntTo (BufferPtr, "\t\tMaxEnergy = ", AgentSet_GetMaxEnergy (Agent));
  AddIntTo (BufferPtr, "\t\tMaxFriends = ", AgentSet_GetMaxFriends (Agent));
  AddIntTo (BufferPtr, "\t\tRangeOfVision = ", AgentSet_GetVisionRange (Agent));
  AddIntTo (BufferPtr, "\t\tRangeOfHearing = ", AgentSet_GetHearingRange (Agent));
  AddIntTo (BufferPtr, "\t\tStrength = ", AgentSet_GetStrength (Agent));
  AddFloatTo (BufferPtr, "\t\tMinEnergyFactor = ", AgentSet_GetMinEnergyFactor (Agent));

  DynaString_Append (BufferPtr, "\t\tEND_ATTRIBUTES\n");
}

/* -------------------------------------------------------------------------- */

void 
AgentSet_FillBufferWithStatusOfAgent (DynaStringType * BufferPtr,
				      AgentIdType Agent)
{
  PositionType position;
  IterStateType friendIter;
  AgentIdType friend;

  DynaString_Append (BufferPtr, "\tSTATUS\n");

  AddIntTo (BufferPtr, "\t\tEnergy = ", AgentSet_GetEnergy (Agent));
  AddIntTo (BufferPtr, "\t\tAge = ", AgentSet_GetAge (Agent));
  AddIntTo (BufferPtr, "\t\tAttacks = ", AgentSet_GetAttacks (Agent));
  AddIntTo (BufferPtr, "\t\tMemory = ", AgentSet_GetMemory (Agent));

  position = AgentSet_GetPosition (Agent);

  AddIntTo (BufferPtr, "\t\tXPosition = ", position.x);
  AddIntTo (BufferPtr, "\t\tYPosition = ", position.y);

  /* DynaString_Append (BufferPtr, "NumberOfFriends =");
     DynaString_AppendInt(BufferPtr, AgentSet_GetNumberOfFriends(Agent)); */

  friendIter = AgentSet_InitFriendIteration (Agent);
  while (friend = AgentSet_GetNextFriend (Agent, &friendIter))
    {
      AddIntTo (BufferPtr, "\t\tFriends = ", friend);
    }

  AddIntTo (BufferPtr, "\t\tOrientation = ", AgentSet_GetOrientation (Agent));
  AddIntTo (BufferPtr, "\t\tScreams = ", AgentSet_GetScreams (Agent));

  DynaString_Append (BufferPtr, "\t\tEND_STATUS\n");
}

/* -------------------------------------------------------------------------- */

void 
AgentSet_DumpSection ()
{
  AgentIdType agent;
  IterStateType iter;
  DynaStringType *bufferPtr;
  ProgramType *program;
  string cstring;

  iter = AgentSet_InitAgentIteration ();

  bufferPtr = DynaString_Alloc (INITIAL_BUFFER_SIZE);

  while (agent = AgentSet_GetNextAgent (&iter))
    {
      DebugLevel (DL, "Dumping Agent %d\n", agent);

      DynaString_Reset (bufferPtr);
      AddIntTo (bufferPtr, "AGENT ", agent);

      Dump_DumpAgent (DynaString_CString (bufferPtr));

      DynaString_Reset (bufferPtr);

      if (Dump_WantDumpAgentAttributes ())
	{
	  AgentSet_FillBufferWithAttributesOfAgent (bufferPtr, agent);
	  cstring = DynaString_CString (bufferPtr);
	  Dump_DumpAgentAttributes (cstring);
	}
      if (Dump_WantDumpAgentStatus ())
	{
	  DynaString_Reset (bufferPtr);
	  AgentSet_FillBufferWithStatusOfAgent (bufferPtr, agent);
	  cstring = DynaString_CString (bufferPtr);
	  Dump_DumpAgentStatus (cstring);
	}

      if (Dump_WantDumpAgentProgram ())
	{
	  ProgramType *program;
	  string program_string = NULL;

	  program = AgentSet_GetProgram (agent);
	  program_string = Interpreter_UnparseProgram (program);
	  if (program_string)
	    {
	      Dump_DumpAgentProgram (program_string);
	      free (program_string);
	    }
	}

      Dump_DumpAgent ("END_AGENT\n\n");
    }

  free (DynaString_CStringAndFree (bufferPtr));
}
