%option lex-compat nostdinit stack
%x	COM MCS FCS ACS ECS

%{

/*_______________________________________________________________________

  DATEI    : AgentParser.l
  AUTOR    : THORSTEN KOHNHORST
  FUNKTION : FLEX-Datei zur Analyse und Auslesen der AgentsSection

_________________________________________________________________________*/

#include  "../misc/TypeDefs.h"
#include  "AgentParser.h"
#include  "AgentSet.h"
#include  "../interpreter/Interpreter.h"
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>

/* Definitions and Deklarations ---------------------------------------- */

#ifndef   FALSE
 #define  FALSE (1 == 0)
 #define  TRUE  (1 == 1)
#endif

int AGENTSparse();

static void ReturnBufferToFile(FILE *File, YY_BUFFER_STATE Buf);
static void SetConstraintFromString(YYSTYPE* val, const string String);

/* LEX-Rules ----------------------------------------------------------- */
%}

string              [a-zA-Z0-9_]+
fliesskomma         [+-]?[0-9]*\.[0-9]*
ganzzahl            [+-]?[0-9]+

constraint			[+\-\&][A-Za-z\*]

%%
<COM>.*"END_COMMENT" { yy_pop_state(); }
<COM>.*             |
<COM>\n             ;


"COMMENT"           { yy_push_state(COM); }

\#.*\n              | 
\n                  |
\t                  /* */
		    ;

[=]                 { return (yytext[0]); 

/* Sections-Token ------------------------------------------------------ */
		    }
"AGENT"             { return (AGENT); }
"END_AGENT"         { return (END_AGENT); }
"AGENTS"            { return (AGENTS); }
"END_AGENTS"        {
                      ReturnBufferToFile (AGENTSin, YY_CURRENT_BUFFER);
                      yyterminate(); 
                    }

"NUMBER"			{ return NUMBER; }
"Number"			{ return NUMBER; }

"FRIEND-CONSTRAINTS" { yy_push_state(FCS); return CONSTRAINT; }
"MATE-CONSTRAINTS"	{ yy_push_state(MCS); return CONSTRAINT; }
"ATTACK-CONSTRAINTS" { yy_push_state(ACS); return CONSTRAINT; }
"EAT-CONSTRAINTS"	{ yy_push_state(ECS); return CONSTRAINT; }

"ATTRIBUTES"        { return (ATTRIBUTES); }
"END_ATTRIBUTES"    { return (END_ATTRIBUTES); }
"STATUS"            { return (STATUS); }
"END_STATUS"        { return (END_STATUS); }
"PROGRAM"           {
                      ProgramType* ptp;
                      ReturnBufferToFile (AGENTSin, YY_CURRENT_BUFFER);
                        ptp = Interpreter_ParseProgram(AGENTSin);     
                        AGENTSlval.typProgramPtr = ptp;  
                      YY_FLUSH_BUFFER;
                      return(PROGRAM);
                    }
"TYPE"              { return (TYPE); }



<FCS>{constraint}	{
						SetConstraintFromString(&AGENTSlval, yytext);
						return FRIEND_CONSTRAINT;
					}

<MCS>{constraint}	{
						SetConstraintFromString(&AGENTSlval, yytext);
						return MATE_CONSTRAINT;
					}

<ACS>{constraint}	{
						SetConstraintFromString(&AGENTSlval, yytext);
						return ATTACK_CONSTRAINT;
					}

<ECS>{constraint}	{
						SetConstraintFromString(&AGENTSlval, yytext);
						return EAT_CONSTRAINT;
					}

<ACS,FCS,MCS,ECS>\n		{
						yy_pop_state();
					}

{fliesskomma}       {
                      sscanf(yytext, "%f", &AGENTSlval.typflt);
                      return(FLT);
                    }
{ganzzahl}          {
                      AGENTSlval.typint=atoi(yytext);	
                      return(INT);
                    }
{string}            {	
                      strncpy(AGENTSlval.typstr, yytext, 255);
                      AGENTSlval.typstr[255]='\0';
                      return (STR);
                    }

\n					|
.                   {}
%%

static void SetConstraintFromString(YYSTYPE* val, const string String)
{
	switch (String[0])
	{
		case '-':
			val->typconstraint.value = IMPOSSIBLE;
		break;
		case '+':
			val->typconstraint.value = POSSIBLE;
		break;
		case '&':
			val->typconstraint.value = AUTOMATIC;
		break;
	}

	switch (String[1])
	{
		case '*':
			val->typconstraint.id = -1;
		break;
		default:
			val->typconstraint.id = (int)yytext[1];
		break;
	}
}

/* --------------------------------------------------------------------- */

int AGENTSwrap ()
{
  return (1);
}

/* --------------------------------------------------------------------- */

static void ReturnBufferToFile(FILE *File, YY_BUFFER_STATE Buf)
{
  /* 
  Diese Funktion liest den uebergebenen lex-Buffer aus
  und haengt alle Zeichen, die aus dem Stream file
  ausgelesen, jedoch noch nicht gematcht wurden, wieder
  in den Stream zurueck. 
  */

  int  length; 
  char *begin;

  length = strlen (Buf->yy_ch_buf);

  begin = &(Buf->yy_ch_buf[length+1]);
 
  length = strlen (begin); 

  fseek (File, -length, 1);
}

/* ---------------------------------------------------------------------- */

bool AgentSet_ParseSection(FILE* File)
{
	if (AGENTSin)
	{
		yyrestart(File);
		YY_FLUSH_BUFFER;
	}
	else
		AGENTSin = File;

	if (0 != AGENTSparse())
	{
		fprintf(stderr, "ERROR[AgentSet_ParseSection] -- ");
		fprintf(stderr, "Fehler beim Parsen der Datei.\n");
		return (FALSE);
	}
	else
		return (TRUE);
}







