%{
/*____________________________________________________________________________

  DATEI    : AgentParser.y
  AUTOR    : THORSTEN KOHNHORST
  FUNKTION : YACC-Datei (BISON) zur Analyse und Auslesen einer AgentsSection

_____________________________________________________________________________*/

#include <stdio.h>
#include <stdlib.h>

#include "../misc/Messages.h"

#include "AgentTypeDefs.h"
#include "AgentSet.h"

#include "../program/AgentPrgFunc.h"

/* -------------------------------------------------------------------- */

#ifndef FALSE
#define FALSE (1 == 0)
#endif
#ifndef TRUE
#define TRUE  (1 == 1)
#endif

#ifdef YYDEBUG
  int yydebug = 1;
#endif

extern int   AGENTSlineno;
extern FILE* AGENTSin;

int AGENTSerror(const string);
int AGENTSlex();

/* -------------------------------------------------------------------- */

/* In dieser Variablen wird festgehalten, ob es sich beim aktuellen     */
/* Agenten um einen Typ oder einen Agenten mit Id handelt, da der Agent */
/* erst nach dem Schluesselwort TYPE (falls vorhanden) erzeugt wird.    */ 

int GlobalKindOfActAgent; 

/* In dieser Variablen wird die Id des aktuell erzeugten Agenten ge-    */
/* speichert, da sie fuer spaetere Parameterzuweisungen benoetigt wird. */

AgentIdType GlobalActAgentId;

/* -------------------------------------------------------------------- */
%}

%union{
        float	typflt;
        int		typint;
		struct _cst
		{
			int	id;
			int	value;
		}		typconstraint;
        char	typstr[1025];
		ProgramType *typProgramPtr;
      }

%token <typstr> STR 
%token <typstr> DAT        
%token <typstr> UHR
%token <typflt> FLT
%token <typint> INT

%token	NUMBER
%token  AGENT
%token  AGENTS
%token  ATTRIBUTES
%token  END_AGENT
%token  END_ATTRIBUTES
%token  END_AGENTS
%token  END_STATUS
%token  <typProgramPtr> PROGRAM
%token  STATUS
%token  TYPE

%token	CONSTRAINT
%token	<typconstraint>	FRIEND_CONSTRAINT
%token	<typconstraint>	MATE_CONSTRAINT
%token	<typconstraint>	ATTACK_CONSTRAINT
%token	<typconstraint>	EAT_CONSTRAINT

%type   <typstr> value
%type   <typint> id_or_type type
%%
 
agents_section          :	/*	*/
						|	AGENTS
							agent_sections
                        ;

agent_sections			:	/*	*/
						|	agent_sections
							agent_section
						;

agent_section			:	AGENT id_or_type type 
							{
								if (GlobalKindOfActAgent == AGENT_TYPE)
								{
									GlobalActAgentId = 
										AgentSet_CreateAgentTypeOfType($2, $3);  
								}
								else
								{
									GlobalActAgentId =
										AgentSet_CreateAgentOfType ($2, $3);
								}
							}
							type_info
							attributes
							status
							program
							END_AGENT
                        ;

id_or_type              :	/*	*/
							{
								/* Der Agent hat keine Id und keinen Typ */ 
								$$ = 0; 
								GlobalKindOfActAgent = AGENT_WITH_ID; 
							}
                        |	STR
							{ 
								/* Der Abschnitt beschreibt einen Typ    */
								$$ = (int)*$1;
								GlobalKindOfActAgent = AGENT_TYPE; 
							}
                        |	INT 
							{
								/* Der Abschnitt beschreibt Agent mit Id */
								$$ =  $1; 
								GlobalKindOfActAgent = AGENT_WITH_ID;
							}
						;

type                    :	/*	*/
							{
								/* Der aktuelle Agent ist von keinem Typ */
								$$ = 0;
							}
                        |	TYPE STR
							{
								/* Der altuelle Agent ist von einem Typ  */
								$$ = (int)*$2;
							}
						;                        

type_info				:	/*	*/
						|	type_info NUMBER '=' INT
							{
								if (GlobalKindOfActAgent == AGENT_TYPE)
								{
									AgentSet_SetNumberOfAgentsOfType
														(GlobalActAgentId, $4);
								}
								else
								{
									Warning("AGENTparser", 
											"constraint for agent given\n");
								}
							}
						|	type_info CONSTRAINT constraints
						;

constraints				:	/*	*/
						|	constraints constraint
						;

constraint				:	FRIEND_CONSTRAINT
							{
/*								if (GlobalKindOfActAgent == AGENT_TYPE)
								{*/
									if ($1.id > 0)
									{
										AgentSet_SetFriendConstraint
											(GlobalActAgentId,$1.id,$1.value);
									}
									else
									{
										int i;

										for (i = 0; i < 256; i++)
										{
											AgentSet_SetFriendConstraint
												(GlobalActAgentId,i,$1.value);
										}
									}
/*								}
								else
								{
									Warning("AGENTparser", 
											"constraint for agent given\n");
								}*/
							}
						|	MATE_CONSTRAINT
							{
/*								if (GlobalKindOfActAgent == AGENT_TYPE)
								{*/
									if ($1.id > 0)
									{
										AgentSet_SetMateConstraint
											(GlobalActAgentId,$1.id,$1.value);
									}
									else
									{
										int i;

										for (i = 0; i < 256; i++)
										{
											AgentSet_SetMateConstraint
												(GlobalActAgentId,i,$1.value);
										}
									}
/*								}
								else
								{
									Warning("AGENTparser", 
											"constraint for agent given\n");
								}*/
							}
						|	ATTACK_CONSTRAINT
							{
/*								if (GlobalKindOfActAgent == AGENT_TYPE)
								{*/
									if ($1.id > 0)
									{
										AgentSet_SetAttackConstraint
											(GlobalActAgentId,$1.id,$1.value);
									}
									else
									{
										int i;

										for (i = 0; i < 256; i++)
										{
											AgentSet_SetAttackConstraint
												(GlobalActAgentId,i,$1.value);
										}
									}
/*								}
								else
								{
									Warning("AGENTparser", 
											"constraint for agent given\n");
								}*/
							}
						|	EAT_CONSTRAINT
							{
								if ($1.id > 0)
								{
									AgentSet_SetEatConstraint
										(GlobalActAgentId,$1.id,$1.value);
								}
								else
								{
									int i;

									for (i = 0; i < 256; i++)
									{
										AgentSet_SetEatConstraint
											(GlobalActAgentId,i,$1.value);
									}
								}
							}
						;

attributes				:	/* */
                        |	ATTRIBUTES
							attribute_list
							END_ATTRIBUTES
						;

attribute_list          :       /* */
                        |	attribute_list
							STR '=' value 
							{ 
								AgentSet_SetParameterValueForAgent(
									GlobalActAgentId, $2, $4,
										GlobalKindOfActAgent); 
							}
                        ;
 

status                  :	/* */
						|	STATUS
							status_list
							END_STATUS 
						;

status_list             :	/* */
                        |	status_list
							STR '=' value
							{ 
								AgentSet_SetParameterValueForAgent(
									GlobalActAgentId, $2, $4, 
										GlobalKindOfActAgent); 
							}
						;

program                 :	/* */
                        |	PROGRAM 
							{
								AgentSet_SetProgramForKind(GlobalActAgentId, 
													$1, GlobalKindOfActAgent);
							}
						;

value					:	STR {strcpy  ($$, $1);}
                        |	INT {sprintf ($$, "%d", $1);}
                        |	FLT {sprintf ($$, "%f", $1);}
						;

%%

int AGENTSerror (char *s)
{
	fprintf(stderr, "Error[AgentSet_ParseSection] %s, line %d\n", 
 				s, AGENTSlineno);
	return 1;
}
