/*-module-*****************************************************************
 *
 * Modul     : AgentSet (FUNKTIONSMODUL)
 *
 * Beschreib.:
 *
 * Autor     : <autor>
 *
 * Datum     : <datum>
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/


/*-includes-*/

#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"
#include "../misc/Memory.h"
#include "../misc/DynaString.h"

#include "AgentSet.h"
#include "AgentSets.h"

#include "../interpreter/Interpreter.h"
#include "../world/World.h"
#include "../dump/Dump.h"
#include "../random/Random.h"

static int DL;

static int AgentSet_Initialized = 0;

static AgentType *DefaultAgent = NULL;
static AgentSet *AgentTypes = NULL;
static AgentSet *Agents = NULL;

static int NumberOfAgents = -1;
static DictType *AgentTypeInfos = NULL;
static AgentTypeInfo DefaultTypeInfo =
{0, 0};

#define INITIAL_BUFFER_SIZE        1024
static DynaStringType *DumpBuffer = NULL;

static AgentType *CreateAgentOfType (AgentIdType TypeId);
static AgentTypeInfo *GetAgentTypeInfo (AgentIdType TypeId, bool Create);
/*****************************************************************************/

inline AgentIdType 
AgentSet_GetType (AgentIdType AgentId)
{
  return Agent_GetType (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetType (AgentIdType AgentId, AgentIdType TypeId)
{
  return Agent_SetType (AgentSets_GetAgent (Agents, AgentId), TypeId);
}


inline CountType 
AgentSet_GetEnergy (AgentIdType AgentId)
{
  return Agent_GetEnergy (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetEnergy (AgentIdType AgentId, EnergyType Energy)
{
  AgentType *agent;

  agent = AgentSets_GetAgent (Agents, AgentId);

  if (Energy > Agent_GetMaxEnergy (agent))
    Energy = Agent_GetMaxEnergy (agent);

  Agent_SetEnergy (agent, Energy);
}


inline CountType 
AgentSet_GetAge (AgentIdType AgentId)
{
  return Agent_GetAge (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetAge (AgentIdType AgentId, CountType Age)
{
  if (Age >= 0)
    Agent_SetAge (AgentSets_GetAgent (Agents, AgentId), Age);
  else
    Warning ("AgentSet_SetAge", "age of agent %d < 0\n", AgentId);
}


inline MemoryType 
AgentSet_GetMemory (AgentIdType AgentId)
{
  return Agent_GetMemory (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetMemory (AgentIdType AgentId, MemoryType Memory)
{
  Agent_SetMemory (AgentSets_GetAgent (Agents, AgentId), Memory);
}


inline PositionType 
AgentSet_GetPosition (AgentIdType AgentId)
{
  return Agent_GetPosition (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetPosition (AgentIdType AgentId, PositionType Position)
{
  Agent_SetPosition (AgentSets_GetAgent (Agents, AgentId), Position);
}


inline OrientationType 
AgentSet_GetOrientation (AgentIdType AgentId)
{
  return Agent_GetOrientation (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_ChangeOrientation (AgentIdType AgentId,
			    OrientationType Orientation)
{
  Orientation = AgentSet_GetOrientation (AgentId) + Orientation;
  AgentSet_SetOrientation (AgentId, Orientation);
}

inline void 
AgentSet_SetOrientation (AgentIdType AgentId,
			 OrientationType Orientation)
{
  if ((Orientation % 45) == 0)
    {
      if (Orientation < 0)
	Orientation = 360 - (Orientation % 360);
      else
	Orientation = Orientation % 360;

      Agent_SetOrientation (AgentSets_GetAgent (Agents, AgentId), Orientation);
    }
  else
    Warning ("AgentSet_SetOrientation", "Orientation not multiple of 45!\n");
}


inline CountType 
AgentSet_GetAttacks (AgentIdType AgentId)
{
  return Agent_GetAttacks (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetAttacks (AgentIdType AgentId, CountType Attacks)
{
  Agent_SetAttacks (AgentSets_GetAgent (Agents, AgentId), Attacks);
}

inline CountType 
AgentSet_GetIdle (AgentIdType AgentId)
{
  return Agent_GetIdle (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetIdle (AgentIdType AgentId, CountType Idle)
{
  Agent_SetIdle (AgentSets_GetAgent (Agents, AgentId), Idle);
}

inline CountType 
AgentSet_GetScreams (AgentIdType AgentId)
{
  return Agent_GetScreams (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetScreams (AgentIdType AgentId, CountType Screams)
{
  Agent_SetScreams (AgentSets_GetAgent (Agents, AgentId), Screams);
}

/*****************************************************************************/

inline IterStateType 
AgentSet_InitFriendIteration (AgentIdType AgentId)
{
  return Agent_InitFriendIteration (AgentSets_GetAgent (Agents, AgentId));
}

inline AgentIdType 
AgentSet_GetNextFriend (AgentIdType AgentId,
			IterStateType * PtrState)
{
  AgentType *agent;
  AgentIdType friendId;

  agent = AgentSets_GetAgent (Agents, AgentId);
  while ((friendId = Agent_GetNextFriend (agent, PtrState)) != 0)
    {
      if (AgentSets_GetAgent (Agents, friendId) != NULL)
	break;
    }

  return friendId;
}

inline bool 
AgentSet_AddToFriends (AgentIdType AgentId, AgentIdType FriendId)
{
  return AgentSet_MakeFriend (AgentId, FriendId);
}

inline bool 
AgentSet_MakeFriend (AgentIdType AgentId, AgentIdType FriendId)
{
  if (Agent_CanBeFriend (AgentSets_GetAgent (Agents, AgentId),
			 AgentSet_GetType (FriendId)))
    return Agent_MakeFriend (AgentSets_GetAgent (Agents, AgentId), FriendId);
  else
    return FALSE;
}

inline bool 
AgentSet_IsFriendOf (AgentIdType AgentId, AgentIdType FriendId)
{
  if (AgentSet_GetFriendConstraint (AgentSet_GetType (AgentId),
				  AgentSet_GetType (FriendId)) != AUTOMATIC)
    return Agent_IsFriendOf (AgentSets_GetAgent (Agents, AgentId), FriendId);
  else
    return TRUE;
}

inline bool 
AgentSet_MeetFriend (AgentIdType AgentId, AgentIdType FriendId)
{
  return Agent_MeetFriend (AgentSets_GetAgent (Agents, AgentId), FriendId);
}

/*****************************************************************************/


inline EnergyType 
AgentSet_GetMaxEnergy (AgentIdType AgentId)
{
  return Agent_GetMaxEnergy (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetMaxEnergy (AgentIdType AgentId, EnergyType MaxEnergy)
{
  Agent_SetMaxEnergy (AgentSets_GetAgent (Agents, AgentId), MaxEnergy);
}

/* added by Martin */
inline CountType 
AgentSet_GetLastMate (AgentIdType AgentId)
{
  return Agent_GetLastMate (AgentSets_GetAgent (Agents, AgentId));
}

/* added by Martin */
inline void 
AgentSet_SetLastMate (AgentIdType AgentId, CountType LastMate)
{
  Agent_SetLastMate (AgentSets_GetAgent (Agents, AgentId), LastMate);
}


inline CountType 
AgentSet_GetMatePause (AgentIdType AgentId)
{
  return Agent_GetMatePause (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetMatePause (AgentIdType AgentId,
		       CountType MatePause)
{
  Agent_SetMatePause (AgentSets_GetAgent (Agents, AgentId), MatePause);
}


inline CountType 
AgentSet_GetMaxFriends (AgentIdType AgentId)
{
  return Agent_GetMaxFriends (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetMaxFriends (AgentIdType AgentId, CountType MaxFriends)
{
  Agent_SetMaxFriends (AgentSets_GetAgent (Agents, AgentId), MaxFriends);
}


inline CountType 
AgentSet_GetVisionRange (AgentIdType AgentId)
{
  return Agent_GetRangeOfVision (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetVisionRange (AgentIdType AgentId, CountType VisionRange)
{
  Agent_SetRangeOfVision (AgentSets_GetAgent (Agents, AgentId), VisionRange);
}

inline CountType 
AgentSet_GetHearingRange (AgentIdType AgentId)
{
  return Agent_GetRangeOfHearing (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetHearingRange (AgentIdType AgentId,
			  CountType HearingRange)
{
  Agent_SetRangeOfHearing (AgentSets_GetAgent (Agents, AgentId), HearingRange);
}


inline StrengthType 
AgentSet_GetStrength (AgentIdType AgentId)
{
  return Agent_GetStrength (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetStrength (AgentIdType AgentId, CountType Strength)
{
  Agent_SetStrength (AgentSets_GetAgent (Agents, AgentId), Strength);
}


inline CountType 
AgentSet_GetNumberOfChildren (AgentIdType AgentId)
{
  return Agent_GetNumberOfChildren (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetNumberOfChildren (AgentIdType AgentId,
			      CountType NumberOfChildren)
{
  Agent_SetNumberOfChildren (AgentSets_GetAgent (Agents, AgentId),
			     NumberOfChildren);
}


inline CountType 
AgentSet_GetMaxAge (AgentIdType AgentId)
{
  return Agent_GetMaxAge (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetMaxAge (AgentIdType AgentId, CountType MaxAge)
{
  Agent_SetMaxAge (AgentSets_GetAgent (Agents, AgentId), MaxAge);
}

inline void 
AgentSet_SetMinMatingAge (AgentIdType AgentId, CountType MinMatingAge)
{
  Agent_SetMinMatingAge (AgentSets_GetAgent (Agents, AgentId), MinMatingAge);
}
inline CountType 
AgentSet_GetMinMatingAge (AgentIdType AgentId)
{
  return Agent_GetMinMatingAge (AgentSets_GetAgent (Agents, AgentId));
}

/*****************************************************************************/


FactorType 
AgentSet_GetMinEnergyFactor (AgentIdType AgentId)
{
  return Agent_GetMinEnergyFactor (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetMinEnergyFactor (AgentIdType AgentId, FactorType MinEnergyFactor)
{
  Agent_SetMinEnergyFactor (AgentSets_GetAgent (Agents, AgentId), MinEnergyFactor);
}

inline FactorType 
AgentSet_GetHungerFactor (AgentIdType AgentId)
{
  return Agent_GetHungerFactor (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetHungerFactor (AgentIdType AgentId, FactorType HungerFactor)
{
  Agent_SetHungerFactor (AgentSets_GetAgent (Agents, AgentId), HungerFactor);
}

inline FactorType 
AgentSet_GetAttackFactor (AgentIdType AgentId)
{
  return Agent_GetAttackFactor (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetAttackFactor (AgentIdType AgentId, FactorType AttackFactor)
{
  Agent_SetAttackFactor (AgentSets_GetAgent (Agents, AgentId), AttackFactor);
}

inline FactorType 
AgentSet_GetDefenseFactor (AgentIdType AgentId)
{
  return Agent_GetDefenseFactor (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetDefenseFactor (AgentIdType AgentId, FactorType DefenseFactor)
{
  Agent_SetDefenseFactor (AgentSets_GetAgent (Agents, AgentId), DefenseFactor);
}

inline EnergyType 
AgentSet_GetCostOfLiving (AgentIdType AgentId)
{
  return Agent_GetCostOfLiving (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetCostOfLiving (AgentIdType AgentId, EnergyType CostOfLiving)
{
  Agent_SetCostOfLiving (AgentSets_GetAgent (Agents, AgentId), CostOfLiving);
}

/*****************************************************************************/

inline ProgramType *
AgentSet_GetProgram (AgentIdType AgentId)
{
  return Agent_GetProgram (AgentSets_GetAgent (Agents, AgentId));
}

void 
AgentSet_SetProgramForKind (AgentIdType AgentId,
			    ProgramType * Program, int KindOfAgent)
{
  AgentType *agent;

  if (KindOfAgent == AGENT_TYPE)
    agent = AgentSets_GetAgent (AgentTypes, AgentId);
  else
    agent = AgentSets_GetAgent (Agents, AgentId);

  Agent_SetProgram (agent, Program);
}

inline void 
AgentSet_SetProgram (AgentIdType AgentId, ProgramType * Program)
{
  Agent_SetProgram (AgentSets_GetAgent (Agents, AgentId), Program);
}

inline CountType 
AgentSet_GetProgramLength (AgentIdType AgentId)
{
  return Agent_GetProgramLength (AgentSets_GetAgent (Agents, AgentId));
}

inline void 
AgentSet_SetProgramLength (AgentIdType AgentId, CountType Length)
{
  Agent_SetProgramLength (AgentSets_GetAgent (Agents, AgentId), Length);
}


/*****************************************************************************/

void 
AgentSet_Initialize ()
{
  if (AgentSet_Initialized == 0)
    {
      DebugLevel (DL, "Initialize AgentSet\n");

      if (!DefaultAgent)
	DefaultAgent = Agent_CreateAgent ();

      if (!AgentTypes)
	AgentTypes = AgentSets_CreateAgentSet ();
      if (!Agents)
	Agents = AgentSets_CreateAgentSet ();

      if (!AgentTypeInfos)
	AgentTypeInfos = Dictionary_Create ();

      if (!DumpBuffer)
	DumpBuffer = DynaString_Alloc (INITIAL_BUFFER_SIZE);

      AgentSet_Initialized = 1;
    }
}

void 
AgentSet_FinishInitializationForType (AgentIdType TypeId)
{
  AgentIdType agentId;

  while (AgentSet_GetAgentCountOfType (TypeId) <
	 AgentSet_GetNumberOfAgentsOfType (TypeId))
    {
      agentId = AgentSet_CreateAgentOfType (0, TypeId);
    }
}

void 
AgentSet_FinishInitialization ()
{
  IterStateType state;
  AgentIdType typeId;

  state = AgentSets_InitAgentIteration (AgentTypes);
  while ((typeId = AgentSets_GetNextAgent (AgentTypes, &state)) > 0)
    {
      AgentSet_FinishInitializationForType (typeId);

      /*
         printf("AgentType %d: wanted %d current %d\n", typeId, 
         AgentSet_GetNumberOfAgentsOfType(typeId), 
         AgentSet_GetAgentCountOfType(typeId));
       */
    }

  if (NumberOfAgents < AgentSet_GetAgentCount ())
    {
      if (NumberOfAgents != -1)
	Warning ("AgentSet_FinishInitialization",
		 "WorldParameter NumberOfAgents exceeded\n");
    }
  else
    {
      DefaultTypeInfo.wantedNumber = NumberOfAgents -
	AgentSet_GetAgentCount ();
      AgentSet_FinishInitializationForType (0);
    }
  /*
     printf("wanted %d current %d\n", 
     NumberOfAgents == -1 ? AgentSet_GetAgentCount() : NumberOfAgents, 
     AgentSet_GetAgentCount());
   */
}

/*-ftn-********************************************************************
 *
 * Funktion  : AgentSet_EndSimulation
 *
 * Beschreib.:
 *
 * Parameter :
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
AgentSet_EndSimulation ()
{
  AgentIdType typeId;
  IterStateType state;
  AgentTypeInfo *info;

  AgentSet_DumpSection ();

  AgentSets_DestroyAgentSet (Agents);


  state = AgentSets_InitAgentIteration (AgentTypes);
  while ((typeId = AgentSets_GetNextAgent (AgentTypes, &state)) > 0)
    {
      if ((info = Dictionary_Find (AgentTypeInfos, typeId)) != NULL)
	Free (info);
    }
  Dictionary_Destroy (AgentTypeInfos);

  AgentSets_DestroyAgentSet (AgentTypes);
  Agent_DestroyAgent (DefaultAgent);
}


bool 
AgentSet_SetParameterValue (const string Parameter, const string Value)
{
  /* initialize module if necessary */
  AgentSet_Initialize ();

  if (strcmp (Parameter, "NumberOfAgents") == 0)
    {
      NumberOfAgents = atoi (Value);
      return TRUE;
    }
  else
    return AgentSet_SetParameterValueForAgent (0, Parameter, Value,
					       AGENT_TYPE);
}

bool 
AgentSet_SetParameterValueForAgent (AgentIdType AgentId,
				 const string Parameter, const string Value,
				    int KindOfAgent)
{
  AgentType *agent;

  if (KindOfAgent == AGENT_TYPE)
    {
      if (AgentId != 0)
	agent = AgentSets_GetAgent (AgentTypes, AgentId);
      else
	agent = DefaultAgent;
    }
  else
    agent = AgentSets_GetAgent (Agents, AgentId);

  if (agent)
    {
      DebugInfo ("--------- AGENT %p ID %d SETTING %s %s\n",
		 agent, AgentId, Parameter, Value);

      return Agent_SetParameterValueForAgent (agent, Parameter, Value);
    }
  return FALSE;
}


static void 
FillBufferWithConstraints (DynaStringType * BufferPtr, char *Name,
			   Constraint * Table)
{
  char constraints[514] =
  {'\0'}, *cs = constraints;
  int *counter;
  int i, m;

  if (Table)
    {
      counter = alloca ((1 + INVALID) * sizeof (int));

      for (i = 0; i <= INVALID; i++)
	counter[i] = 0;

      for (i = 0; i < 256; i++)
	{
	  if (Table[i] < INVALID)
	    counter[Table[i]]++;
	  else
	    counter[INVALID]++;
	}

      if (counter[INVALID])
	Warning ("AgentSet: FillBufferWithConstraints",
		 "%d INVALID constraint values\n", counter[INVALID]);

      for (i = 0, m = 0; i < INVALID; i++)
	if (counter[i] > counter[m])
	  m = i;

      cs[0] = (m == 0 ? '-' : (m == 1 ? '+' : '&'));
      cs[1] = '*';
      cs[2] = '\0';
      cs += 2;

      for (i = 0; i < 256; i++)
	{
	  if (Table[i] != m)
	    {
	      cs[0] = (Table[i] == 0 ? '-' : (Table[i] == 1 ? '+' : '&'));
	      cs[1] = i;
	      cs[2] = '\0';
	      cs += 2;
	    }
	}

      DynaString_Append (BufferPtr, Name);
      DynaString_Append (BufferPtr, constraints);
      DynaString_Append (BufferPtr, "\n");
    }
}

static void 
FillBufferWithAttributesOfAgent (DynaStringType * BufferPtr,
				 AgentIdType AgentId)
{
  char agent_type;
  AgentType *agent;

  agent = AgentSets_GetAgent (Agents, AgentId);

/*      FillBufferWithConstraints(DumpBuffer, "\tFRIEND-CONSTRAINTS ", 
   agent->friendConstraints);

   FillBufferWithConstraints(DumpBuffer, "\tMATE-CONSTRAINTS ", 
   agent->mateConstraints);

   FillBufferWithConstraints(DumpBuffer, "\tATTACK_CONSTRAINTS ", 
   agent->attackConstraints);
 */

  DynaString_Append (BufferPtr, "\tATTRIBUTES\n");

  DynaString_AppendFormat (BufferPtr, 64, "\t\tMaxAge = \t%d\n",
			   AgentSet_GetMaxAge (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tMaxEnergy = \t%d\n",
			   AgentSet_GetMaxEnergy (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tMaxFriends = \t%d\n",
			   AgentSet_GetMaxFriends (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tRangeOfVision = \t%d\n",
			   AgentSet_GetVisionRange (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tRangeOfHearing = \t%d\n",
			   AgentSet_GetHearingRange (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tStrength = \t%d\n",
			   AgentSet_GetStrength (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tMatePause = \t%d\n",
			   AgentSet_GetMatePause (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tMinMatingAge = \t%d\n",
			   AgentSet_GetMinMatingAge (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tMinEnergyFactor = \t%f\n",
			   AgentSet_GetMinEnergyFactor (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tHungerFactor = \t%f\n",
			   AgentSet_GetHungerFactor (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tAttackFactor = \t%f\n",
			   AgentSet_GetAttackFactor (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tDefenseFactor = \t%f\n",
			   AgentSet_GetDefenseFactor (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tCostOfLiving = \t%d\n",
			   AgentSet_GetCostOfLiving (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tNrOfChildren = \t%d\n",
			   AgentSet_GetNumberOfChildren (AgentId));

  DynaString_Append (BufferPtr, "\tEND_ATTRIBUTES\n");
}

static void 
FillBufferWithStatusOfAgent (DynaStringType * BufferPtr,
			     AgentIdType AgentId)
{
  PositionType position;
  IterStateType friendIter;
  AgentIdType friend;
  AgentType *agent;

  DynaString_Append (BufferPtr, "\tSTATUS\n");

  DynaString_AppendFormat (BufferPtr, 64, "\t\tEnergy = \t%d\n",
			   AgentSet_GetEnergy (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tAge = \t\t%d\n",
			   AgentSet_GetAge (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tAttacks = \t%d\n",
			   AgentSet_GetAttacks (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tIdle = \t%d\n",
			   AgentSet_GetIdle (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tMemory = \t%d\n",
			   AgentSet_GetMemory (AgentId));

  position = AgentSet_GetPosition (AgentId);
  DynaString_AppendFormat (BufferPtr, 64, "\t\tXPosition = \t%d\n", position.x);
  DynaString_AppendFormat (BufferPtr, 64, "\t\tYPosition = \t%d\n", position.y);

  agent = AgentSets_GetAgent (Agents, AgentId);
  friendIter = Agent_InitFriendIteration (agent);
  while ((friend = Agent_GetNextFriend (agent, &friendIter)) != 0)
    DynaString_AppendFormat (BufferPtr, 64, "\t\tFriends = \t%d\n", friend);

  DynaString_AppendFormat (BufferPtr, 64, "\t\tOrientation = \t%d\n",
			   AgentSet_GetOrientation (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tScreams = \t%d\n",
			   AgentSet_GetScreams (AgentId));
  DynaString_AppendFormat (BufferPtr, 64, "\t\tLastMate = \t%d\n",
			   AgentSet_GetLastMate (AgentId));

  DynaString_Append (BufferPtr, "\tEND_STATUS\n");
}


void 
AgentSet_DumpSection ()
{
  AgentIdType agent;
  IterStateType iter;
  char agent_type;
  AgentType *agentp;

  Dump_DumpAgent ("AGENTS\n\n");

  iter = AgentSet_InitAgentIteration ();
  while ((agent = AgentSet_GetNextAgent (&iter)) != 0)
    {
      DebugLevel (DL, "Dumping Agent %d\n", agent);

      DynaString_Reset (DumpBuffer);
      DynaString_AppendFormat (DumpBuffer, 64, "AGENT %d\n", agent);

      agentp = AgentSets_GetAgent (Agents, agent);

      if ((agent_type = Agent_GetType (agentp)) != 0)
	DynaString_AppendFormat (DumpBuffer, 64, "\tTYPE %c\n", agent_type);
      Dump_DumpAgent (DynaString_CString (DumpBuffer));

      if (Dump_WantDumpAgentAttributes ())
	{
	  DynaString_Reset (DumpBuffer);
	  FillBufferWithAttributesOfAgent (DumpBuffer, agent);
	  Dump_DumpAgentAttributes (DynaString_CString (DumpBuffer));
	}

      if (Dump_WantDumpAgentStatus ())
	{
	  DynaString_Reset (DumpBuffer);
	  FillBufferWithStatusOfAgent (DumpBuffer, agent);
	  Dump_DumpAgentStatus (DynaString_CString (DumpBuffer));
	}

      if (Dump_WantDumpAgentProgram ())
	{
	  ProgramType *program;
	  string program_string = NULL;

	  program = AgentSet_GetProgram (agent);
	  program_string = Interpreter_UnparseProgram (program);
	  if (program_string)
	    {
	      Dump_DumpAgentProgram (program_string);
	      Free (program_string);
	    }
	}

      Dump_DumpAgent ("END_AGENT\n\n");
    }

  Dump_DumpAgent ("END_AGENTS\n\n");
}


/*-ftn-********************************************************************
 *
 * Funktion  : AgentSet_InitTimestep
 *
 * Beschreib.:
 *
 * Parameter : (i)Timestep : CountType
 *
 * Return    : * keiner *
 *
 *************************************************************************/

void 
AgentSet_InitTimestep (CountType Timestep)
{
  IterStateType state;
  AgentIdType agent;
  unsigned dead = 0, alive = 0;

  state = AgentSet_InitAgentIteration ();
  agent = AgentSet_GetNextAgent (&state);
  while (agent && Timestep == 1)
    if (Agent_Timestep (AgentSets_GetAgent (Agents, agent)) != AGENT_ALIVE)
      {
	/* Todesflle von vor dem ersten Timestep bearbeiten */
	AgentIdType next_agent;
	dead++;
	World_AgentFoodConvert (agent, AgentSet_GetEnergy (agent), AgentSet_GetPosition (agent));
	next_agent = AgentSet_GetNextAgent (&state);
	AgentSet_DestroyAgent (agent);
	agent = next_agent;
      }
    else
      {
	alive++;
	agent = AgentSet_GetNextAgent (&state);
      }

  if (Dump_WantDumpStatistics ())
    {
      DynaString_Reset (DumpBuffer);

      DynaString_AppendFormat (DumpBuffer, 64, "#! Agents = %d\n",
			       AgentSet_GetAgentCount ());

      DynaString_AppendFormat (DumpBuffer, 64, "#! Type 0 = %d\n",
			       AgentSet_GetAgentCountOfType (0));
      state = AgentSets_InitAgentIteration (AgentTypes);
      while ((agent = AgentSets_GetNextAgent (AgentTypes, &state)))
	{
	  DynaString_AppendFormat (DumpBuffer, 64, "#! Type %c = %d\n",
		 agent ? agent : '*', AgentSet_GetAgentCountOfType (agent));
	}

      Dump_DumpStatistics (DynaString_CString (DumpBuffer));
    }

  DebugLevel (DL, "Timestep %d: %d Agents died, %d still alive\n",
	      Timestep, dead, alive);
}

void 
AgentSet_ConcludeTimestep (CountType Timestep)
{
  IterStateType state;
  AgentIdType agent;
  AgentType *PtrAgent;

  /* Das ganze Procedere nochmal, damit im Dump eines Timesteps die Todesflle 
     angezeigt werden, die in diesem Timestep passiert sind */

  state = AgentSet_InitAgentIteration ();
  agent = AgentSet_GetNextAgent (&state);
  while (agent)
    {

      PtrAgent = AgentSets_GetAgent (Agents, agent);
      Agent_ConsumeEnergy (PtrAgent, Agent_GetCostOfLiving (PtrAgent));
      Agent_IncreaseAge (PtrAgent, 1);
      if (Agent_Timestep (PtrAgent) != AGENT_ALIVE)
	{
	  AgentIdType next_agent;
	  World_AgentFoodConvert (agent, AgentSet_GetEnergy (agent), AgentSet_GetPosition (agent));
	  next_agent = AgentSet_GetNextAgent (&state);
	  AgentSet_DestroyAgent (agent);
	  agent = next_agent;
	}
      else
	agent = AgentSet_GetNextAgent (&state);
    }

  if (Dump_WantDumpAgentsSection ())
    AgentSet_DumpSection ();
}


/*-ftn-********************************************************************
 *
 * Funktion  : AgentSet_InitAgentIteration
 *
 * Beschreib.:
 *
 * Parameter :
 *
 * Return    : AgentIdType
 *
 *************************************************************************/
inline IterStateType 
AgentSet_InitAgentIteration ()
{
  return AgentSets_InitAgentIteration (Agents);
}

/*-ftn-********************************************************************
 *
 * Funktion  : AgentSet_GetNextAgent
 *
 * Beschreib.: 
 *
 * Parameter : (i)PtrState : IterStateType
 *
 * Return    : AgentIdType
 *
 *************************************************************************/
inline AgentIdType 
AgentSet_GetNextAgent (IterStateType * PtrState)
{
  return AgentSets_GetNextAgent (Agents, PtrState);
}

inline CountType 
AgentSet_GetAgentCount ()
{
  return AgentSets_GetCount (Agents);
}


void 
AgentSet_FinishAgentType (AgentIdType AgentId)
{
  AgentType *agent;

  if (AgentId == 0)
    agent = DefaultAgent;
  else
    agent = AgentSets_GetAgent (AgentTypes, AgentId);

  if (agent)
    {
      EnergyType energy;
      CountType age, range, friends;
      StrengthType strength;

      if (Agent_GetMinEnergyFactor (agent) == -1.0)
	{
	  Agent_SetMinEnergyFactor (agent, (FactorType) ((Random_Rand () % 100) + 1));
	  DebugInfo ("type %d: using MinEnergyFactor = %f\n", AgentId,
		     Agent_GetMinEnergyFactor (agent));
	}

      if (Agent_GetHungerFactor (agent) == -1.0)
	{
	  Agent_SetHungerFactor (agent, (FactorType) (Random_Rand () % 101));
	  DebugInfo ("type %d: using HungerFactor = %f\n", AgentId,
		     Agent_GetHungerFactor (agent));
	}

      if (Agent_GetAttackFactor (agent) == -1.0)
	{
	  Agent_SetAttackFactor (agent, (FactorType) (Random_Rand () % 101));
	  DebugInfo ("type %d: using AttackFactor = %f\n", AgentId,
		     Agent_GetAttackFactor (agent));
	}

      if (Agent_GetDefenseFactor (agent) == -1.0)
	{
	  Agent_SetDefenseFactor (agent, (FactorType) (Random_Rand () % 101));
	  DebugInfo ("type %d: using DefenseFactor = %f\n", AgentId,
		     Agent_GetDefenseFactor (agent));
	}

      if (Agent_GetNumberOfChildren (agent) == -1)
	{
	  Agent_SetNumberOfChildren (agent, (Random_Rand () % 5) + 1);
	  DebugInfo ("type %d: using NumberOfChildren = %d\n", AgentId,
		     Agent_GetNumberOfChildren (agent));
	}

      if ((age = Agent_GetMaxAge (agent)) < 0)
	{
	  Agent_SetMaxAge (agent, age = ((Random_Rand () % -age) + 1));
	  DebugInfo ("type %d: using MaxAge = %d\n", AgentId, age);
	}
      else if (age == 0)
	Warning ("AgentSet_FinishAgentType", "type %d: MaxAge = 0",
		 AgentId);
      if ((energy = Agent_GetMaxEnergy (agent)) < 0)
	{
	  Agent_SetMaxEnergy (agent, energy = ((Random_Rand () % -energy) + 1));
	  DebugInfo ("type %d: using MaxEnergy = %d\n", AgentId, energy);
	}
      else if (energy == 0)
	Warning ("AgentSet_FinishAgentType", "type %d: MaxEnergy = 0",
		 AgentId);
      if ((strength = Agent_GetStrength (agent)) < 0)
	{
	  Agent_SetStrength (agent, strength = ((Random_Rand () % -strength) + 1));
	  DebugInfo ("type %d: using Strength = %d\n", AgentId, strength);
	}

      if ((range = Agent_GetRangeOfVision (agent)) < 0)
	{
	  Agent_SetRangeOfVision (agent, range = ((Random_Rand () % -range) + 1));
	  DebugInfo ("type %d: using RangeOfVision = %d\n", AgentId, range);
	}

      if ((range = Agent_GetRangeOfHearing (agent)) < 0)
	{
	  Agent_SetRangeOfHearing (agent, range = ((Random_Rand () % -range) + 1));
	  DebugInfo ("type %d: using RangeOfHearing = %d\n", AgentId, range);
	}

      if ((friends = Agent_GetMaxFriends (agent)) < 0)
	{
	  Agent_SetMaxFriends (agent, energy = ((Random_Rand () % -friends) + 1));
	  DebugInfo ("type %d: using MaxFriends = %d\n", AgentId, friends);
	}
    }
}
void 
AgentSet_FinishAgent (AgentIdType AgentId)
{
  AgentType *agent;

  if ((agent = AgentSets_GetAgent (Agents, AgentId)) != NULL)
    {
      EnergyType energy;
      CountType age;
      if ((energy = Agent_GetEnergy (agent)) <= 0)
	{
	  if (energy < 0)
	    energy = (Random_Rand () % AgentSet_GetMaxEnergy (AgentId)) + 1;
	  else if (energy == 0)
	    energy = Agent_GetMaxEnergy (agent);

	  Agent_SetEnergy (agent, energy);
	  DebugInfo ("agent %d: using Energy = %d\n", AgentId, energy);
	}

      if ((age = Agent_GetAge (agent)) < 0)
	{
	  if (age < 0)
	    age = Random_Rand () % AgentSet_GetMaxAge (AgentId);

	  Agent_SetAge (agent, age);
	  DebugInfo ("agent %d: using Age = %d\n", AgentId, age);
	}

      if (Agent_GetOrientation (agent) < 0)
	{
	  Agent_SetOrientation (agent, (Random_Rand () % 8) * 45);
	  DebugInfo ("agent %d: using Orientation = %d\n", AgentId,
		     Agent_GetOrientation (agent));
	}
    }
}


static AgentType *
CreateAgentOfType (AgentIdType TypeId)
{
  AgentType *agent, *type;

  agent = Agent_CreateAgent ();
  if (TypeId != 0)
    {
      if (!(type = AgentSets_GetAgent (AgentTypes, TypeId)))
	{
	  Warning ("CreateAgentOfType",
		   "agent type %d does not exist. using default.\n",
		   TypeId);
	  TypeId = 0;
	  type = DefaultAgent;
	}
    }
  else
    type = DefaultAgent;

  AgentSet_FinishAgentType (TypeId);
  Agent_CopyAgent (agent, type);
  Agent_SetType (agent, TypeId);
  return agent;
}

AgentIdType 
AgentSet_CreateAgentTypeOfType (AgentIdType NewTypeId,
				AgentIdType TypeId)
{
  AgentType *agent;
  AgentTypeInfo *info;

  agent = CreateAgentOfType (TypeId);
  if ((NewTypeId = AgentSets_InsertAgent (AgentTypes, NewTypeId, agent)) != 0)
    {
      Agent_SetId (agent, NewTypeId);

      info = Allocate (sizeof (AgentTypeInfo));
      info->wantedNumber = 0;
      info->currentNumber = 0;
      Dictionary_InsertItem (AgentTypeInfos, info, NewTypeId);

      DebugLevel (DL, "------- AgentSet_CreateAgentTypeOfType %d\n", NewTypeId);
      AgentSet_FinishAgentType (NewTypeId);
    }
  else
    Agent_DestroyAgent (agent);

  return NewTypeId;
}

AgentIdType 
AgentSet_CreateAgentOfType (AgentIdType AgentId, AgentIdType TypeId)
{
  AgentType *agent;
  AgentTypeInfo *info;

  AgentSet_FinishAgentType (0);

  agent = CreateAgentOfType (TypeId);
  AgentId = AgentSets_InsertAgent (Agents, AgentId, agent);
  if (AgentId > 0)
    {
      Agent_SetId (agent, AgentId);

      if ((info = GetAgentTypeInfo (TypeId, TRUE)) != NULL)
	info->currentNumber++;

      DebugLevel (DL, "------- AgentSet_CreateAgentOfType %d\n", AgentId);
      AgentSet_FinishAgent (AgentId);
    }
  else
    Agent_DestroyAgent (agent);

  return AgentId;
}

inline AgentIdType 
AgentSet_CreateAgent (AgentIdType AgentId)
{
  return AgentSet_CreateAgentOfType (AgentId, 0);
}

inline void 
AgentSet_DestroyAgentType (AgentIdType TypeId)
{
  Agent_DestroyAgent (AgentSets_RemoveAgent (AgentTypes, TypeId));
}

inline void 
AgentSet_DestroyAgent (AgentIdType AgentId)
{
  AgentIdType typeId;
  AgentTypeInfo *typeInfo;

  typeId = AgentSet_GetType (AgentId);

  typeInfo = GetAgentTypeInfo (typeId, FALSE);
  typeInfo->currentNumber--;
  Agent_DestroyAgent (AgentSets_RemoveAgent (Agents, AgentId));
}


inline bool 
AgentSet_IsAgentType (AgentIdType TypeId)
{
  if (TypeId != 0)
    return AgentSets_GetAgent (AgentTypes, TypeId) != NULL;
  else
    return DefaultAgent != NULL;
}

inline bool 
AgentSet_IsAgent (AgentIdType AgentId)
{
  return AgentSets_GetAgent (Agents, AgentId) != NULL;
}


static AgentTypeInfo *
GetAgentTypeInfo (AgentIdType TypeId, bool Create)
{
  AgentTypeInfo *info;

  if (TypeId != 0)
    {
      if (!(info = Dictionary_Find (AgentTypeInfos, TypeId)))
	{
	  if (Create)
	    {
	      info = Allocate (sizeof (AgentTypeInfo));

	      info->wantedNumber = info->currentNumber = 0;

	      Dictionary_InsertItem (AgentTypeInfos, info, TypeId);
	    }
	}

      return info;
    }
  else
    return &DefaultTypeInfo;
}

void 
AgentSet_SetNumberOfAgentsOfType (AgentIdType TypeId, CountType Number)
{
  AgentTypeInfo *info;

  if (TypeId != 0)
    {
      info = GetAgentTypeInfo (TypeId, TRUE);
      info->wantedNumber = Number;
    }
  else
    DefaultTypeInfo.wantedNumber = Number;
}

CountType 
AgentSet_GetNumberOfAgentsOfType (AgentIdType TypeId)
{
  AgentTypeInfo *info;

  if (TypeId != 0)
    {
      if ((info = GetAgentTypeInfo (TypeId, FALSE)))
	return info->wantedNumber;
      else
	return 0;
    }
  else
    return DefaultTypeInfo.wantedNumber;
}

CountType 
AgentSet_GetAgentCountOfType (AgentIdType TypeId)
{
  AgentTypeInfo *info;

  if (TypeId != 0)
    {
      if ((info = GetAgentTypeInfo (TypeId, FALSE)))
	return info->currentNumber;
      else
	return 0;
    }
  else
    return DefaultTypeInfo.currentNumber;
}


void 
AgentSet_SetFriendConstraint (AgentIdType TypeId,
			    AgentIdType FriendTypeId, Constraint constraint)
{
  AgentType *type;

  if (TypeId)
    type = AgentSets_GetAgent (AgentTypes, TypeId);
  else
    type = DefaultAgent;

  Agent_SetFriendConstraint (type, FriendTypeId, constraint);
}

Constraint 
AgentSet_GetFriendConstraint (AgentIdType TypeId,
			      AgentIdType FriendTypeId)
{
  AgentType *type;

  if (TypeId)
    type = AgentSets_GetAgent (AgentTypes, TypeId);
  else
    type = DefaultAgent;

  return Agent_GetFriendConstraint (type, FriendTypeId);
}


void 
AgentSet_SetMateConstraint (AgentIdType TypeId,
			    AgentIdType MateTypeId, Constraint constraint)
{
  AgentType *type;

  if (TypeId)
    type = AgentSets_GetAgent (AgentTypes, TypeId);
  else
    type = DefaultAgent;

  Agent_SetMateConstraint (type, MateTypeId, constraint);
}

Constraint 
AgentSet_GetMateConstraint (AgentIdType TypeId,
			    AgentIdType MateTypeId)
{
  AgentType *type;

  if (TypeId)
    type = AgentSets_GetAgent (AgentTypes, TypeId);
  else
    type = DefaultAgent;

  return Agent_GetMateConstraint (type, MateTypeId);
}


void 
AgentSet_SetAttackConstraint (AgentIdType TypeId,
			    AgentIdType AttackTypeId, Constraint constraint)
{
  AgentType *type;

  if (TypeId)
    type = AgentSets_GetAgent (AgentTypes, TypeId);
  else
    type = DefaultAgent;

  Agent_SetAttackConstraint (type, AttackTypeId, constraint);
}

Constraint 
AgentSet_GetAttackConstraint (AgentIdType TypeId,
			      AgentIdType AttackTypeId)
{
  AgentType *type;

  if (TypeId)
    type = AgentSets_GetAgent (AgentTypes, TypeId);
  else
    type = DefaultAgent;

  return Agent_GetAttackConstraint (type, AttackTypeId);
}

void 
AgentSet_SetEatConstraint (AgentIdType TypeId,
			   AgentIdType Food, Constraint constraint)
{
  AgentType *type;
  if (TypeId)
    type = AgentSets_GetAgent (AgentTypes, TypeId);
  else
    type = DefaultAgent;

  Agent_SetEatConstraint (type, Food, constraint);
}

Constraint 
AgentSet_GetEatConstraint (AgentIdType TypeId, AgentIdType Food)
{
  AgentType *type;

  if (TypeId)
    type = AgentSets_GetAgent (AgentTypes, TypeId);
  else
    type = DefaultAgent;

  return Agent_GetEatConstraint (type, Food);
}

inline bool 
AgentSet_CanAttack (AgentIdType AttackerId, AgentIdType DefenderId)
{
  return Agent_CanAttack (AgentSets_GetAgent (Agents, AttackerId),
			  AgentSet_GetType (DefenderId));
}

inline bool 
AgentSet_CanMate (AgentIdType AgentId, AgentIdType MateId)
{
  return Agent_CanMate (AgentSets_GetAgent (Agents, AgentId),
			AgentSet_GetType (MateId));
}

inline bool 
AgentSet_CanBeFriend (AgentIdType AgentId, AgentIdType FriendId)
{
  return Agent_CanMate (AgentSets_GetAgent (Agents, AgentId),
			AgentSet_GetType (FriendId));
}

inline bool 
AgentSet_CanEat (AgentIdType AgentId, AgentIdType Food)
{
  return Agent_CanEat (AgentSets_GetAgent (Agents, AgentId), Food);
}
