#ifndef AGENT_SET_H
#define AGENT_SET_H

/*-module-*****************************************************************
 *
 * Modul     : AgentSet (Funktionsmodul-Definitionen)
 *
 * Beschreib.: Diese Header-Datei dient zur Bereitstellung der
 *             Exportschnittstelle der gleichnamigen C-Datei.
 *             Sie muss in den Modulen includet werden, in denen
 *             die beschriebenen Ressourcen importiert werden sollen.
 *
 * Autor     : <autor>
 *
 * Datum     : <datum>
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/


/*-includes-*/

#include "../misc/TypeDefs.h"
#include "Agent.h"
#include "AgentList.h"

/*-defines-*/

/*-import-*/

inline bool AgentSet_SetParameterValue (const string Parameter,
					const string Value);
inline bool AgentSet_SetParameterValueForAgent (AgentIdType AgentId,
				 const string Parameter, const string Value,
						const int KindOfAgent);

inline IterStateType AgentSet_InitAgentIteration ();
inline AgentIdType AgentSet_GetNextAgent (IterStateType * PtrState);
inline CountType AgentSet_GetAgentCount ();


void AgentSet_Initialize ();
void AgentSet_EndSimulation ();

void AgentSet_InitTimestep (CountType Timestep);
void AgentSet_ConcludeTimestep (CountType Timestep);

bool AgentSet_ParseSection ();
void AgentSet_DumpSection ();

AgentIdType AgentSet_CreateAgentTypeOfType (AgentIdType NewTypeId,
					    AgentIdType TypeId);
AgentIdType AgentSet_CreateAgentOfType (AgentIdType AgentId,
					AgentIdType TypeId);
inline AgentIdType AgentSet_CreateAgent (AgentIdType AgentId);

inline void AgentSet_DestroyAgentType (AgentIdType TypeId);
inline void AgentSet_DestroyAgent (AgentIdType AgentId);

inline bool AgentSet_IsAgentType (AgentIdType TypeId);
inline bool AgentSet_IsAgent (AgentIdType AgentId);

inline AgentIdType AgentSet_GetType (AgentIdType AgentId);
inline void AgentSet_SetType (AgentIdType AgentId, AgentIdType TypeId);

inline EnergyType AgentSet_GetEnergy (AgentIdType AgentId);
inline void AgentSet_SetEnergy (AgentIdType AgentId, EnergyType Energy);

inline CountType AgentSet_GetAge (AgentIdType AgentId);
inline void AgentSet_SetAge (AgentIdType AgentId, CountType Age);

inline MemoryType AgentSet_GetMemory (AgentIdType AgentId);
inline void AgentSet_SetMemory (AgentIdType AgentId, MemoryType Memory);

inline PositionType AgentSet_GetPosition (AgentIdType AgentId);
inline void AgentSet_SetPosition (AgentIdType AgentId,
				  PositionType Position);

inline OrientationType AgentSet_GetOrientation (AgentIdType AgentId);
inline void AgentSet_ChangeOrientation (AgentIdType AgentId,
					OrientationType Orientation);
inline void AgentSet_SetOrientation (AgentIdType AgentId,
				     OrientationType Orientation);

inline CountType AgentSet_GetAttacks (AgentIdType AgentId);
inline void AgentSet_SetAttacks (AgentIdType AgentId, CountType Attacks);

inline CountType AgentSet_GetIdle (AgentIdType AgentId);
inline void AgentSet_SetIdle (AgentIdType AgentId,
			      CountType Idle);

inline CountType AgentSet_GetScreams (AgentIdType AgentId);
inline void AgentSet_SetScreams (AgentIdType AgentId,
				 CountType Screams);

inline IterStateType AgentSet_InitFriendIteration (AgentIdType AgentId);
inline AgentIdType AgentSet_GetNextFriend (AgentIdType AgentId,
					   IterStateType * PtrState);
inline bool AgentSet_AddToFriends (AgentIdType AgentId,
				   AgentIdType FriendId);
inline bool AgentSet_MakeFriend (AgentIdType AgentId,
				 AgentIdType FriendId);
inline bool AgentSet_IsFriendOf (AgentIdType AgentId,
				 AgentIdType FriendId);
inline bool AgentSet_MeetFriend (AgentIdType AgentId,
				 AgentIdType FriendId);

inline EnergyType AgentSet_GetMaxEnergy (AgentIdType AgentId);
inline void AgentSet_SetMaxEnergy (AgentIdType AgentId,
				   EnergyType MaxEnergy);

inline CountType AgentSet_GetMaxFriends (AgentIdType AgentId);
inline void AgentSet_SetMaxFriends (AgentIdType AgentId,
				    CountType MaxFriends);

inline CountType AgentSet_GetVisionRange (AgentIdType AgentId);
inline void AgentSet_SetVisionRange (AgentIdType AgentId,
				     CountType VisionRange);

inline CountType AgentSet_GetHearingRange (AgentIdType AgentId);
inline void AgentSet_SetHearingRange (AgentIdType AgentId,
				      CountType HearingRange);

inline StrengthType AgentSet_GetStrength (AgentIdType AgentId);
inline void AgentSet_SetStrength (AgentIdType AgentId,
				  CountType Strength);

inline CountType AgentSet_GetNumberOfChildren (AgentIdType AgentId);
inline void AgentSet_SetNumberOfChildren (AgentIdType AgentId,
					  CountType NumberOfChildren);

inline CountType AgentSet_GetMaxAge (AgentIdType AgentId);
inline void AgentSet_SetMaxAge (AgentIdType AgentId, CountType MaxAge);

inline void AgentSet_SetMinMatingAge (AgentIdType AgentId,
				      CountType MinMatingAge);
inline CountType AgentSet_GetMinMatingAge (AgentIdType AgentId);

/* added by Martin */
inline CountType AgentSet_GetLastMate (AgentIdType AgentId);
inline void AgentSet_SetLastMate (AgentIdType AgentId,
				  CountType LastMate);

inline CountType AgentSet_GetMatePause (AgentIdType AgentId);
inline void AgentSet_SetMatePause (AgentIdType AgentId,
				   CountType MatePause);
inline FactorType AgentSet_GetMinEnergyFactor (AgentIdType AgentId);
inline void AgentSet_SetMinEnergyFactor (AgentIdType AgentId, FactorType MinEnergyFactor);

inline FactorType AgentSet_GetHungerFactor (AgentIdType AgentId);
inline void AgentSet_SetHungerFactor (AgentIdType AgentId, FactorType HungerFactor);

inline FactorType AgentSet_GetAttackFactor (AgentIdType AgentId);
inline void AgentSet_SetAttackFactor (AgentIdType AgentId, FactorType AttackFactor);

inline FactorType AgentSet_GetDefenseFactor (AgentIdType AgentId);
inline void AgentSet_SetDefenseFactor (AgentIdType AgentId, FactorType DefenseFactor);

inline EnergyType AgentSet_GetCostOfLiving (AgentIdType AgentId);
inline void AgentSet_SetCostOfLiving (AgentIdType AgentId, EnergyType CostOfLiving);


inline ProgramType *AgentSet_GetProgram (AgentIdType AgentId);
void AgentSet_SetProgramForKind (AgentIdType AgentId,
				 ProgramType * Program, int KindOfAgent);
inline void AgentSet_SetProgram (AgentIdType AgentId,
				 ProgramType * Program);
inline CountType AgentSet_GetProgramLength (AgentIdType AgentId);
inline void AgentSet_SetProgramLength (AgentIdType AgentId,
				       CountType Length);

typedef struct
  {
    unsigned int wantedNumber;
    unsigned int currentNumber;
  }
AgentTypeInfo;


void AgentSet_FinishInitialization ();

void AgentSet_SetNumberOfAgentsOfType (AgentIdType TypeId,
				       CountType Number);
CountType AgentSet_GetNumberOfAgentsOfType (AgentIdType TypeId);
CountType AgentSet_GetAgentCountOfType (AgentIdType TypeId);

Constraint AgentSet_GetFriendConstraint (AgentIdType TypeId,
					 AgentIdType FriendTypeId);
void AgentSet_SetFriendConstraint (AgentIdType TypeId,
			   AgentIdType FriendTypeId, Constraint constraint);

Constraint AgentSet_GetMateConstraint (AgentIdType TypeId,
				       AgentIdType MateTypeId);
void AgentSet_SetMateConstraint (AgentIdType TypeId,
			     AgentIdType MateTypeId, Constraint constraint);

Constraint AgentSet_GetAttackConstraint (AgentIdType TypeId,
					 AgentIdType AttackTypeId);
void AgentSet_SetAttackConstraint (AgentIdType TypeId,
			   AgentIdType AttackTypeId, Constraint constraint);
void AgentSet_SetEatConstraint (AgentIdType TypeId,
				AgentIdType Food, Constraint constraint);
Constraint AgentSet_GetEatConstraint (AgentIdType TypeId, AgentIdType Food);



inline bool AgentSet_CanAttack (AgentIdType AttackerId,
				AgentIdType DefenderId);
inline bool AgentSet_CanMate (AgentIdType AgentId, AgentIdType MateId);
inline bool AgentSet_CanBeFriend (AgentIdType AgentId,
				  AgentIdType FriendId);
inline bool AgentSet_CanEat (AgentIdType AgentId, AgentIdType Food);

#endif
