







#include "../dictionary/Dictionary.h"
#include <stdlib.h>

/* dummy for testing */
/*typedef void  DictType; */

typedef struct
  {
    AgentIdType currentId;
    DictType *map;
    AgentIdListType list;
  }
AgentSet;


static AgentType *Default_Agent;

static AgentSet *Default_Agents;
static AgentSet *Agents;

/*-local-functions-*/

static AgentType *
GetAgentWithIdFromSet (AgentSet * PtrSet, AgentIdType AgentId)
{
  AgentIdNode *nodePtr;

  if (PtrSet)
    {
      if (AgentId > 0)
	{
	  /* Test (ohne Dictionary) */

/*        nodePtr = AgentList_FindAgentId(&(PtrSet->list), AgentId);

   if (nodePtr != NULL) 
   return (AgentType*)nodePtr->PtrAgent;
   else                 
   return NULL; */

	  return Dictionary_Find (PtrSet->map, (DictKeyType) AgentId);
	}
    }
  else if (AgentId == 0)
    return Default_Agent;

  return NULL;
}

static AgentType *
GetAgentWithId (AgentIdType AgentId)
{
  return GetAgentWithIdFromSet (Agents, AgentId);
}


static bool 
AddAgentWithIdToSet (AgentSet * PtrSet, AgentType * PtrAgent, AgentIdType AgentId)
{
  if (Dictionary_InsertItem (&PtrSet->map, PtrAgent, (DictKeyType) AgentId))
    {
      AgentType *dummy;

      AgentList_Append (&PtrSet->list, AgentId);
      printf ("Agent %d %p inserted\n", AgentId, PtrAgent);

      if ((dummy = Dictionary_Find (PtrSet->map, (DictKeyType) AgentId)) != PtrAgent)
	fprintf (stderr, "Dictionary Error: %d: %p != %p\n", AgentId, PtrAgent, dummy);

      return 1;
    }
  else
    {
      return 0;
    }

  /* Zu Testzwecken wird der AgentTypePtr in die AgentList mit eingehaengt ... */

/*        AgentList_AppendTest (&(PtrSet->list)
   , AgentId, (void*)PtrAgent); */

  return 1;
}

static bool 
AddAgentWithId (AgentType * PtrAgent, AgentIdType AgentId)
{
  return AddAgentWithIdToSet (Agents, PtrAgent, AgentId);
}

static AgentIdType 
GetIdInSet (AgentSet * PtrSet)
{
  ++PtrSet->currentId;
  return PtrSet->currentId;
}

static AgentIdType 
GetId ()
{
  return GetIdInSet (Agents);
}

static bool 
UnusedIdInSet (AgentSet * PtrSet, AgentIdType AgentId)
{
  return GetAgentWithIdFromSet (PtrSet, AgentId) == NULL;
}

static bool 
UnusedId (AgentIdType AgentId)
{
  return UnusedIdInSet (Agents, AgentId);
}


static AgentType *
CreateAgentOfType (AgentIdType DefaultTypeId)
{
  AgentType *DefaultAgent;
  AgentType *PtrAgent;

  PtrAgent = Agent_CreateAgent ();
  if (DefaultTypeId != 0)
    {
      if ((DefaultAgent = GetAgentWithIdFromSet (Default_Agents, DefaultTypeId)) != NULL)
	{
	  Agent_CopyAgent (PtrAgent, DefaultAgent);
	  return PtrAgent;
	}
      else
	AS_WARNING ("Agent type does not exist. Using default");
    }

  Agent_CopyAgent (PtrAgent, Default_Agent);
  return PtrAgent;
}

static AgentType *
CreateAgent ()
{
  return CreateAgentOfType (0);
}


AgentIdType 
AgentSet_CreateAgentTypeOfType (AgentIdType AgentId,
				AgentIdType DefaultTypeId)
{
  AgentType *PtrAgent;

  if (AgentId == 0)
    AgentId = GetIdInSet (Default_Agents);

  if (AgentId != 0)
    {
      PtrAgent = CreateAgentOfType (DefaultTypeId);

      if (!(AddAgentWithIdToSet (Default_Agents, PtrAgent, AgentId)))
	{
	  AS_WARNING ("Cannot insert id in Dictionary! Serious...");
	  /*Agent_DestroyAgent(PtrAgent);
	     return 0; */
	}

      printf ("------- AgentSet_CreateAgentTypeOfType %d\n", AgentId);
      return AgentId;
    }
  else
    {
      AS_WARNING ("Couldn't create Agent 'cause there is no id left");
      return 0;
    }
}

AgentIdType 
AgentSet_CreateAgentOfType (AgentIdType AgentId,
			    AgentIdType DefaultTypeId)
{
  AgentType *PtrAgent;

  if (AgentId == 0)
    AgentId = GetId ();

  if (AgentId != 0)
    {
      PtrAgent = CreateAgentOfType (DefaultTypeId);

      if (!(AddAgentWithId (PtrAgent, AgentId)))
	{
	  AS_WARNING ("Cannot insert id in Dictionary! Serious...");
	  /*Agent_DestroyAgent(PtrAgent); */
	  /*return 0; */
	}

      printf ("------- AgentSet_CreateAgentOfType %d %d\n", AgentId, DefaultTypeId);
      return AgentId;
    }
  else
    {
      AS_WARNING ("Couldn't create Agent 'cause there is no id left");
      return 0;
    }
}

/*-ftn-********************************************************************
 *
 * Funktion  : AgentSet_CreateAgent
 *
 * Beschreib.: Falls noch kein Agent existiert wird die AgentId 1 zurueckgeliefert,
 *             ansonsten die naechste freie Id.
 *
 * Parameter :
 *
 * Return    : AgentIdType
 *
 *************************************************************************/
AgentIdType 
AgentSet_CreateAgent (AgentIdType AgentId)
/*--*/
{
  AgentId = AgentSet_CreateAgentOfType (AgentId, 0);

  printf ("------- CREATE AGENT ID  %d\n", AgentId);

  return AgentId;
}
/*-end-ftn-*/


AgentSet *
AgentSet_CreateAgentSet ()
{
  AgentSet *set;

  fprintf (stdout, "Create AgentSet\n");

  set = (AgentSet *) malloc (sizeof (AgentSet));
  AS_CHECKMEM (set);
  set->currentId = 0;
  set->map = Dictionary_Create ();
  set->list.PtrHead = NULL;
  return set;
}

void 
AgentSet_DestroyAgentFromSet (AgentSet * PtrSet, AgentIdType Agent)
{
  AgentType *agent_ptr = GetAgentWithIdFromSet (PtrSet, Agent);

  if (agent_ptr)
    {
      if (PtrSet)
	{
	  AgentList_Remove (&PtrSet->list, Agent);
	  Dictionary_Delete (&PtrSet->map, Agent);
	}
      Agent_DestroyAgent (agent_ptr);
    }
  else
    AS_WARNING ("Trying to destroy invalid agent");
}

void 
AgentSet_DestroyAgent (AgentIdType Agent)
{
  AgentSet_DestroyAgentFromSet (Agents, Agent);
}

void 
AgentSet_DestroyAgentSet (AgentSet * PtrSet)
{
}

void 
AgentSet_DestroyAgents ()
{
}

void 
AgentSet_Initialize ()
{
  fprintf (stdout, "Initialize AgentSet\n");

  if (!Default_Agent)
    Default_Agent = Agent_CreateAgent ();

  if (!Default_Agents)
    Default_Agents = AgentSet_CreateAgentSet ();
  if (!Agents)
    Agents = AgentSet_CreateAgentSet ();
}
