
#include "../dictionary/Dictionary.h"
#include <stdlib.h>

static int DL = 100;

/* dummy for testing */
/*typedef void  DictType; */

typedef struct
  {
    AgentIdType currentId;
    DictType *map;
    AgentIdListType list;
  }
AgentSet;

static int AgentSet_Initialized = 0;

static AgentType *Default_Agent;

static AgentSet *Default_Agents;
static AgentSet *Agents;

/*-local-functions-*/

static void 
UseIdInSet (AgentSet * PtrSet, AgentIdType AgentId)
{
  if (AgentId > PtrSet->currentId)
    PtrSet->currentId = AgentId;
}

static AgentType *
GetAgentWithIdFromSet (AgentSet * PtrSet, AgentIdType AgentId)
{
  AgentIdNode *nodePtr;

  if (PtrSet)
    {
      if (AgentId > 0)
	{
	  /* Test (ohne Dictionary) */

/*        nodePtr = AgentList_FindAgentId(&(PtrSet->list), AgentId);

   if (nodePtr != NULL) 
   return (AgentType*)nodePtr->PtrAgent;
   else                 
   return NULL; */

	  return Dictionary_Find (PtrSet->map, (DictKeyType) AgentId);
	}
    }
  else if (AgentId == 0)
    return Default_Agent;

  return NULL;
}

static AgentType *
GetAgentWithId (AgentIdType AgentId)
{
  return GetAgentWithIdFromSet (Agents, AgentId);
}

AgentIdType 
GetIdOfAgentInSet (AgentSet * PtrSet, AgentType * PtrAgent)
{
  IterStateType state;
  AgentIdType foundId;

  state = AgentList_InitIteration (&PtrSet->list);
  while (foundId = AgentList_Next (&PtrSet->list, &state))
    {
      if (PtrAgent == GetAgentWithIdFromSet (PtrSet, foundId))
	return foundId;
    }
  return 0;
}

AgentIdType 
GetIdOfAgent (AgentType * PtrAgent)
{
  return GetIdOfAgentInSet (Agents, PtrAgent);
}

static bool 
AddAgentWithIdToSet (AgentSet * PtrSet, AgentType * PtrAgent, AgentIdType AgentId)
{
  if (Dictionary_InsertItem (&PtrSet->map, PtrAgent, (DictKeyType) AgentId))
    {
      UseIdInSet (PtrSet, AgentId);
      AgentList_Append (&PtrSet->list, AgentId);
      return 1;
    }
  else
    {
      return 0;
    }

  /* Zu Testzwecken wird der AgentTypePtr in die AgentList mit eingehaengt ... */

/*        AgentList_AppendTest (&(PtrSet->list)
   , AgentId, (void*)PtrAgent); */

  return 1;
}

static bool 
AddAgentWithId (AgentType * PtrAgent, AgentIdType AgentId)
{
  return AddAgentWithIdToSet (Agents, PtrAgent, AgentId);
}

static AgentIdType 
GetIdInSet (AgentSet * PtrSet)
{
  ++PtrSet->currentId;
  return PtrSet->currentId;
}

static AgentIdType 
GetId ()
{
  return GetIdInSet (Agents);
}

static bool 
UnusedIdInSet (AgentSet * PtrSet, AgentIdType AgentId)
{
  return GetAgentWithIdFromSet (PtrSet, AgentId) == NULL;
}

static bool 
UnusedId (AgentIdType AgentId)
{
  return UnusedIdInSet (Agents, AgentId);
}


static AgentType *
CreateAgentOfType (AgentIdType DefaultTypeId)
{
  AgentType *DefaultAgent;
  AgentType *PtrAgent;

  PtrAgent = Agent_CreateAgent ();
  if (DefaultTypeId != 0)
    {
      if ((DefaultAgent = GetAgentWithIdFromSet (Default_Agents, DefaultTypeId)) != NULL)
	{
	  Agent_CopyAgent (PtrAgent, DefaultAgent);
	  return PtrAgent;
	}
      else
	Warning ("CreateAgentOfType", "Agent type does not exist. Using default.\n");
    }

  Agent_CopyAgent (PtrAgent, Default_Agent);
  return PtrAgent;
}

static AgentType *
CreateAgent ()
{
  return CreateAgentOfType (0);
}


AgentIdType 
AgentSet_CreateAgentTypeOfType (AgentIdType AgentId,
				AgentIdType DefaultTypeId)
{
  AgentType *PtrAgent;

  if (AgentId == 0)
    AgentId = GetIdInSet (Default_Agents);

  if (AgentId != 0)
    {
      PtrAgent = CreateAgentOfType (DefaultTypeId);

      if (!(AddAgentWithIdToSet (Default_Agents, PtrAgent, AgentId)))
	{
	  Warning ("AgentSet_CreateAgentTypeOfType",
		   "Cannot insert id in Dictionary! Serious...\n");
	  Agent_DestroyAgent (PtrAgent);
	  return 0;
	}
      DebugLevel (DL, "------- AgentSet_CreateAgentTypeOfType %d\n", AgentId);

      return AgentId;
    }
  else
    {
      Warning ("AgentSet_CreateAgentTypeOfType",
	       "Couldn't create Agent 'cause there is no id left.\n");
      return 0;
    }
}

AgentIdType 
AgentSet_CreateAgentOfType (AgentIdType AgentId,
			    AgentIdType DefaultTypeId)
{
  AgentType *PtrAgent;

  if (AgentId == 0)
    AgentId = GetId ();

  if (AgentId != 0)
    {
      PtrAgent = CreateAgentOfType (DefaultTypeId);

      if (!(AddAgentWithId (PtrAgent, AgentId)))
	{
	  Warning ("AgentSet_CreateAgentOfType",
		   "Cannot insert id in Dictionary! Serious...\n");
	  /*Agent_DestroyAgent(PtrAgent); */
	  /*return 0; */
	}
      DebugLevel (DL, "------- AgentSet_CreateAgentOfType %d\n", AgentId);
      return AgentId;
    }
  else
    {
      Warning ("AgentSet_CreateAgentOfType",
	       "Couldn't create Agent 'cause there is no id left.\n");
      return 0;
    }
}

/*-ftn-********************************************************************
 *
 * Funktion  : AgentSet_CreateAgent
 *
 * Beschreib.: Falls noch kein Agent existiert wird die AgentId 1 zurueckgeliefert,
 *             ansonsten die naechste freie Id.
 *
 * Parameter :
 *
 * Return    : AgentIdType
 *
 *************************************************************************/
AgentIdType 
AgentSet_CreateAgent (AgentIdType AgentId)

{
  AgentId = AgentSet_CreateAgentOfType (AgentId, 0);

  DebugLevel (DL, "------- CREATE AGENT ID  %d\n", AgentId);

  return AgentId;
}

AgentSet *
AgentSet_CreateAgentSet ()
{
  AgentSet *set;

  DebugLevel (DL, "Create AgentSet\n");

  set = (AgentSet *) malloc (sizeof (AgentSet));
  AS_CHECKMEM (set);
  set->currentId = 0;
  set->map = Dictionary_Create ();
  set->list.PtrHead = NULL;
  return set;
}

void 
AgentSet_DestroyAgentFromSet (AgentSet * PtrSet, AgentIdType Agent)
{
  AgentType *agent_ptr = GetAgentWithIdFromSet (PtrSet, Agent);

  if (agent_ptr)
    {
      if (PtrSet)
	{
	  AgentList_Remove (&PtrSet->list, Agent);
	  Dictionary_Delete (&PtrSet->map, Agent);
	}
      Agent_DestroyAgent (agent_ptr);
    }
  else
    Warning ("AgentSet_DestroyAgentFromSet", "Trying to destroy invalid agent.\n");
}

void 
AgentSet_DestroyAgent (AgentIdType Agent)
{
  AgentSet_DestroyAgentFromSet (Agents, Agent);
}

void 
AgentSet_DestroyAgentSet (AgentSet * PtrSet)
{
  IterStateType iter;
  AgentIdType agent;

  iter = AgentList_InitIteration (&PtrSet->list);
  agent = AgentList_Next (&PtrSet->list, &iter);
  while (agent)
    {
      AgentIdType next_agent;

      next_agent = AgentList_Next (&PtrSet->list, &iter);
      AgentSet_DestroyAgentFromSet (PtrSet, agent);
      agent = next_agent;
    }
}

void 
AgentSet_DestroyAgents ()
{
  AgentSet_DestroyAgentSet (Agents);
}

void 
AgentSet_Initialize ()
{
  if (AgentSet_Initialized == 0)
    {
      DebugLevel (DL, "Initialize AgentSet\n");

      if (!Default_Agent)
	{
	  Default_Agent = Agent_CreateAgent ();
	  Agent_SetXPosition (Default_Agent, -1);
	  Agent_SetYPosition (Default_Agent, -1);
	  Default_Agent->Orientation = -1;
	  Default_Agent->Age = -1;
	  Default_Agent->Energy = -1.0;
	}

      if (!Default_Agents)
	Default_Agents = AgentSet_CreateAgentSet ();
      if (!Agents)
	Agents = AgentSet_CreateAgentSet ();

      AgentSet_Initialized = 1;
    }
}

bool 
AgentSet_IsAgentType (AgentIdType AgentId)
{
  if (AgentId != 0)
    return GetAgentWithIdFromSet (Default_Agents, AgentId) != NULL;
  else
    return Default_Agent != NULL;
}

bool 
AgentSet_IsAgent (AgentIdType AgentId)
{
  return GetAgentWithId (AgentId) != NULL;
}
