
/*-module-*****************************************************************
 *
 * Modul     : Control (FUNKTIONSMODUL)
 *
 * Beschreib.:
 *
 * Autor     : Oliver Dolezal
 *
 * Datum     : 22.01.96
 *
 * Aender.   : 
 *
 *   geaendert am 07.02.96         von O.Dolezal          Bemerkungen
 *
 *--**********************************************************************/

/*-includes-*/

#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"
#include "../loader/Loader.h"
#include "../agents/AgentSet.h"
#include "../world/World.h"
#include "../interpreter/Interpreter.h"
#include "../dump/Dump.h"
#include "../agents/AgentTypeDefs.h"
#include "../io/IO.h"
#include "../random/Random.h"
#include "../random/DCount.h"

#include "../interpreter/FuncSem.h"
#include <string.h>
#include "Control.h"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#ifdef NeXT
#include <libc.h>
#else

#include <sys/stat.h>
#include <sys/param.h>

#if SYSTEM != SUNOS
int getopt (int argc, char **argv, char *optstring);
#endif
#if !(SYSTEM == linux)		/* Untersttzung fr Speicherberwachung auf SUNOS. */

#include <sys/types.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#include <unistd.h>

typedef struct prmap
  {
    caddr_t pr_vaddr;		/* Virtual address */
    u_long pr_size;		/* Size of mapping in bytes */
    u_long pr_pagesize;		/* pagesize in bytes for this mapping */
    off_t pr_off;		/* Offset into mapped object, if any */
    long pr_mflags;		/* Protection and attribute flags */
  }
prmap_t;

static int memfd;
#endif

extern char *optarg;
extern int optind;
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#define STR_EQUAL 0

/*-local-vars-*/
volatile static bool ContinueSimulation = TRUE;
static int MaxPosX = 0, MaxPosY = 0;
static unsigned NumberOfAgents = 0, Seed = 0;
static dcount WindUp;

int Global_WantVerboseOutput = 0;
int Global_UsePort = 0;
int Global_MemMax = 0;

static bool complete_map_and_agents ();
void signal_handler (int signo);


#if !(SYSTEM == linux)
int 
virtual_mem ()
{
  int pnr, i, memcount = 0;
  prmap_t **pmap_list;

  /* Speicherplatz reservieren, und Array fllen, size-Werte aufaddieren
   */
  pnr = ioctl (memfd, PIOCNMAP, pmap_list);
  pmap_list = malloc (sizeof (prmap_t *) * (pnr + 1));
  for (i = 0; i <= pnr; i++)
    pmap_list[i] = malloc (sizeof (prmap_t));
  ioctl (memfd, PIOCMAP, pmap_list);
  for (i = 0; i <= pnr; i++)
    {
      memcount += (pmap_list[i])->pr_size;
      free (pmap_list[i]);
    }
  free (pmap_list);
  return memcount;
}
#endif

bool 
Control_SetParameterValue (string Key, string Value)
	/*
	   Diese Funktion speichert globale Parameter aus dem Worldfile,
	   die hier benoetigt werden lokal, um auf diese zugreifen zu
	   koennen.
	 */
{
  bool wantedValue = FALSE;

  if (strcmp (Key, "WorldWidth") == 0)
    {
      MaxPosX = atoi (Value);
      wantedValue = TRUE;
    }
  else if (strcmp (Key, "RandomWindUpLow") == 0)
    {
      DCount_SetLow (&WindUp, strtoul (Value, (char **) NULL, 10));
      wantedValue = TRUE;
    }
  else if (strcmp (Key, "RandomWindUpHigh") == 0)
    {
      DCount_SetHigh (&WindUp, strtoul (Value, (char **) NULL, 10));
      wantedValue = TRUE;
    }
  else if (strcmp (Key, "WorldHeight") == 0)
    {
      MaxPosY = atoi (Value);
      wantedValue = TRUE;
    }
  else if (strcmp (Key, "NumberOfAgents") == 0)
    {
      NumberOfAgents = atoi (Value);
      wantedValue = TRUE;
    }
  else if (strcmp (Key, "Seed") == 0)
    {
      if (Seed == 0)
	Seed = atoi (Value);
      wantedValue = TRUE;
    }
  return wantedValue;
}

static bool 
init_simulation (char *filename)
{
#if !(SYSTEM == linux)
  char pidpath[30];
#endif

  if (filename)
    if (Loader_LoadFile (filename) == FALSE)
      return FALSE;

  /* Initialisierung des Signal-Handlers fuer
   * Abbruch der Simulation via CTRL-C oder
   * KILL
   */
  signal (SIGTERM, signal_handler);
  signal (SIGINT, signal_handler);

#if !(SYSTEM == linux)		/* ffnen des Proc Dateisystems zur Speicherberwachung */
  sprintf (pidpath, "/proc/%d\0", (int) getpid ());
  if ((memfd = open (pidpath, O_RDONLY)) == -1)
    Error ("Control", "Could not open /proc-fs");
#endif


/*  if (!complete_map_and_agents())
   return FALSE;
 */
  if (Seed == 0)
    {
      char buf[32];

      Seed = (unsigned) Random_Init (0L);
      Message ("Control", "Using seed %d (random value)\n", Seed);

      sprintf (buf, "%d", Seed);
      Loader_SetParameterValue ("Seed", buf);
    }
  else
    {
      Random_Init (Seed);
    }
  if (!DCount_IsNull (WindUp))
    {
      Message ("Control", "Winding up random generator times.\n");
      Random_WindUp (WindUp);
    }
  if (!complete_map_and_agents ())
    return FALSE;
  /* Vorher sollte der rand55 initialisiert sein. */
  Print ("\nUsing Seed: %u\n", Seed);

  DebugInfo ("Simulation initialized.\n");

  return TRUE;
}

static void
set_randomCount ()
{  
  char buf[12];
  dcount temp;
  
  temp = Random_Counted ();
  sprintf (buf, "%lu", DCount_GetLow (temp));
  Dump_SetParameterValue ("RandomWindUpLow", buf);
  sprintf (buf, "%lu", DCount_GetHigh (temp));
  Dump_SetParameterValue ("RandomWindUpHigh", buf);
}
  


static void 
end_simulation ()
{
  DebugInfo ("Finish simulation.\n");

  set_randomCount ();
  Dump_EndSimulation ();
  AgentSet_EndSimulation ();
  World_EndSimulation ();

  Dump_CloseFiles ();
}



void 
signal_handler (int signo)
     /*
        Diese Funktion ermoeglicht den Abbruch des Simulators ueber Signals.
        Wird dem Simulationsprozess das Signal SIGTERM geschickt, wird das
        Simulationssystem normal beendet, d.h. inklusive Dump. Beim Signal
        SIGINT bzw. Control-C wird die Ausfuehrung sofort abgebrochen.
      */
{
  ContinueSimulation = FALSE;

  Print ("\n-- User Abort -- got signal %d.\n", signo);

  /* das hier habe ich auskommentiert, da die Simulation ordentlich 
     beendet werden sollte. Zumindest die Ausgabekanaele sollten geschlossen
     werden, sonst haben wir unvollstaendige Dumpfiles. kodi 
     fflush(stdout);
     if (signo==SIGINT)
     exit(2);
     Der Timestep wird also ordentlich zuende gefhrt! jenne 
   */
}


static bool 
complete_agents ()
{
  PositionType currentPos;
  IterStateType state;
  AgentIdType currentId;

  state = AgentSet_InitAgentIteration ();
  while ((currentId = AgentSet_GetNextAgent (&state)) > 0)
    {
      AgentIdType worldId;

      currentPos = AgentSet_GetPosition (currentId);

      if ((currentPos.x != -1) && (currentPos.y != -1))
	{
	  worldId = World_AgentThere (currentPos);
	  if (worldId != currentId)
	    {
	      if (worldId == 0)
		World_SetAgent (currentId, currentPos);
	      else
		Error ("complete_map_and_agents",
		       "inconsistent data at (%d, %d): Agent %d != Map %d\n",
		       currentPos.x, currentPos.y, currentId, worldId);
	    }

	  if (!AgentSet_GetProgram (currentId))
	    AgentSet_SetProgram (currentId,
				 Interpreter_CreateRandomProgram ());
	}

    }
  return TRUE;
}

static bool 
complete_random_agents ()
{
  PositionType currentPos;
  IterStateType state;
  AgentIdType currentId;

  state = AgentSet_InitAgentIteration ();
  while ((currentId = AgentSet_GetNextAgent (&state)) > 0)
    {
      AgentIdType worldId;

      currentPos = AgentSet_GetPosition (currentId);

      if ((currentPos.x == -1) || (currentPos.y == -1))
	{
	  if (FuncSem_FindFreeRandomPosition (&currentPos, 100))
	    {
	      AgentSet_SetPosition (currentId, currentPos);
	      DebugInfo ("placing agent %d at (%d %d)\n",
			 currentId, currentPos.x, currentPos.y);
	    }
	  else
	    Error ("complete_map_and_agents",
		   "couldn't find position for Agent %d\n",
		   currentId);

	  worldId = World_AgentThere (currentPos);
	  if (worldId != currentId)
	    {
	      if (worldId == 0)
		World_SetAgent (currentId, currentPos);
	      else
		Error ("complete_map_and_agents",
		       "inconsistent data at (%d, %d): Agent %d != Map %d\n",
		       currentPos.x, currentPos.y, currentId, worldId);
	    }

	  if (!AgentSet_GetProgram (currentId))
	    AgentSet_SetProgram (currentId,
				 Interpreter_CreateRandomProgram ());
	}
    }
  return TRUE;
}

static bool 
complete_map_and_agents ()
{
  /* Dies ist eine erweiterte Funktion von classes2agents(),
     die nicht nur Typen in 'Instanzen' umwandelt, sondern
     auch die Map und die Agents angleicht und evtl.
     komplettiert
   */

  PositionType currentPos;
  AgentIdType currentId;

/*      while (NumberOfAgents > AgentSet_GetAgentCount())
   {
   AgentIdType          agentId;

   if (FuncSem_FindFreeRandomPosition(&currentPos, 1000))
   {
   agentId = AgentSet_CreateAgent(0);
   AgentSet_SetPosition(agentId, currentPos);
   World_SetAgent(agentId, currentPos);
   DebugInfo("placing random agent %d at position (%d, %d)\n", 
   agentId, currentPos.x, currentPos.y);
   }
   else
   Error("[Control]complete_map_and_agents", 
   "couldn't find position for random agent", agentId);
   } */
  AgentSet_FinishInitialization ();

  for (currentPos.y = 0; currentPos.y < MaxPosY; currentPos.y++)
    {
      for (currentPos.x = 0; currentPos.x < MaxPosX; currentPos.x++)
	{
	  currentId = World_AgentThere (currentPos);
	  if (currentId < 0)	/* Typ */
	    {
	      /* in der map ist ein Typ angegeben:
	         * erzeuge einen neuen Agenten von diesem Typ
	         * und setze ihn an die aktuelle Position */
	      AgentIdType newId;

	      newId = AgentSet_CreateAgentOfType (0, -currentId);
	      AgentSet_SetPosition (newId, currentPos);
	      World_SetAgent (newId, currentPos);
	    }
	  else if (currentId > 0)	/* Agent */
	    {
	      /* in der map ist ein Agent angegeben:
	         * wenn kein Agent mit der aktuellen Id existiert,
	         * wird einer erzeugt und die aktuelle Position gesetzt;
	         * sonst werden die beiden Positionen verglichen;
	         * bei Ungleichheit mit einem Fehler abgebrochen */

	      if (!AgentSet_IsAgent (currentId))
		{
		  AgentSet_CreateAgent (currentId);
		  AgentSet_SetPosition (currentId, currentPos);
		}
	      else
		{
		  PositionType agentPos;

		  agentPos = AgentSet_GetPosition (currentId);

		  if ((agentPos.x == -1) || (agentPos.y == -1))
		    {
		      AgentSet_SetPosition (currentId, currentPos);
		    }
		  else if ((agentPos.x != currentPos.x) ||
			   (agentPos.y != currentPos.y))
		    {
		      Error ("[Control]complete_map_and_agents",
			     "inconsistent data: agent %d\n",
			     currentId);
		    }
		}		/* if (!AgentSet...) else */
	    }			/* else if (currentId > 0) */
	}			/* for (CurrentPos.x... */
    }				/* for (CurrentPos.y... */

  complete_agents ();
  complete_random_agents ();

  return TRUE;
}



static void 
init_timestep (CountType TimeStep)
{
  DebugInfo ("Initialize timestep: %d\n", (int) TimeStep);

#if !(SYSTEM == linux)
  if (Global_MemMax < virtual_mem ())
    Error ("Control", "Virtual Memory reached Maximum.\n");
#endif
  set_randomCount ();
  IO_InitTimestep (TimeStep);
  Dump_InitTimeStep (TimeStep);
  AgentSet_InitTimestep (TimeStep);
  World_InitTimestep (TimeStep);
  Interpreter_InitTimestep (TimeStep);

  DebugInfo ("Initialization successful.\n");
}



static void 
simulate_timestep (CountType TimeStep)
{
  IterStateType IterState;
  AgentIdType CurrentAgentId;

  IterState = AgentSet_InitAgentIteration ();
  CurrentAgentId = AgentSet_GetNextAgent (&IterState);
  while (CurrentAgentId > 0)
    {
      Interpreter_EvalAgent (CurrentAgentId);
      CurrentAgentId = AgentSet_GetNextAgent (&IterState);
    }
}


static void 
conclude_timestep (CountType TimeStep)
{
  DebugInfo ("Concluding timestep: %d\n", (int) TimeStep);

  /* Achtung! Das ist zwar haesslich, aber die Reihenfolge der Aufrufe
     ist von Bedeutung. Also bitte nicht aendern. */

  set_randomCount ();
  Dump_ConcludeTimeStep (TimeStep);
  AgentSet_ConcludeTimestep (TimeStep);
  Interpreter_ConcludeTimestep (TimeStep);
  World_ConcludeTimestep (TimeStep);
  IO_ConcludeTimestep (TimeStep);

  DebugInfo ("Timestep finished.\n");
}


static void 
usage (const char *Name)
{
  fprintf (stderr, "%s [-s Timesteps] [-r Randomizer Seed] [-d] "
	   "[-w Random Count] [-m MaxMemUsage] [-D DebugLevel] [-v] worldfile\n", Name);
}

static bool 
read_defaults ()
{
  char *path;
  char defaults_file[MAXPATHLEN + 1];
  struct stat stat_buf;

  if (stat ("./.philia", &stat_buf) == 0)
    return Loader_LoadFile ("./.philia");

  if ((path = getenv ("PHILIAPATH")))
    {
      strcpy (defaults_file, path);
      strcat (defaults_file, "/.philia");
      if (stat (defaults_file, &stat_buf) == 0)
	return Loader_LoadFile (defaults_file);
    }

  if ((path = getenv ("HOME")))
    {
      strcpy (defaults_file, path);
      strcat (defaults_file, "/.philia");
      if (stat (defaults_file, &stat_buf) == 0)
	return Loader_LoadFile (defaults_file);
    }

  Warning ("Control", "couldn't find '.philia'\n");
  return FALSE;
}

int 
main (int argc, char *argv[])
{
  CountType CurrentTimeStep, MaxTimeStep;
  char optchar;
  long MaxTimeInSeconds;
  bool WantTimeWatch = FALSE, forever, opterr = FALSE;


  MaxTimeStep = 0;
  Global_WantDebugOutput = 0;
  Global_DebugLevel = 0;
  Global_WantVerboseOutput = 0;

  Random_Reset();
  DCount_Reset (&WindUp);
  while ((optchar = getopt (argc, argv, "s:r:n:dvhm:t:D:")) != EOF)
    switch (optchar)
      {
      case 's':
	MaxTimeStep = (CountType) atoi (optarg);
	break;
      case 'r':
	Seed = atoi (optarg);
	break;
      case 'n':
	Global_UsePort = atoi (optarg);
	break;
      case 'd':
	Global_WantDebugOutput = TRUE;
	break;
      case 't':
	WantTimeWatch = TRUE;
	MaxTimeInSeconds = atol (optarg) * 60;
	break;
      case 'D':
	Global_WantDebugOutput = TRUE;
	Global_DebugLevel = atoi (optarg);
	break;
      case 'm':
	Global_MemMax = atoi (optarg);
	break;
      case 'v':
	Global_WantVerboseOutput = TRUE;
	break;
      case 'h':
      default:
	opterr = TRUE;
	break;
      }

  if (opterr || (optind + 1 < argc))
    {
      usage (argv[0]);
      exit (2);
    }

  read_defaults ();

  if (MaxTimeStep < 0)
    {
      Error ("main", "Number of timesteps has to be non-negative.\n");
    }
  Message ("Control", "MaxTimeStep: %d\n", (int) MaxTimeStep);

  if (MaxTimeStep == 0)
    forever = TRUE;
  else
    forever = FALSE;

  DebugInfo ("forever: %d\n", (int) forever);

  /* Initialisierung des IO_Ports fuer den Dump */
  IO_MakePort ();

  if (optind > (argc - 2))
    {
    }

  if (!init_simulation (argv[optind]))
    {
      Error ("main", "Problems at initialization.\n");
    }

  CurrentTimeStep = 1;

  while (ContinueSimulation && (forever || (CurrentTimeStep <= MaxTimeStep)))
    /* 
       Die Schleife bricht ab, wenn die gewuenschte Anzahl an Simulationslaeufen
       stattgefunden hat,oder wenn das SIGNAL zum Abbruch des Prozesses gesendet
       wurde.
     */
    {
      if (Global_WantVerboseOutput)
	{
	  printf ("-----------------------\n"
		  "Current timestep: %d\n", (int) CurrentTimeStep);
	  printf ("agents: %d\n", AgentSet_GetAgentCount ());

	}
      else
	{
	  if ((CurrentTimeStep - 1) % 10 == 0)
	    {
	      if ((CurrentTimeStep - 1) % 100 == 0)
		fprintf (stdout, "\ntime %8d, agents %5d,  ",
			 CurrentTimeStep, AgentSet_GetAgentCount ());
	      fprintf (stdout, "#");
	      fflush (stdout);
	    }
	}

      init_timestep (CurrentTimeStep);
      simulate_timestep (CurrentTimeStep);
      if (AgentSet_GetAgentCount () == 0)
	{
	  printf ("Agents died out.\n");

	  if (Dump_WantDumpStatistics ())
	    {

	      char buf[200];

	      sprintf (buf, "#! Agents died out after %d timesteps\n",
		       CurrentTimeStep);

	      Dump_DumpStatistics (buf);
	    }

	  break;
	}


      conclude_timestep (CurrentTimeStep);
      CurrentTimeStep++;
      /* Breche ab wenn Zeit berschritten */
      if (WantTimeWatch && ((clock () / CLOCKS_PER_SEC) >= MaxTimeInSeconds))
	{
	  puts ("\nTimeout.");
	  break;
	}
    }
  end_simulation ();
  puts ("\nSimulation end.\n");

  return 0;
}
