%option lex-compat stack nostdinit
%x COM HDR QUT

%{

#include "../misc/TypeDefs.h"
#include "../misc/DynaString.h"
#include "../misc/Messages.h"
#include "../io/IO.h"
#include "DumpPrivate.h"
#include "DumpParser.h"
#include <strings.h>
#include <stdio.h>

#ifndef FALSE
#	define FALSE	(1 == 0)
#endif
#ifndef TRUE
#	define TRUE		(1 == 1)
#endif

static void ReturnBufferToFile(FILE *File, YY_BUFFER_STATE Buf);

static char		current_quote_char = '\0';
static char		is_matching_quote(char quote_start, char quote_end);

int		DUMPparse(void);			/* in DumpParser.y */

/* LEX-Rules ----------------------------------------------------------- */
%}

word				[[:alnum:]]+
string				[[:alnum:]/][[:graph:]]*

%%

<COM>"END_COMMENT"		{ yy_pop_state();	}

<INITIAL,COM>"COMMENT"	{ yy_push_state(COM); }

\#.*\n              	;

<COM>.*             |
<COM>\n             ;


"HEADER"			{
						yy_push_state(HDR);

						return HEADER;
					}

<HDR>"END_HEADER"	{
						yy_pop_state();
						ReturnBufferToFile(DUMPin, YY_CURRENT_BUFFER);
						yyterminate();
					}

[=]					{ return (yytext[0]); }

"DUMP"				{ return DUMP; }
"END_DUMP"			{
						ReturnBufferToFile(DUMPin, YY_CURRENT_BUFFER);
						yyterminate();
					}

"CHANNEL"			{ return CHANNEL; }
"END_CHANNEL"		{ return END_CHANNEL; }

"TYPE"				{ return TYPE; }
"FILE"				{ DUMPlval.typint = IO_FileChannel; return INT; }
"SOCKET"			{ DUMPlval.typint = IO_SocketChannel; return INT; }
"PIPE"				{ DUMPlval.typint = IO_PipeChannel; return INT; }
"COMMAND"			{ DUMPlval.typint = IO_CommandChannel; return INT; }
"CLOSE"				{ DUMPlval.typint = -1; return INT; }
"REOPEN"			{ DUMPlval.typint = -2; return INT; }

<INITIAL>"MODE"		{ return MODE; }
<INITIAL>"APPEND"	{ return APPEND; }
<INITIAL>"OVERWRITE" { return OVERWRITE; }

<INITIAL>"FLUSH"	{ DUMPlval.typint = _IONBF; return FLUSH; }
<INITIAL>"NO_BUFFER"	{ DUMPlval.typint = _IONBF; return FLUSH; }
<INITIAL>"LINE_BUFFER"	{ DUMPlval.typint = _IOLBF; return FLUSH; }
<INITIAL>"BUFFER"		{ DUMPlval.typint = _IOFBF; return FLUSH; }

{string}			{
						strncpy(DUMPlval.typstr, yytext, 1024);
						return STR;
					}

<HDR>.*				{
						strncpy(DUMPlval.typstr, yytext, 1024);
						return STR;
					};


[ \t\n]				{}

<QUT>[[:graph:]]+	{
						if (is_matching_quote(current_quote_char, 
											yytext[yyleng - 1]))
						{
							yy_pop_state();
							strncpy(DUMPlval.typstr, yytext, yyleng - 1);
							return QUOTE_END;
						}
						else
						{
							yymore();
						}
					}

<QUT>[[:blank:]]		{
						yymore();
					}

[^[:alnum:]]		{
						current_quote_char = yytext[0];
						yy_push_state(QUT);
						return QUOTE_START;
					}
					;

%%

/* --------------------------------------------------------------------- */

static bool is_matching_quote(char quote_start, char quote_end)
{
	switch (quote_start)
	{
		case '(':
			return quote_end == ')';
		case '[':
			return quote_end == ']';
		case '{':
			return quote_end == '}';
		case '<':
			return quote_end == '>';
		default:
			return quote_start == quote_end;
	}
}

/* --------------------------------------------------------------------- */

int DUMPwrap()
{
	return 1;
}

static void ReturnBufferToFile(FILE *File, YY_BUFFER_STATE Buf)
{
  /* 
  Diese Funktion liest den uebergebenen lex-Buffer aus
  und haengt alle Zeichen, die aus dem Stream file
  ausgelesen, jedoch noch nicht gematcht wurden, wieder
  in den Stream zurueck. 
  */

  int  length; 
  char *begin;

  length = strlen (Buf->yy_ch_buf);

  begin = &(Buf->yy_ch_buf[length+1]);
 
  length = strlen (begin); 

  fseek (File, -length, 1);
}


bool Dump_ParseSection(FILE* Stream)
{
	if (DUMPin)
	{
		yyrestart(Stream);
		YY_FLUSH_BUFFER;
	}
	else
		DUMPin = Stream;

	return DUMPparse() == 0;
}

/*
int main(int argc, char* argv[])
{
	FILE* stream = fopen(argv[1], "r");

	return Dump_ParseDump(stream);
}*/
