
%{

#include "DumpPrivate.h"
#include "../misc/DynaString.h"
#include "../misc/Messages.h"

#ifdef YYDEBUG
  int yydebug = 1;
#endif

extern int   			DUMPlineno;
extern FILE* 			DUMPin;

static DynaStringType*	currentHeader = NULL;
static DumpChannel*		currentDumpChannel = NULL;

static int				currentType;
static char				currentMode[2];
static int				currentFlush;


int						DUMPlex(void);
int						DUMPerror(char*);

%}

%union
{
	int				typint;
	char			typstr[1025];
}

%token	<typint>	INT
%token	<typstr>	STR

%token				HEADER
%token				END_HEADER

%token				DUMP
%token				END_DUMP
%token				CHANNEL
%token				END_CHANNEL

%token				TYPE
%token				MODE
%token				APPEND
%token				OVERWRITE
%token	<typint>	FLUSH

%token				QUOTE_START
%token	<typstr>	QUOTE_END

%type	<typstr>	string;

%%

section					:	header_section
						|	dump_section
						;

header_section			:	HEADER
							{
								if (!currentHeader)
									currentHeader = DynaString_Alloc(512);
								else
									DynaString_Reset(currentHeader);
							}
							strings
						;

strings					:	/*	*/
						|	strings STR
							{
								DynaString_Append(currentHeader, "#");
								DynaString_Append(currentHeader, $2);
								DynaString_Append(currentHeader, "\n");
							}
						;

dump_section			:	DUMP
							channel_sections
						;

channel_sections		:	/* */
						|	channel_sections
							CHANNEL
							string
							{
								strcpy(currentMode, "w");
								currentType = 1;
								currentFlush = 0;
							}
							type_mode
							{
								if (currentType >= 0)
								{
									Message("Creating Channel %s %d\n", 
													$3, currentType, 
													currentMode, currentFlush);

									currentDumpChannel = 
										Dump_CreateChannel($3, currentType, 
													currentMode, currentFlush);

									Dump_SetChannelHeader(currentDumpChannel, 
										currentHeader);
								}
								else if (currentType == -1)
								{
									Message("Closing Channel %s\n", $3);
									Dump_CloseChannel($3);
								}
								else if (currentType == -2)
								{
									Message("Reopening Channel %s\n", $3);
									currentDumpChannel = 
										Dump_ReopenChannel($3, currentMode);
								}
							}
							display_vars
							END_CHANNEL
						;

type_mode				:	/* */
						|	type_mode TYPE INT
							{
								currentType = $3;
							}
						|	type_mode MODE APPEND
							{
								strcpy(currentMode, "a");
							}
						|	type_mode MODE OVERWRITE
							{
								strcpy(currentMode, "w");
							}
						|	type_mode FLUSH
							{
								currentFlush = 1;
							} 
						;


display_vars			:	/* */
						|	display_vars
							display_var
						;

display_var				:	STR '=' string
							{
								Message("	%s = %s\n", $1, $3);
								Dump_SetParameterValueForChannel
												(currentDumpChannel, $1, $3);
							}
						;

string					:	STR
							{
								strcpy($$, $1);
							}
						|	QUOTE_START QUOTE_END
							{
								strcpy($$, $2);
							}
						;


%%

int DUMPerror (char *s)
{
	Error("DUMPerror", "%s, line %d\n", s, DUMPlineno);
	return 0;
}

