%option lex-compat stack
%x COM

%{

#include "../misc/TypeDefs.h"
#include "../misc/DynaString.h"
#include "DumpPrivate.h"
#include <strings.h>

static void ReturnBufferToFile(FILE *File, YY_BUFFER_STATE Buf);

DynaStringType*				headerBuffer;

/* LEX-Rules ----------------------------------------------------------- */
%}

string				[a-zA-Z0-9_.]+
ganzzahl			[+-]?[0-9]+

%%

<COM>"END_COMMENT"  	yy_pop_state();

<INITIAL,COM>"COMMENT"	yy_push_state(COM);

\#.*\n              | 
		    ;

<INITIAL>"HEADER"	;
<INITIAL>"END_HEADER"	{
						DebugInfo("%s", DynaString_CString(headerBuffer));

						ReturnBufferToFile(HEADERin, YY_CURRENT_BUFFER);
						yyterminate();
					}

.*					{
						DynaString_Append(headerBuffer, yytext);
						DynaString_Append(headerBuffer, "\n");
					}
%%

/* --------------------------------------------------------------------- */

int HEADERerror (char *s)
{
 Error("HEADERerror", "%s, line %d\n", s, HEADERlineno);
 return 0;
}


int HEADERwrap()
{
	return 1;
}

static void ReturnBufferToFile(FILE *File, YY_BUFFER_STATE Buf)
{
  /* 
  Diese Funktion liest den uebergebenen lex-Buffer aus
  und haengt alle Zeichen, die aus dem Stream file
  ausgelesen, jedoch noch nicht gematcht wurden, wieder
  in den Stream zurueck. 
  */

  int  length; 
  char *begin;

  length = strlen (Buf->yy_ch_buf);

  begin = &(Buf->yy_ch_buf[length+1]);
 
  length = strlen (begin); 

  fseek (File, -length, 1);
}

bool Dump_ParseHeader(FILE* Stream)
{
	WorldFileType*		worldfile;

	Message("[Dump]Dump_ParseDump: called with %p (was %p)\n", 
			Stream, HEADERin);

	worldfile = Dump_GetWorldFileForStream(Stream);
	if (!worldfile)
		Dump_AddWorldFile(Dump_CreateWorldFile(NULL, Stream, NULL));

	if (!worldfile->scanBuffer)
		worldfile->scanBuffer = yy_create_buffer(Stream, 40);

	headerBuffer = Dump_GetHeaderBufferForWorldFile(worldfile);

	Message("Worldfile: %p Buffer %p\n", worldfile, worldfile->scanBuffer);

	HEADERin = Stream;
	yy_switch_to_buffer(worldfile->scanBuffer);

	while (HEADERlex() != 0);

	yy_flush_buffer(worldfile->scanBuffer);
	return TRUE;
}

/*
int main(int argc, char* argv[])
{
	FILE* stream = fopen(argv[1], "r");

	Dump_ParseHeader(stream);
}*/
