
#include "DumpPrivate.h"
#include "../misc/Memory.h"
#include <stdlib.h>

WorldFileType *WorldFiles = NULL;

WorldFileType *
Dump_CreateWorldFile (const char *Name, FILE * Stream, void *Buffer)
{
  WorldFileType *worldfile;

  worldfile = Allocate (sizeof (WorldFileType));

  if (Name && *Name)
    strcpy (worldfile->name, Name);
  else
    worldfile->name[0] = '\0';

  worldfile->inStream = Stream;
  worldfile->scanBuffer = Buffer;

  worldfile->header = worldfile->parameters = NULL;

  worldfile->next = NULL;

  return worldfile;
}

void 
Dump_DestroyWorldFiles ()
{
  WorldFileType *worldfile;

  for (worldfile = WorldFiles; worldfile;)
    {
      WorldFiles = worldfile->next;

      if (worldfile->header)
	DynaString_Free (worldfile->header);
      if (worldfile->parameters)
	DynaString_Free (worldfile->parameters);
      Free (worldfile);

      worldfile = WorldFiles;
    }
}

void 
Dump_AddWorldFile (WorldFileType * worldfile)
{
  if (worldfile->next && WorldFiles)
    {
      WorldFileType *wf;

      wf = WorldFiles;
      while (wf->next != worldfile)
	wf = wf->next;

      wf->next = worldfile->next;
    }

  worldfile->next = WorldFiles;
  WorldFiles = worldfile;
}

WorldFileType *
Dump_GetWorldFileForStream (FILE * Stream)
{
  WorldFileType *worldfile;

  for (worldfile = WorldFiles; worldfile; worldfile = worldfile->next)
    if (Stream == worldfile->inStream)
      return worldfile;

/*      worldfile = Dump_CreateWorldFile(NULL, Stream, NULL);
   Dump_AddWorldFile(worldfile);
   return worldfile; */
  return NULL;
}

DynaStringType *
Dump_GetHeaderBufferForWorldFile (WorldFileType * WorldFile)
{
  if (WorldFile)
    {
      if (!WorldFile->header)
	WorldFile->header = DynaString_Alloc (256);

      return WorldFile->header;
    }
  return NULL;
}

DynaStringType *
Dump_GetChannelBufferForWorldFile (WorldFileType * WorldFile)
{
  if (!WorldFile->channels)
    WorldFile->channels = DynaString_Alloc (256);

  return WorldFile->channels;
}

DynaStringType *
Dump_GetParameterBufferForWorldFile (WorldFileType * WorldFile)
{
  if (!WorldFile->parameters)
    WorldFile->parameters = DynaString_Alloc (128);

  return WorldFile->parameters;
}
