/*-module-*****************************************************************
 *
 * Modul     : FuncSem (FUNKTIONSMODUL)
 *
 * Beschreib.: In diesem File werden die einzelnen Befehle der
 *			   Strategiesprache implementiert. In der Funktion MATE
 *			   befindet sich der GP-Teil des Simulators.
 *
 * Autor     : Johannes & Martin
 *
 * Datum     : 15.01.1996
 *
 * Aender.   : 08.05.1996 Martin
 * 
 *
 *--**********************************************************************/

#include <stdlib.h>
#include <math.h>
#include <string.h>


#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"
#include "../misc/Memory.h"

#include "../world/World.h"
#include "../agents/AgentSet.h"
#include "../agents/AgentTypeDefs.h"
#include "../program/AgentPrgFunc.h"
#include "../dump/Dump.h"
#include "../random/Random.h"

#include "FuncSem.h"
#include "Interpreter.h"
#include "InterpreterPrivate.h"
/*--*/


/*-defines-*/

#define RECTANGLE	0
#define CYLINDER	1
#define TORUS		2
#define RANDOM(range) ((long) Random_LRand((long) range))

static bool Mute = 0;
static double Mutation = 0.0;
static bool CMute = 0;
static double CMutation = 0.0;

static EnergyType AttackFailureEnergy = 0;

static int DefenderKeepEnergy = 0;

 /* static bool TurnType = 0; */

static FactorType IdleFactor = 0.0;

static int WorldSizeX, WorldSizeY, Topology, CheapMating = 1, FastAttack = 0,
  MaxDepth = 0, MinDepth = 0;

static CountType OffspringRadius, MaxProgramDepth;

static FactorType MateProb = 1.0;


int 
sign (int zahl)
{
  if (zahl < 0)
    return -1;
  else
    return 1;
}

int 
min (int a, int b)
{
  return (a > b) ? b : a;
}



/*************************************************************************
 * Diese Funktion liefert zu einer gegebenen Position und Orientierung
 * die Koordinaten des direkt vor der Position befindlichen Felds.
 * Bewegt man sich ueber den Rand hinaus (abhaengig von der Topologie),
 * so wird kein Fehler ausgegeben, sondern es werden nicht existierende
 * rein rechnerische Positionen ermittelt.
 * Diese Funktion sollte auf jeden Fall mit der Funktion
 * "IsPositionValid" verwendet werden, da sie diese Eigenheiten von
 * z.B. negativen Positionen entsprechend beruecksichtigt.
 *************************************************************************/


static PositionType 
GetHerePosition (PositionType Position,
		 OrientationType Orientation)
{
  switch (Orientation)
    {
    case 0:
      if (Position.y <= 0)
	switch (Topology)
	  {
	  case CYLINDER:
	  case RECTANGLE:
	    Position.y += -1;
	    break;
	  case TORUS:
	    Position.y = WorldSizeY - 1;
	  }
      else
	Position.y--;
      break;
    case 45:
      if (Position.y <= 0)
	switch (Topology)
	  {
	  case CYLINDER:
	  case RECTANGLE:
	    Position.y += -1;
	    break;
	  case TORUS:
	    Position.y = WorldSizeY - 1;
	  }
      else
	Position.y--;
      if (Position.x >= WorldSizeX - 1)
	switch (Topology)
	  {
	  case RECTANGLE:
	    Position.x += 1;
	    break;
	  case CYLINDER:
	  case TORUS:
	    Position.x = 0;
	  }
      else
	Position.x++;
      break;
    case 90:
      if (Position.x >= WorldSizeX - 1)
	switch (Topology)
	  {
	  case RECTANGLE:
	    Position.x += 1;
	    break;
	  case CYLINDER:
	  case TORUS:
	    Position.x = 0;
	  }
      else
	Position.x++;
      break;
    case 135:
      if (Position.y >= WorldSizeY - 1)
	switch (Topology)
	  {
	  case CYLINDER:
	  case RECTANGLE:
	    Position.y += 1;
	    break;
	  case TORUS:
	    Position.y = 0;
	  }
      else
	Position.y++;
      if (Position.x >= WorldSizeX - 1)
	switch (Topology)
	  {
	  case RECTANGLE:
	    Position.x += 1;
	    break;
	  case CYLINDER:
	  case TORUS:
	    Position.x = 0;
	  }
      else
	Position.x++;
      break;
    case 180:
      if (Position.y >= WorldSizeY - 1)
	switch (Topology)
	  {
	  case CYLINDER:
	  case RECTANGLE:
	    Position.y += 1;
	    break;
	  case TORUS:
	    Position.y = 0;
	  }
      else
	Position.y++;
      break;
    case 225:
      if (Position.y >= WorldSizeY - 1)
	switch (Topology)
	  {
	  case CYLINDER:
	  case RECTANGLE:
	    Position.y += 1;
	    break;
	  case TORUS:
	    Position.y = 0;
	  }
      else
	Position.y++;
      if (Position.x <= 0)
	switch (Topology)
	  {
	  case RECTANGLE:
	    Position.x += -1;
	    break;
	  case CYLINDER:
	  case TORUS:
	    Position.x = WorldSizeX - 1;
	  }
      else
	Position.x--;
      break;
    case 270:
      if (Position.x <= 0)
	switch (Topology)
	  {
	  case RECTANGLE:
	    Position.x += -1;
	    break;
	  case CYLINDER:
	  case TORUS:
	    Position.x = WorldSizeX - 1;
	  }
      else
	Position.x--;
      break;
    case 315:
      if (Position.y <= 0)
	switch (Topology)
	  {
	  case CYLINDER:
	  case RECTANGLE:
	    Position.y += -1;
	    break;
	  case TORUS:
	    Position.y = WorldSizeY - 1;
	  }
      else
	Position.y--;
      if (Position.x <= 0)
	switch (Topology)
	  {
	  case RECTANGLE:
	    Position.x += -1;
	    break;
	  case CYLINDER:
	  case TORUS:
	    Position.x = WorldSizeX - 1;
	  }
      else
	Position.x--;
      break;
    default:

      break;
    }
  return Position;
}

/**************************************************************************
 *
 * Diese Funktion ueberprueft, ob eine Position gueltig ist, d.h. ob
 * sich die Koordinaten zwischen 0 und WorldSize bewegen. Sie kann
 * sinnvollerweise zusammen mit der Funktion GetHerePosition verwendet
 * werden.
 *
 *************************************************************************/
static bool 
IsPositionValid (PositionType Position)
{
  if ((Position.x >= 0) &&
      (Position.x < WorldSizeX) &&
      (Position.y >= 0) &&
      (Position.y < WorldSizeY))
    return TRUE;
  else
    return FALSE;
}


/**************************************************************************
 *
 * Diese Funktion ermittelt eine Position fuer zu erzeugenden
 * Nachkommen unter Beruecksichtigung des "OffspringRadius".
 * Gibt es kein freies Feld, so wird "flag" auf "FALSE" gesetzt,
 * ansonsten auf "TRUE".
 *
 * GetOffspringPosition wird nur von FuncSem_Mate aufgerufen.
 *
 *************************************************************************/
PositionType 
GetOffspringPosition (PositionType Position, bool * flag)
{
  int i;
  int j;
  int k;
  PositionType OffspringPosition;
  OrientationType OffspringOrientation;

  *flag = TRUE;
  OffspringPosition = Position;
  for (i = 1; i <= OffspringRadius; i++)
    {
      OffspringPosition = GetHerePosition (OffspringPosition, 45);
      OffspringOrientation = 180;
      for (j = 1; j <= 4; j++)
	{

	  for (k = 1; k <= 2 * i; k++)
	    {
	      OffspringPosition = GetHerePosition (OffspringPosition,
						   OffspringOrientation);
	      if (IsPositionValid (OffspringPosition))
		if ((World_AgentThere (OffspringPosition) == 0) &&
		    (World_GetEnergy (OffspringPosition) == 0))
		  return OffspringPosition;
	    }
	  OffspringOrientation += 90;
	  if (OffspringOrientation == 360)
	    OffspringOrientation = 0;
	}
    }
  *flag = FALSE;
  return OffspringPosition;
}

bool 
FuncSem_FindFreePositionAround (PositionType * Position, unsigned radius)
{
  return FALSE;
}

bool 
FuncSem_FindFreeRandomPosition (PositionType * Position, unsigned tries)
{
  
  while (tries--)
    {
      Position->x = RANDOM (WorldSizeX);
      Position->y = RANDOM (WorldSizeY);
      if (!World_AgentThere (*Position) && !World_GetEnergy (*Position))
	return TRUE;
    }
  return FALSE;
}



/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Hungry
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/

bool
FuncSem_Hungry (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  
  if (AgentSet_GetEnergy (AgentID) <=
      AgentSet_GetHungerFactor (AgentID) * AgentSet_GetMaxEnergy (AgentID))
    return TRUE;
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_UnderAttack
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_UnderAttack (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  
  if (AgentSet_GetAttacks (AgentID) == 0)
    return FALSE;
  else
    return TRUE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Sound
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_Sound (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{

  int i;
  CountType MaxFriends;
  CountType Range;
  AgentIdType FriendID;
  PositionType PositionAgent;
  PositionType PositionFriend;
  IterStateType FriendIter;
  int DiffX;
  int DiffY;


  MaxFriends = AgentSet_GetMaxFriends (AgentID);
  Range = AgentSet_GetHearingRange (AgentID);
  PositionAgent = AgentSet_GetPosition (AgentID);
  FriendIter = AgentSet_InitFriendIteration (AgentID);

  for (i = 1; i <= MaxFriends; i++)
    {
      FriendID = AgentSet_GetNextFriend (AgentID, &FriendIter);
      if ((FriendID != 0) && (AgentSet_GetScreams (FriendID)))
	{
	  PositionFriend = AgentSet_GetPosition (FriendID);
	  DiffX = abs (PositionAgent.x - PositionFriend.x);
	  DiffY = abs (PositionAgent.y - PositionFriend.y);
	  switch (Topology)
	    {
	    case 0:
	      if ((DiffX <= Range) && (DiffY <= Range))
		return TRUE;
	      break;

	    case 1:
	      if (
		   ((DiffX <= Range) && (DiffY <= Range))
		   ||
		   (((WorldSizeX - DiffX) <= Range) && (DiffY <= Range))
		)
		return TRUE;
	      break;

	    case 2:
	      if (
		   ((DiffX <= Range) || ((WorldSizeX - DiffX) <= Range))
		   &&
		   ((DiffY <= Range) || ((WorldSizeY - DiffY) <= Range))
		)
		return TRUE;
	      break;
	    }
	}
    }
  return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Mem
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_Mem (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{

  return AgentSet_GetMemory (AgentID);

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_AgentHere
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_AgentHere (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  PositionType Position;
  AgentIdType AgentID2;


  Position = GetHerePosition (AgentSet_GetPosition (AgentID),
			      AgentSet_GetOrientation (AgentID));
  if (IsPositionValid (Position))
    {
      AgentID2 = World_AgentThere (Position);
      if (AgentID2 != 0)
	{
	  AgentSet_MeetFriend (AgentID, AgentID2);
	  return TRUE;
	}
      else
	return FALSE;
    }
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_AgentAhead
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_AgentAhead (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  int flag = 0;
  CountType VisionRange;
  CountType Lauf = 1;
  PositionType Position;
  OrientationType Orientation;
  AgentIdType AgentID2;


  VisionRange = AgentSet_GetVisionRange (AgentID);
  Position = AgentSet_GetPosition (AgentID);
  Orientation = AgentSet_GetOrientation (AgentID);

  while ((flag == 0) && (Lauf <= VisionRange))
    {
      Position = GetHerePosition (Position, Orientation);
      if (IsPositionValid (Position))
	{
	  AgentID2 = World_AgentThere (Position);
	  if (AgentID2 != 0)
	    {
	      AgentSet_MeetFriend (AgentID, AgentID2);
	      flag = 1;
	    }
	  else if (World_GetEnergy (Position) != 0)
	    flag = 2;
	}
      else
	flag = 3;
      Lauf++;
    }

  if (flag == 1)
    return TRUE;
  else
    return FALSE;
}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_FoodHere
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_FoodHere (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  PositionType Position;


  Position = GetHerePosition (AgentSet_GetPosition (AgentID),
			      AgentSet_GetOrientation (AgentID));
  if (IsPositionValid (Position))
    if ((World_GetEnergy (Position) != 0) && AgentSet_CanEat (AgentID, World_GetFoodCarrionType (Position)))
      return TRUE;
    else
      return FALSE;
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_FoodAhead
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_FoodAhead (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  int flag = 0;
  CountType VisionRange;
  CountType Lauf = 1;
  PositionType Position;
  OrientationType Orientation;


  VisionRange = AgentSet_GetVisionRange (AgentID);
  Position = AgentSet_GetPosition (AgentID);
  Orientation = AgentSet_GetOrientation (AgentID);

  while ((flag == 0) && (Lauf <= VisionRange))
    {
      Position = GetHerePosition (Position, Orientation);
      if (IsPositionValid (Position))
	{
	  if ((World_GetEnergy (Position) != 0) && AgentSet_CanEat (AgentID, World_GetFoodCarrionType (Position)))
	    flag = 1;
	  else if ((World_AgentThere (Position) != 0) || (World_GetEnergy (Position) != 0))
	    flag = 2;
	}
      else
	flag = 3;
      Lauf++;
    }

  if (flag == 1)
    return TRUE;
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_FriendHere
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_FriendHere (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  PositionType Position;
  AgentIdType AgentID2;


  Position = GetHerePosition (AgentSet_GetPosition (AgentID),
			      AgentSet_GetOrientation (AgentID));
  if (IsPositionValid (Position))
    {
      AgentID2 = World_AgentThere (Position);
      if (AgentID2 && AgentSet_IsFriendOf (AgentID, AgentID2))
	{
	  AgentSet_MeetFriend (AgentID, AgentID2);
	  return TRUE;
	}
      else
	return FALSE;
    }
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_FriendAhead
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_FriendAhead (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  int flag = 0;
  CountType VisionRange;
  CountType Lauf = 1;
  PositionType Position;
  OrientationType Orientation;
  AgentIdType AgentID2;


  VisionRange = AgentSet_GetVisionRange (AgentID);
  Position = AgentSet_GetPosition (AgentID);
  Orientation = AgentSet_GetOrientation (AgentID);

  while ((flag == 0) && (Lauf <= VisionRange))
    {
      Position = GetHerePosition (Position, Orientation);
      if (IsPositionValid (Position))
	{
	  AgentID2 = World_AgentThere (Position);
	  if (AgentID2 != 0)
	    flag = 1;
	  else if (World_GetEnergy (Position) != 0)
	    flag = 2;
	}
      else
	flag = 3;
      Lauf++;
    }

  if ((flag == 1) && (AgentSet_IsFriendOf (AgentID, AgentID2)))
    {
      AgentSet_MeetFriend (AgentID, AgentID2);
      return TRUE;
    }
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_NothingHere
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_NothingHere (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  PositionType Position;


  Position = GetHerePosition (AgentSet_GetPosition (AgentID),
			      AgentSet_GetOrientation (AgentID));
  if (IsPositionValid (Position))
    if ((World_AgentThere (Position) == 0) &&
	(World_GetEnergy (Position) == 0))
      return TRUE;
    else
      return FALSE;
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_NothingAhead
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_NothingAhead (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  int flag = 0;
  CountType VisionRange;
  CountType Lauf = 1;
  PositionType Position;
  OrientationType Orientation;


  VisionRange = AgentSet_GetVisionRange (AgentID);
  Position = AgentSet_GetPosition (AgentID);
  Orientation = AgentSet_GetOrientation (AgentID);
     
  while ((flag == 0) && (Lauf <= VisionRange))
    {
      Position = GetHerePosition (Position, Orientation);
      if (IsPositionValid (Position))
	{
	  if ((World_AgentThere (Position) != 0) ||
	      (World_GetEnergy (Position) != 0))
	    flag = 1;
	}
      else
	flag = 3;
      Lauf++;
    }

  if (flag == 0)
    return TRUE;
  else
    return FALSE;

}
/*-end-ftn-*/

void 
setEnergy (AgentIdType AgentId, EnergyType MaxEnergy, EnergyType Energy)
{
  if (Energy > MaxEnergy)
    {
      AgentSet_SetIdle (AgentId, (CountType) floor (IdleFactor * (Energy - MaxEnergy)));
      Energy = MaxEnergy;
    }
  AgentSet_SetEnergy (AgentId, Energy);
}

/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Attack
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/


bool
FuncSem_Attack (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  int flag = 0;
  PositionType Position;
  PositionType AttackField;
  OrientationType Orientation;
  AgentIdType DefenderID, DeadID;
  EnergyType AgentEnergy;
  EnergyType DefenderEnergy;
  CountType AgentStrength;
  CountType DefenderStrength;
  AgentIdType AgentType, DefenderType;
  EnergyType AgentMinEnergy, DefenderMinEnergy;
  EnergyType DiffEnergy;


  Position = AgentSet_GetPosition (AgentID);
  Orientation = AgentSet_GetOrientation (AgentID);
  AttackField = GetHerePosition (Position, Orientation);

  if (IsPositionValid (AttackField))
    {
      DefenderID = World_AgentThere (AttackField);
/*              if (DefenderID != 0) */
      if ((DefenderID != 0) && (AgentSet_CanAttack (AgentID, DefenderID)))
	{
	  AgentEnergy = AgentSet_GetEnergy (AgentID);
	  DefenderEnergy = AgentSet_GetEnergy (DefenderID);
	  AgentStrength = AgentSet_GetStrength (AgentID);
	  DefenderStrength = AgentSet_GetStrength (DefenderID);

	  if (DefenderStrength == 0)
	    DefenderStrength = 1;
	  if (AgentStrength == 0)
	    AgentStrength = 1;

	  if (FastAttack == 0)
	    {
	      setEnergy (DefenderID, AgentSet_GetMaxEnergy (DefenderID),
		       DefenderEnergy - (int) ceil ((double) AgentStrength /
						 (double) DefenderStrength *
			      (double) AgentSet_GetAttackFactor (AgentID)));
	      setEnergy (AgentID, AgentSet_GetMaxEnergy (AgentID),
			 AgentEnergy - (int)
			 ceil ((double) DefenderStrength /
			       (double) AgentStrength *
			  (double) AgentSet_GetDefenseFactor (DefenderID)));
	      AgentSet_SetAttacks (DefenderID, AgentSet_GetAttacks (DefenderID) + 1);

	      flag = 1;
	    }
	  else
	    {
	      AgentType = AgentSet_GetType (AgentID);
	      DefenderType = AgentSet_GetType (DefenderID);
	      /* 2. Argument fr AgentSet_CanEat() mu ein _Typ_ sein (keine ID)! */
	      AgentMinEnergy = (int) ceil ((double) AgentSet_GetMaxEnergy (AgentID) * (double) AgentSet_GetMinEnergyFactor (AgentID));
	      DefenderMinEnergy = (int) ceil ((double) AgentSet_GetMaxEnergy (DefenderID) * (double) AgentSet_GetMinEnergyFactor (DefenderID));
	      if ((AgentStrength > DefenderStrength) && AgentSet_CanEat (AgentID, DefenderType) && (DefenderMinEnergy <= DefenderEnergy))
	        {
		  DiffEnergy = DefenderEnergy - DefenderMinEnergy + 1;
		  if (DefenderKeepEnergy > 0 && DiffEnergy > AgentSet_GetMaxEnergy (AgentID) - AgentEnergy)
		    {
		      DiffEnergy = AgentSet_GetMaxEnergy (AgentID) - AgentEnergy;
		      if (DefenderKeepEnergy == 1)
		        AgentSet_SetAttacks (DefenderID, AgentSet_GetAttacks (DefenderID) + 1);
		      else 
		        AgentSet_SetAttacks (DefenderID, -1);
		    }
		     
		  setEnergy (AgentID,
			     AgentSet_GetMaxEnergy (AgentID),
			     AgentEnergy +
			     DiffEnergy 
		    ); /* + 1 wegen setEnergy DefMinEn - 1 */
		  setEnergy (DefenderID, AgentSet_GetMaxEnergy (DefenderID), DefenderEnergy - DiffEnergy);
		  DeadID = DefenderID;
		  flag = 1;
		}

	      if (AgentStrength == DefenderStrength)
		flag = 0;

	      if ((AgentStrength < DefenderStrength) && AgentSet_CanEat (DefenderID, AgentType) && (AgentMinEnergy <= AgentEnergy))
		{
		  if (DefenderKeepEnergy)
		    Error ("Attack: DefenderKeepEnergy: Not implemented for this case!\n");
		  setEnergy (DefenderID,
			     AgentSet_GetMaxEnergy (DefenderID),
			     DefenderEnergy +
			     AgentEnergy - AgentMinEnergy + 1
		    ); /* s.o. */
		  setEnergy (AgentID, AgentSet_GetMaxEnergy (AgentID), AgentMinEnergy - 1
		    );
		    DeadID = AgentID;
		  flag = 1;
		}

	    }
	}
    }

  if (flag == 1)
  /*{
  World_AgentFoodConvert (DeadID, AgentSet_GetEnergy (DeadID), AgentSet_GetPosition (DeadID));
    AgentSet_DestroyAgent (DeadID);*/ 
    return TRUE;
   /*} */
  else
    {
      AgentEnergy = AgentSet_GetEnergy (AgentID) - AttackFailureEnergy;
      AgentSet_SetEnergy (AgentID, (AgentEnergy < 0) ? 0 : AgentEnergy);
      return FALSE;
    }

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Mate
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/


bool
FuncSem_Mate (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  int Lauf, EnergyAgent, EnergyMate, EnergyAvailable;
  bool flag = FALSE;
  PositionType Position;
  PositionType OffspringPosition;
  PositionType MatePosition;
  OrientationType Orientation;
  AgentIdType MateID;
  AgentIdType OffspringID;
  AgentIdType TempID;
  AgentIdType InheritID;
  AgentIdType CrossoverID;
  NodePtrType CrossoverDest;
  NodePtrType CrossoverSource;
  NodePtrType OffspringNodePtr;
  NodePtrType CMuteNode;
  NodePtrType CMuteDest;
  unsigned child;
  CountType NumChildren;
  AgentIdType AgentType, MateType, OffspringType;
  bool mute = 0;

  if ((RANDOM (101) / 100.0) > MateProb) 
    return FALSE;
  
  Position = AgentSet_GetPosition (AgentID);
  Orientation = AgentSet_GetOrientation (AgentID);
  MatePosition = GetHerePosition (Position, Orientation);

  if (!IsPositionValid (MatePosition))
    return FALSE;
  else
    {
      MateID = World_AgentThere (MatePosition);
/*      if (MateID == 0) */

      if ((MateID == 0) || (!AgentSet_CanMate (AgentID, MateID))) 
	return FALSE;
      else
	{
	  if ((AgentSet_GetAge (AgentID) <= AgentSet_GetMinMatingAge (AgentID)) ||
	  (AgentSet_GetAge (MateID) <= AgentSet_GetMinMatingAge (MateID)) ||
	      (AgentSet_GetLastMate (AgentID) != 0))  
	    return FALSE;
	  else
	    {
	      EnergyAgent = AgentSet_GetEnergy (AgentID);
	      EnergyMate = AgentSet_GetEnergy (MateID);
	      EnergyAvailable = ceil (0.5 * EnergyAgent + 0.5 * EnergyMate);

	      if (CheapMating == 0)
		{
		  AgentSet_SetEnergy (AgentID, ceil (0.5 * EnergyAgent));
		  AgentSet_SetEnergy (MateID, ceil (0.5 * EnergyMate));
		}

	      AgentSet_SetLastMate (AgentID, AgentSet_GetMatePause (AgentID));
	      NumChildren = AgentSet_GetNumberOfChildren (AgentID);
	      for (Lauf = 1; Lauf <= NumChildren; Lauf++)
		{
		  OffspringPosition = GetOffspringPosition (Position, &flag);
		  if (flag == TRUE)
		    {
		      DebugInfo ("MATING ----------------------- \n");


		      if (RANDOM (2) == 0)
			{
			  TempID = AgentID;

			  DebugInfo ("      Child inherits attributes from agent %d (father)\n", AgentID);
			}
		      else
			{
			  TempID = MateID;

			  DebugInfo ("      Child inherits attributes from agent %d (mother)\n", MateID);
			}

		      OffspringID = AgentSet_CreateAgentOfType (0,
						 AgentSet_GetType (TempID));

		      DebugInfo ("      Creating agent %d as child of agents %d (father) and %d (mother)\n",
				 OffspringID, AgentID, MateID);
		      DebugInfo ("      Childs position will be %d %d\n",
				 OffspringPosition.x, OffspringPosition.y);
		         AgentSet_SetMaxFriends(OffspringID,
		         AgentSet_GetMaxFriends(TempID));
		         AgentSet_SetStrength(OffspringID,
		         AgentSet_GetStrength(TempID));
		         AgentSet_SetMaxEnergy(OffspringID,
		         AgentSet_GetMaxEnergy(TempID));
		         AgentSet_SetVisionRange(OffspringID,
		         AgentSet_GetVisionRange(TempID));
		         AgentSet_SetHearingRange(OffspringID,
		         AgentSet_GetHearingRange(TempID));
		      
		      /* AgentSet_InitAgent(OffspringID); */

		      AgentSet_SetAge (OffspringID, 0);

		      if (CheapMating == 1)
			AgentSet_SetEnergy (OffspringID, AgentSet_GetMaxEnergy (OffspringID));
		      else
			AgentSet_SetEnergy (OffspringID, ceil (EnergyAvailable / NumChildren));

		      AgentSet_SetAttacks (OffspringID, 0);
		      AgentSet_SetScreams (OffspringID, 0);
		      AgentSet_AddToFriends (OffspringID, AgentID);
		      AgentSet_AddToFriends (OffspringID, MateID);
		      AgentSet_SetPosition (OffspringID, OffspringPosition);
		      AgentSet_SetOrientation (OffspringID, RANDOM (8) * 45);
		      World_SetAgent (OffspringID, OffspringPosition);

		      if (RANDOM (2) == 0)
			{
			  InheritID = AgentID;
			  CrossoverID = MateID;
			}
		      else
			{
			  InheritID = MateID;
			  CrossoverID = AgentID;
			}

		      OffspringNodePtr =
			AgentPrgFunc_CopyTree (AgentPrgFunc_Program2NodePtr
					 (AgentSet_GetProgram (InheritID)));

		      /* die Zielknotennummer darf nicht 1 sein, weil replace dann
		         nicht funktioniert. ist aber vielleicht auch nicht so schlimm, 
		         dass nicht der ganze Baum kopiert werden darf. erik */

		      CrossoverDest = AgentPrgFunc_GetNode (OffspringNodePtr,
							    RANDOM (AgentSet_GetProgramLength (InheritID) - 1) + 2);
		
		      CrossoverSource = AgentPrgFunc_GetNode (AgentPrgFunc_Program2NodePtr
					(AgentSet_GetProgram (CrossoverID)),
					   RANDOM (AgentSet_GetProgramLength
						   (CrossoverID)) + 1);

		      AgentPrgFunc_Replace (OffspringNodePtr, CrossoverDest, CrossoverSource);


		      if (CMute && (Random_DRand () < CMutation))
		      {
		      	CMuteNode = Interpreter_CreateRandomNode (0, MaxDepth - MinDepth );
		      	CMuteDest = AgentPrgFunc_GetNode (OffspringNodePtr, 
		      						RANDOM (AgentPrgFunc_NodeCount (OffspringNodePtr) - 1) + 2);
		      	AgentPrgFunc_Replace (OffspringNodePtr, CMuteDest, CMuteNode);
		      	for (child = 0; child < 3; child++)
		      		AgentPrgFunc_Del (CMuteNode, child);
		        free (CMuteNode);
			if (ActionDumpBuffer != NULL)
			  {
			    DynaString_Append (ActionDumpBuffer, "\t\t#! CMutation Agent: ");
			    DynaString_AppendInt (ActionDumpBuffer, OffspringID);
			    DynaString_Append (ActionDumpBuffer, "\n");
			  }
		      } 
		      
		      if (Mute && (Random_DRand () < Mutation))
			{
			  mute = 1;
			  if (ActionDumpBuffer != NULL)
			    {
			      DynaString_Append (ActionDumpBuffer, "\t\t#! Mutation Agent: ");
			      DynaString_AppendInt (ActionDumpBuffer, OffspringID);
			      DynaString_Append (ActionDumpBuffer, "\n");
			    }
			}

		
		      AgentPrgFunc_Prune (OffspringNodePtr, MaxDepth, mute);

		      AgentSet_SetProgram (OffspringID,
					   AgentPrgFunc_NodePtr2Program
					   (OffspringNodePtr,
				AgentPrgFunc_NodeCount (OffspringNodePtr)));

#ifdef DEBUG
/*                    DebugInfo("      Resulting program:\n      ");
   Interpreter_PrintNodes(stderr, 
   AgentPrgFunc_Program2NodePtr(AgentSet_GetProgram(OffspringID)));
   DebugInfo("END of MATING ---------------------\n\n"); */
#endif
		      if (Dump_WantDumpBirthAndDeath ())
			{
			  DynaStringType *text;
			  AgentType = AgentSet_GetType (AgentID);
			  MateType = AgentSet_GetType (MateID);
			  OffspringType = AgentSet_GetType (OffspringID);

			  text = DynaString_Alloc (400);
			  DynaString_AppendFormat (text, 50, "#! Birth = %06d %c {%06d %c, %06d %c} ",
						   OffspringID, OffspringType, AgentID, AgentType, MateID, MateType);
/*                              Interpreter_UnparseProgramInto(
   AgentSet_GetProgram(OffspringID), text); */
			  DynaString_Append (text, "\n");
			  Dump_DumpBirthAndDeath (DynaString_CString (text));
			  DynaString_Free (text);
			}
		    }
		  else if (Lauf > 1)
		    return TRUE;
		  else
		    return FALSE;
		}
	      return TRUE;
	    }
	}
    }
}


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Eat
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/


bool
FuncSem_Eat (AgentIdType AgentID, NodePtrArrayType Args)
{
  EnergyType Energy;
  EnergyType MaxEnergy;
  EnergyType Food;
  OrientationType Orientation;
  PositionType Position;
  FoodType Foodtype;


  Position = AgentSet_GetPosition (AgentID);
  Orientation = AgentSet_GetOrientation (AgentID);
  Position = GetHerePosition (Position, Orientation);
  if (IsPositionValid (Position) == FALSE)
    return FALSE;
  Food = World_GetEnergy (Position);
  Foodtype = World_GetFoodType (Position);

  if ((Foodtype != NONE) && (Food > 0) && ((Foodtype == PLANT && AgentSet_CanEat (AgentID, PLANT)) || (Foodtype == CARRION && AgentSet_CanEat (AgentID, World_GetCarrionType (Position)))))
    /* Ein Paar konsistenzpruefungen anbei ... */
    {
      DebugLevel (2, "\nType %c has eaten %c\n", (int) AgentSet_GetType (AgentID), Foodtype);
      MaxEnergy = AgentSet_GetMaxEnergy (AgentID);
      Energy = AgentSet_GetEnergy (AgentID) + Food;
      setEnergy (AgentID, MaxEnergy, Energy);
      World_FoodRemove (Position, Energy - MaxEnergy);
      /* Behandelt World_FoodRemove schon korrekt */
      return TRUE;
    }
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_MakeFriend
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_MakeFriend (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  int flag = 0;
  PositionType Position;
  AgentIdType FriendID;


  Position = GetHerePosition (AgentSet_GetPosition (AgentID),
			      AgentSet_GetOrientation (AgentID));
  if (IsPositionValid (Position))
    {
      FriendID = World_AgentThere (Position);
      if (FriendID != 0 && AgentSet_CanBeFriend (AgentID, FriendID))
	{
	  AgentSet_MakeFriend (AgentID, FriendID);	/* this is enough */
	  flag = 1;
	}
/*                      if (!AgentSet_IsFriendOf(AgentID, FriendID))
   {
   AgentSet_AddToFriends(AgentID, FriendID);
   flag = 1;
   } */
    }

  if (flag == 1)
    return TRUE;
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Scream
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_Scream (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{

  AgentSet_SetScreams (AgentID, AgentSet_GetScreams (AgentID) + 1);
  return TRUE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Move
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_Move (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  PositionType OldPosition;
  PositionType NewPosition;


  OldPosition = AgentSet_GetPosition (AgentID);
  NewPosition = GetHerePosition (OldPosition,
				 AgentSet_GetOrientation (AgentID));
  if ((IsPositionValid (NewPosition)) &&
      (World_AgentThere (NewPosition) == 0) &&
      (World_GetEnergy (NewPosition) == 0))
    {
      AgentSet_SetPosition (AgentID, NewPosition);
      World_AgentMove (AgentID, OldPosition, NewPosition);
      return TRUE;
    }
  else
    return FALSE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_TurnLeft
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_TurnLeft (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{

  OrientationType Orientation;

  Orientation = AgentSet_GetOrientation (AgentID);
  if (Orientation == 0)
    AgentSet_SetOrientation (AgentID, 315);
  else
    AgentSet_SetOrientation (AgentID, Orientation - 45);
  return TRUE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_TurnRight
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_TurnRight (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{

  OrientationType Orientation;

  Orientation = AgentSet_GetOrientation (AgentID);
  if (Orientation == 315)
    AgentSet_SetOrientation (AgentID, 0);
  else
    AgentSet_SetOrientation (AgentID, Orientation + 45);
  return TRUE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_TurnSound
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool 
FuncSem_TurnSound (AgentIdType AgentID, NodePtrArrayType Args)
{
  IterStateType FriendIter;

  struct Listenstruktur
    {
      PositionType Position;
      struct Listenstruktur *Nachfolger;
    };

  struct Listenstruktur *Aktuell;
  struct Listenstruktur *Temp;
  struct Listenstruktur *Liste;

  int ListCount = 0;
  int i, DiffX, DiffY, Anzahl, AgentPos;
  double Winkel;
  PositionType PositionAgent, PositionFriend;
  CountType Range;
  AgentIdType FriendID;
  bool InRange;

  Aktuell = Liste = NULL;
  Range = AgentSet_GetHearingRange (AgentID);
  PositionAgent = AgentSet_GetPosition (AgentID);
  FriendIter = AgentSet_InitFriendIteration (AgentID);

  while ((FriendID = AgentSet_GetNextFriend (AgentID, &FriendIter)) != 0)
    {
      InRange = FALSE;
      PositionFriend = AgentSet_GetPosition (FriendID);
      DiffX = abs (PositionAgent.x - PositionFriend.x);
      DiffY = abs (PositionAgent.y - PositionFriend.y);

      switch (Topology)
	{
	case RECTANGLE:
	  if ((DiffX <= Range) && (DiffY <= Range))
	    InRange = TRUE;
	  break;
	case CYLINDER:
	  if (((DiffX <= Range) && (DiffY <= Range)) ||
	      (((WorldSizeX - DiffX) <= Range) && (DiffY <= Range)))
	    {
	      InRange = TRUE;
	    }
	  break;

	case TORUS:
	  if (((DiffX <= Range) || ((WorldSizeX - DiffX) <= Range)) &&
	      ((DiffY <= Range) || ((WorldSizeY - DiffY) <= Range)))
	    {
	      InRange = TRUE;
	    }
	  break;
	}

      if (InRange)
	{
	  Anzahl = AgentSet_GetScreams (FriendID);
	  for (i = 0; i < Anzahl; i++)
	    {
	      Temp = Allocate (sizeof (struct Listenstruktur));
	      Temp->Position = PositionFriend;
	      Temp->Nachfolger = NULL;

	      if (!Liste)
		Liste = Aktuell = Temp;
	      else
		{
		  Aktuell->Nachfolger = Temp;
		  Aktuell = Temp;
		}

	      ListCount++;
	    }
	}
    }

  if (ListCount > 0)
    {
      AgentPos = RANDOM (ListCount) + 1;
      Aktuell = Liste;
      for (i = 1; i < AgentPos; i++)
	Aktuell = Aktuell->Nachfolger;

      DiffX = PositionAgent.x - Aktuell->Position.x;
      DiffY = PositionAgent.y - Aktuell->Position.y;
      if (abs (DiffX) > Range)
	DiffX -= sign (DiffX) * WorldSizeX;
      if (abs (DiffY) > Range)
	DiffY -= sign (DiffY) * WorldSizeY;

      Winkel = ceil (
		      acos (DiffY /
			  sqrt ((double) (DiffX * DiffX + DiffY * DiffY))) /
		      (2 * 3.14159) * 360);

      if (DiffX > 0)
	Winkel = 360 - Winkel;

      if ((Winkel - floor (Winkel / 45) * 45) > 22)
	Winkel = (floor (Winkel / 45) + 1) * 45;
      else
	Winkel = floor (Winkel / 45) * 45;
      AgentSet_SetOrientation (AgentID, (OrientationType) Winkel);

      while (Liste)
	{
	  Aktuell = Liste->Nachfolger;
	  Free (Liste);
	  Liste = Aktuell;
	}
      return TRUE;
    }
  else
    return FALSE;
}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_SetMem
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_SetMem (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  MemoryType OldMem;


  OldMem = AgentSet_GetMemory (AgentID);
  AgentSet_SetMemory (AgentID, Interpreter_Eval (AgentID, Args[0]));
  return (bool) OldMem;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Nop
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_Nop (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{

  return TRUE;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_If
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_If (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  bool result;

  result = Interpreter_Eval (AgentID, Args[0]);
  if (result == TRUE)
    result = Interpreter_Eval (AgentID, Args[1]);
  else
    result = Interpreter_Eval (AgentID, Args[2]);
  return result;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Seq
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_Seq (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  bool result;

  result = Interpreter_Eval (AgentID, Args[0]);
  result = Interpreter_Eval (AgentID, Args[1]);
  result = Interpreter_Eval (AgentID, Args[2]);
  return result;

}
/*-end-ftn-*/


/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_And
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_And (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  bool result;

  result = Interpreter_Eval (AgentID, Args[0]) &&
    Interpreter_Eval (AgentID, Args[1]);
  return result;


}
/*-end-ftn-*/

/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Or
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/
bool
FuncSem_Or (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  bool result;

  result = Interpreter_Eval (AgentID, Args[0]) ||
    Interpreter_Eval (AgentID, Args[1]);
  return result;


}
/*-end-ftn-*/

/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_Not
 *
 * Beschreib.:
 *
 * Parameter : (i)Args : NodePtrArrayType
 *
 * Return    : bool
 *
 *************************************************************************/



bool
FuncSem_Not (AgentIdType AgentID, NodePtrArrayType Args)

/*--*/
{
  bool result;

  result = !(Interpreter_Eval (AgentID, Args[0]));
  return result;


}
/*-end-ftn-*/



/*-ftn-********************************************************************
 *
 * Funktion  : FuncSem_SetParameterValue
 *
 * Beschreib.:
 *
 * Parameter :
 *
 * Return    : 
 *
 *************************************************************************/

bool 
FuncSem_SetParameterValue (string Key, string Value)
	/*
	   Diese Funktion speichert globale Parameter aus dem Worldfile,
	   die hier benoetigt werden lokal, um auf diese zugreifen zu
	   koennen.
	 */
{
  char *p;


/*      if(strcmp(Key,"WorldSize") == 0)
   {
   p = strtok(Value, " ");
   WorldSizeX = atoi(p);
   p = strtok(NULL, " ");
   WorldSizeY = atoi(p);
   } */
  if (strcmp (Key, "WorldWidth") == 0)
    {
      p = strtok (Value, " ");
      WorldSizeX = atoi (p);
      return (TRUE);
    }
  if (strcmp (Key, "WorldHeight") == 0)
    {
      p = strtok (Value, " ");
      WorldSizeY = atoi (p);
      return (TRUE);
    }
  if (strcmp (Key, "IdleFactor") == 0)
    {
      IdleFactor = atof (Value);
      return (TRUE);
    }
  if (strcmp (Key, "AttackFailureEnergy") == 0)
    {
      AttackFailureEnergy = atoi (Value);
      return (TRUE);
    }
  if (strcmp (Key, "Topology") == 0)
    {
      Topology = atoi (Value);
      return (TRUE);
    }
  if (strcmp (Key, "OffspringRadius") == 0)
    {
      OffspringRadius = atoi (Value);
      return (TRUE);
    }
  if (strcmp (Key, "TurnType") == 0)
    {
/*              TurnType = atoi(Value); */
      return (TRUE);
    }
  if (strcmp (Key, "MateProb") == 0)
    {
      MateProb = atof (Value);
      return (TRUE);
    }
  if (strcmp (Key, "CheapMating") == 0)
    {
      CheapMating = atoi (Value);
      return (TRUE);
    }
  if (strcmp (Key, "FastAttack") == 0)
    {
      FastAttack = atoi (Value);
      return (TRUE);
    }
  if (strcmp (Key, "DefenderKeepEnergy") == 0)
    {
      DefenderKeepEnergy = atoi (Value);
      return (TRUE);
    }
  if (strcmp (Key, "MaxProgramDepth") == 0)
    {
      MaxDepth = atoi (Value);
      return (TRUE);
    }
  if (strcmp (Key, "MinProgramDepth") == 0)
    {
      MinDepth = atoi (Value);
      return TRUE;
    }
  if (strcmp (Key, "Mutation") == 0)
    {
      Mutation = atof (Value);
      Mute = 1;
      return (TRUE);
    }
  if (strcmp (Key, "CMutation") == 0)
    {
      CMutation = atof (Value);
      CMute = 1;
      return TRUE;
    }



  return (FALSE);
}
