
/*-module-*****************************************************************
 *
 * Modul     : Interpreter (FUNKTIONSMODUL)
 *
 * Beschreib.: Agentenprogramm Lexer-Submodul
 *
 * Autor     : erik
 *
 * Datum     : 26.1.1996
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/
 
%{

#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"

#include "InterpreterPrivate.h"
#include "InterpreterParser.h"

void Interpreter_FinishReading (FILE *File, YY_BUFFER_STATE Buf);

int	Interpreter_LineNo;
/*
#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) \
	{ \
	int	c = getc(yyin); \
	result = (c==EOF) ? YY_NULL : (buf[0] = c, 1); \
	}
*/
%}

/*-ftn-********************************************************************
 *
 * Agenten Programm Tokenizer
 * 
 * In den Agentprogrammen kommen drei verschiedene Tokentypen vor,
 * Funktionsnamen, die oeffnende und die schliessende Klammer. Klammern
 * werde ohne speziell definiert zu sein als Einzeichen-Token direkt an 
 * die Grammatik uebergeben. Funktionsnamen werden als FUNC Token ueber-
 * geben. Ihr Wert ist die dem Namen entsprechende Funktionsnummer.
 *
 * Ausserdem existiert noch die Endmarkierung 'END_PROGRAM.'
 *
 * #fertig
 *
 *************************************************************************/

%%

[\t ]

[\n]		  { Interpreter_LineNo++; }

END_PROGRAM	  { Interpreter_FinishReading(yyin, YY_CURRENT_BUFFER);
				YY_FLUSH_BUFFER;
				return EO_PROGRAM;}

[A-Za-z?_]+	  { Interpreter_lval.yString = strdup(yytext);
				return FUNC; 
			  }

.		  	  { return yytext[0]; 
			  }

%%

/*-ftn-********************************************************************
 *
 * Interpreter_wrap (PRIVAT)
 *
 * Diese Funktion wird aufgerufen, wenn der Lexer das Dateiende erreicht.
 * Es wird immer 1 zurueckgeben, was signalisiert, dass die Eingabe 
 * wirklich zuende ist. 
 *
 * #fertig
 *
 *************************************************************************/

int Interpreter_wrap()
	{
	return 1;
	}



/*-ftn-********************************************************************
 *
 * Interpreter_FinishReading (PRIVAT)
 *
 * Schreibt die Zeichen, die schon gelesen aber nicht gematcht wurden
 * zurueck in die Eingabedatei.
 *
 * #fertig
 *
 *************************************************************************/

void Interpreter_FinishReading (FILE *File, YY_BUFFER_STATE Buf)
	{
	int  length; 
	char *begin;

	length = strlen (Buf->yy_ch_buf);
	begin = &(Buf->yy_ch_buf[length+1]);
	length = strlen (begin); 
	fseek (File, -length, 1);
	}

