%option lex-compat nostdinit
%Start  COM ACT

%{

#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"
#include "../agents/AgentSet.h"
#include "../dump/Dump.h"
#include "../world/World.h"
#include "LoaderParser.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef FALSE
#define FALSE (1 == 0)
#define TRUE  (1 == 1)
#endif

int LOADERparse();

int Standard;

typedef FILE* FilePointer;

void L_ReturnBufferToFile(FILE *File, YY_BUFFER_STATE Buf);

%}

string		[a-zA-Z0-9_.-]+\??
fliesskomma	[+-]?[0-9]*\.[0-9]*
ganzzahl	[+-]?[0-9]+

%%

<COM>.*"END_COMMENT"        BEGIN 0;
<ACT>.*"END_ACTIONS"        BEGIN 0;

<COM>.*                     |
<COM>\n                     ;

<ACT>.*                     |
<ACT>\n                     ;

"COMMENT"                   BEGIN COM;
"ACTIONS"                   BEGIN ACT;


\#.*\n           | 
\n		 |
\t               /* */
		 ;

[=]              {           
                   return(L_EQUAL);
                 }

"PARAMETERS"     {           
                   return(L_PARAMETERS);
                 }

"END_PARAMETERS" {           
                   return(L_END_PARAMETERS);
                 }

"HEADER"		{
					yyless(0);
					L_ReturnBufferToFile(LOADERin,YY_CURRENT_BUFFER);
/*					Dump_ParseHeader(LOADERin); */
					Dump_ParseSection(LOADERin);
					YY_FLUSH_BUFFER; 
					return(L_HEADER);
				}

"DUMP"			{
					yyless(0);
					L_ReturnBufferToFile(LOADERin,YY_CURRENT_BUFFER);
/*					Dump_ParseDump(LOADERin); */
					Dump_ParseSection(LOADERin);
					YY_FLUSH_BUFFER;
					return(L_DUMP);
				}

"AGENTS"		{
					yyless(0);
					L_ReturnBufferToFile   (LOADERin, YY_CURRENT_BUFFER); 
					AgentSet_ParseSection(LOADERin); 
					YY_FLUSH_BUFFER;
					return(L_AGENTS); 
				}

"MAP"			{
					yyless(0);
					L_ReturnBufferToFile  (LOADERin, YY_CURRENT_BUFFER);
					World_ParseSection(LOADERin);YY_FLUSH_BUFFER;
					YY_FLUSH_BUFFER;
					return(L_MAP);
				}

{string}		{	
					strncpy(LOADERlval.typstr, yytext, 255);
					LOADERlval.typstr[255]='\0';          
					return (L_STR);
				}

[\n\t]			|
.				;
%%

void L_ReturnBufferToFile(FILE *File, YY_BUFFER_STATE Buf)
{
  int  length; 
  char *begin;

  length = strlen (Buf->yy_ch_buf);

  begin = &(Buf->yy_ch_buf[length+1]);
  
  length = strlen (begin); 

  fseek (File, -(length + (yy_hold_char == EOF ? 0 : 1)), 1);
}

bool Loader_ParseSection(FILE* stream)
{
	if (LOADERin)
	{
		yyrestart(stream);
		YY_FLUSH_BUFFER;
	}
	else	
		LOADERin = stream;

	return LOADERparse() == 0;
}

bool Loader_LoadFile(const string Filename)
{
	FILE*	stream;
	bool	flag;

	if (Filename)
	{
		if (strcmp(Filename, "-") == 0)
		{
			stream = stdin;
		}
		else if (!(stream = fopen(Filename, "r")))
		{
			Warning("[Loader]Loader_LoadFile:", "couldn't open %s\n", 
					Filename);
			return FALSE;
		}
	}
	else
		return FALSE;

	Message("[Loader]Loader_LoadFile: parsing %s\n", Filename);

	flag = Loader_ParseSection(stream);

	if (stream != stdin)
		fclose(stream);

	return flag;
}


int LOADERwrap ()
{
  return (1);
}
