%{

#include "Loader.h"

#include <stdio.h>
#include <stdlib.h>
#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"

#include "../dump/Dump.h"
#include "../control/Control.h"
#include "../interpreter/Interpreter.h"
#include "../interpreter/FuncSem.h"
#include "../world/World.h"
#include "../agents/AgentSet.h"

#include <sys/param.h>

#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif

typedef FILE* FilePointer;

extern int   LOADERlineno;
extern FILE* LOADERin;

int LOADERlex();

static int LOADERerror(const string);

%}

%union {
	float	typflt;
	int		typint;
	char	typstr[256];
}

%token <typstr>		L_STR 
%token <typflt>		L_FLT
%token <typint>		L_INT
%token        		L_HEADER
%token				L_DUMP
%token         		L_PARAMETERS
%token         		L_END_PARAMETERS
%token         		L_AGENTS
%token         		L_MAP
%token         		L_EQUAL

%type   <typstr> value

%%


worldfile		:	header
					dump
					parameter
					agent
					map
				;

header			:	/*	*/
				|	L_HEADER
					{
						Message("Section HEADER parsed\n");
					}
				;

dump			:	/*	*/
				|	L_DUMP
					{
						Message("Section DUMP parsed\n");
					}
				;
				
parameter		:	/*	*/
				|	L_PARAMETERS
					parameters
					L_END_PARAMETERS
					{
						Message("Section PARAMETERS parsed\n");
					}
				;

agent			:	/*	*/
				|	L_AGENTS
					{
						Message("Section AGENTS parsed\n");
					}
				;

map				:	/*	*/
				|	L_MAP
					{
						Message("Section MAP parsed\n");
					}
				;

parameters		:	/* */
				|	parameters
					L_STR L_EQUAL value 
					{
						Loader_SetParameterValue($2, $4);
					}                        
				;


value			:	L_STR {strcpy  ($$, $1);}
				|	L_INT {sprintf ($$, "%d", $1);}
				|	L_FLT {sprintf ($$, "%f", $1);}
				;

%%



static int LOADERerror(const string s)
{
	Error("LOADER[LoadFile]", "%s, line %d\n", s, LOADERlineno);
	return 0;
}



void Loader_SetParameterValue(const string Key, const string Value)
	{
	bool	wantedValue;
		
	Debug(DEBUG_MODULE_LOADER | DEBUG_ACTION_EXE, DEBUG_DETAIL_MEDIUM, 
			"\tSetting Parameter %s to %s\n", Key, Value);

	wantedValue = FALSE;

	wantedValue = AgentSet_SetParameterValue(Key,Value) || wantedValue;
	wantedValue = Dump_SetParameterValue(Key,Value) || wantedValue;
	wantedValue = Interpreter_SetParameterValue(Key,Value) || wantedValue;
	wantedValue = World_SetParameterValue(Key,Value) || wantedValue;
	wantedValue = FuncSem_SetParameterValue(Key,Value) || wantedValue;
	wantedValue = Control_SetParameterValue(Key,Value) || wantedValue;

	if(!wantedValue)
		Error("LOADER[LoadFile]", "unrecognised parameter: found '%s'\n", Key); 
	}
