
/*-module-*****************************************************************
 *
 * Modul     : DynaString (FUNKTIONSMODUL)
 *
 * Beschreib.: Funktionen zur effizienten Verwaltung von dynamisch
 *			   wachsenden Strings
 *
 * Autor     : erik
 *
 * Datum     : 5.2.1996
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "CDefines.h"
#include "TypeDefs.h"


typedef struct
  {
    string buffer;
    int pos;
    int size;
    int left;
  }
DynaStringType;


/*-ftn-********************************************************************
 *
 * DynaString_Alloc
 *
 * Initialisiert einen dynamischen String der Groesse InitialSize.
 * Wird spaeter mehr Platz benoetigt, wird die Puffergroesse jeweils
 * verdoppelt.
 *
 * #fertig
 *
 *************************************************************************/

DynaStringType *
DynaString_Alloc (int InitialSize)
{
  DynaStringType *strg;

  strg = malloc (sizeof (DynaStringType));
  strg->buffer = malloc (InitialSize);
  *(strg->buffer) = '\0';
  strg->left = strg->size = InitialSize;
  strg->pos = 0;

  return strg;
}



/*-ftn-********************************************************************
 *
 * DynaString_CStringAndFree
 *
 * Liefert einen normalen C-String (also ein Array von char durch '\0'
 * terminiert) und gibt den dynamischen String frei. Der C-String muss,
 * wenn er nicht mehr gebraucht wird, mit free freigegeben werden.
 *
 * #fertig
 *
 *************************************************************************/

string 
DynaString_CStringAndFree (DynaStringType * DString)
{
  string cstring;

  cstring = DString->buffer;
  free (DString);

  return cstring;
}



/*-ftn-********************************************************************
 *
 * DynaString_Free
 *
 * Gibt den dynamischen String frei.
 *
 * #fertig
 *
 *************************************************************************/

void 
DynaString_Free (DynaStringType * DString)
{
  free (DString->buffer);
  free (DString);
}



/*-ftn-********************************************************************
 *
 * DynaString_CString
 *
 * Liefert einen normalen C-String (also ein Array von char durch '\0'
 * terminiert) Der C-String darf nicht geaendert werden, braucht dafuer
 * aber auch nicht freigegeben werden.
 *
 * #fertig
 *
 *************************************************************************/

string 
DynaString_CString (DynaStringType * DString)
{
  return DString->buffer;
}



/*-ftn-********************************************************************
 *
 * DynaString_Length
 *
 * Liefert die Laenge des dynamischen Strings.
 *
 * #fertig
 *
 *************************************************************************/

int 
DynaString_Length (DynaStringType * DString)
{
  return DString->pos;
}



/*-ftn-********************************************************************
 *
 * DynaString_Append
 *
 * Haengt einen normalen C-String (s.o.) an den dynamischen String.
 *
 * #fertig
 *
 *************************************************************************/

void 
DynaString_Append (DynaStringType * DString, const string CString)
{
  char *p, *source;

  p = DString->buffer + DString->pos;
  source = (char *) CString;	/* gcc stellt sich etwas an */
  while (*source)
    {
      *p++ = *source++;
      DString->left--;

      if (DString->left == 0)
	{
	  DString->pos = (int) (p - DString->buffer);
	  DString->left = DString->size;
	  DString->size *= 2;
	  DString->buffer = realloc (DString->buffer, DString->size);
	  p = DString->buffer + DString->pos;
	}
    }
  *p = '\0';
  DString->pos = (int) (p - DString->buffer);
}


void 
DynaString_AppendFormat (DynaStringType * DString, int MaxLen,
			 const string fmt,...)
{
  va_list args;
  string buf;

  va_start (args, fmt);
  buf = malloc (MaxLen);
  vsprintf (buf, fmt, args);
  DynaString_Append (DString, buf);
  free (buf);
}


void 
DynaString_AppendInt (DynaStringType * DString, int Value)
{
  char b[256];

  sprintf (b, "%d", Value);
  DynaString_Append (DString, b);
}


void 
DynaString_AppendFloat (DynaStringType * DString, float Value)
{
  char b[256];

  sprintf (b, "%f", Value);
  DynaString_Append (DString, b);
}


void 
DynaString_Reset (DynaStringType * DString)
{
  DString->pos = 0;
  DString->left = DString->size;
  DString->buffer[0] = '\0';
}
