#include <stdio.h>
#include "Messages.h"
#include <varargs.h>

#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif


extern int Global_WantVerboseOutput;

int Global_WantDebugOutput = 0;
int Global_DebugLevel = 0;

int Global_DebugModules = 0;
int Global_DebugType = DEBUG_DETAIL_MEDIUM;

static string ModuleNames[] =
{
  "AgentSet", "Control", "Dictionary", "Dump", "Interpreter", "IO",
  "Loader", "Misc", "Program", "World", "Random"
};

static string StateNames[] =
{
  "Initialization", "Parse", "Scan", "Execution", "Dump"
};

int Global_IsInteractive = TRUE;

void 
DebugInfo (va_alist)
     va_dcl
{
  char *template;
  va_list arglist;

  if (Global_WantDebugOutput)
    {
      va_start (arglist);
      template = va_arg (arglist, char *);
      vfprintf (stderr, template, arglist);
      va_end (arglist);
    }
}


/* -------------------------------------------------------------------------- */
/*
   DebugLevel (int, template, p3, p4, ...)

   Diese Funktion dient der selektiven Debug-Ausgabe.
   Sie erwartet mindestens 2 Parameter. 

   Der erste ist vom Typ int und gibt den gewuenschten DebugLevel an, auf dem
   der folgende Text (Parameter 2 bis n) ausgegeben werden soll.

   Also, das geht so: Wird Failure mit der Option -D # aufgerufen, so ist der
   DebugLevel fuer diesen Lauf gleich #. Alle Aufrufe dieser Funktion mit einem
   ersten Parameter kleiner gleich # fuehren zur Ausgabe des DebugTextes. 

   Der DebugText wird wie bei der Funktion printf ausgegeben, template gibt 
   das Format an, die optionalen Parameter p3 bis pn enthalten evtl. die Werte
   fuer die im Formatstring template angegebenen Platzhalter.

 */
/* -------------------------------------------------------------------------- */

void 
DebugLevel (va_alist)
     va_dcl
{
  char *template;
  va_list arglist;
  int debug_level;

  if (Global_WantDebugOutput)
    {
      va_start (arglist);

      debug_level = va_arg (arglist, int);

      if (debug_level <= Global_DebugLevel)
	{
	  template = va_arg (arglist, char *);
	  vfprintf (stderr, template, arglist);
	}
      va_end (arglist);
    }
}

void 
Error (va_alist)
     va_dcl
{
  char *template;
  va_list arglist;

  va_start (arglist);

  fprintf (stderr, "ERROR[%s] -- ", va_arg (arglist, char *));
  template = va_arg (arglist, char *);
  vfprintf (stderr, template, arglist);

  va_end (arglist);

  exit (1);
}

void 
Warning (va_alist)
     va_dcl
{
  char *template;
  va_list arglist;

  va_start (arglist);

  fprintf (stderr, "WARNING[%s] -- ", va_arg (arglist, char *));
  template = va_arg (arglist, char *);
  vfprintf (stderr, template, arglist);

  va_end (arglist);
}


void 
Print (va_alist)
     va_dcl
{
  char *template;
  va_list arglist;

  va_start (arglist);
  template = va_arg (arglist, char *);
  vfprintf (stdout, template, arglist);
  va_end (arglist);
  fflush (stdout);
}

void 
Message (va_alist)
     va_dcl
{
  char *template;
  va_list arglist;

  if (Global_WantVerboseOutput)
    {
      va_start (arglist);
      template = va_arg (arglist, char *);
      vfprintf (stderr, template, arglist);
      va_end (arglist);
      fflush (stderr);
    }
}


static bool 
is_single_bit_set (unsigned bits)
{
  int firstBit = ffs (bits);

  if (firstBit != 0)
    {
      bits &= ~(1 << firstBit);
      return ffs (bits) == 0;
    }
  else
    return FALSE;
}

bool 
Debug (va_alist)
     va_dcl
{
  unsigned type, level;
  char *template;
  va_list arglist;

  va_start (arglist);
  type = va_arg (arglist, int);
  level = va_arg (arglist, int);

  if (Global_WantDebugOutput)
    {
      if ((type & Global_DebugModules) && (level <= Global_DebugType))
	{
	  if (is_single_bit_set (type & DEBUG_MODULES))
	    fprintf (stderr, "%s:", ModuleNames[type & DEBUG_MODULES]);
	  if (is_single_bit_set (type & DEBUG_ACTION))
	    fprintf (stderr, "[%s]", StateNames[type & DEBUG_ACTION]);

	  template = va_arg (arglist, char *);
	  vfprintf (stderr, template, arglist);

	  return TRUE;
	}
    }

  va_end (arglist);
  return FALSE;
}

#ifdef MESSAGE_TEST
int 
main (int argc, char *argv[])
{
  if (argc == 3)
    {
      Global_DebugLevel = atoi (argv[1]);
      Global_DebugType = atoi (argv[2]);
      return 0;
    }
  else
    return argc - 3;
}
#endif
