
#include "StringList.h"
#include "Memory.h"
#include <string.h>

static void 
StringList_Realloc (StringListType * StringList,
		    unsigned NumberOfStrings)
{
  if (NumberOfStrings == 0)
    NumberOfStrings = StringList->Allocated;

  StringList->Strings = Reallocate (StringList->Strings,
	       (StringList->Allocated + NumberOfStrings) * sizeof (string));
  bzero (StringList->Strings + StringList->Allocated,
	 NumberOfStrings * sizeof (string));
  StringList->Allocated += NumberOfStrings;
}

StringListType *
StringList_Alloc (unsigned NumberOfStrings)
{
  StringListType *newStringList;

  newStringList = Allocate (sizeof (StringListType));
  newStringList->Allocated = NumberOfStrings;
  newStringList->Used = 0;
  newStringList->Strings = Callocate (NumberOfStrings, sizeof (string));

  return newStringList;
}

void 
StringList_Destroy (StringListType * StringList, bool WithContents)
{
  StringList_Empty (StringList, WithContents);
  Free (StringList);
}

void 
StringList_Empty (StringListType * StringList, bool WithContents)
{
  if (WithContents)
    {
      int i;

      for (i = 0; i < StringList->Used; i++)
	Free (StringList->Strings[i]);
    }
}

unsigned 
StringList_Count (StringListType * StringList)
{
  return StringList->Used;
}

string 
StringList_StringAt (StringListType * StringList, int index)
{
  if (index < 0)
    index = StringList->Used + index - 1;

  if ((index >= 0) && (index < StringList->Used))
    return StringList->Strings[index];
  else
    return NULL;
}

int 
StringList_FindString (StringListType * StringList, const string String)
{
  return StringList_MatchString (StringList, String, strcmp);
}

int 
StringList_MatchString (StringListType * StringList, const string String,
			int (*comp) (const string, const string))
{
  int i;

  for (i = 0; i < StringList->Used; i++)
    {
      if ((*comp) (StringList->Strings[i], String) == 0)
	return i;
    }

  return -1;
}

unsigned 
StringList_AddString (StringListType * StringList, string String)
{
  return StringList_InsertStringAt (StringList, String, StringList->Used);
}

void 
StringList_DeleteStringAt (StringListType * StringList, int index)
{
  string s;

  if ((s = StringList_RemoveStringAt (StringList, index)))
    Free (s);
}

int 
StringList_InsertStringAt (StringListType * StringList,
			   string String, int index)
{
  if (StringList->Used == StringList->Allocated)
    StringList_Realloc (StringList, 0);

  if (index < 0)
    index = StringList->Used + index - 1;

  if ((index >= 0) && (index <= StringList->Used))
    {
      if ((index + 1) < StringList->Used)
	{
	  bcopy (StringList->Strings + index, StringList->Strings + index + 1,
		 (StringList->Used - (index + 1)) * sizeof (string));
	}

      StringList->Strings[index] = String;
      StringList->Used++;
      return index;
    }
  else
    return -1;
}

string 
StringList_RemoveStringAt (StringListType * StringList, int index)
{
  unsigned count;
  string s;

  count = StringList_Count (StringList);

  if (index < 0)
    index = count + index - 1;

  if ((s = StringList->Strings[index]))
    {
      if ((index + 1) < count)
	{
	  bcopy (StringList->Strings + index + 1, StringList->Strings + index,
		 (count - (index + 1)) * sizeof (string));
	}
      StringList->Used--;
    }

  return s;
}

string 
StringList_ReplaceWithString (StringListType * StringList,
			      string String, int index)
{
  unsigned count;
  string s;

  count = StringList_Count (StringList);

  if (index < 0)
    index = count + index - 1;

  if ((s = StringList_StringAt (StringList, index)))
    {
      StringList->Strings[index] = String;
      return s;
    }
  else
    return NULL;
}
