/*-module-*****************************************************************
 *
 * Modul     : AgentPrgFunc (FUNKTIONSMODUL)
 *
 * Beschreib.:  Dieses Modul enthaelt Funktionen mit Hilfe denen, das Baum von
 *              Agenten erzeugt werden kann.
 *              Die Knoten eines Baumes sind Structuren von "NodePtrType".NodePtrType
 *              besteht aus einem Befehl von "CommandType" und drei Zeiger auf Knoten.
 *
 *
 * Autor     : <autor>
 *
 * Datum     : <15.01.>
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/

/*-includes-*/

#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"
#include "../misc/Messages.h"
#include "../interpreter/Interpreter.h"

#include "AgentPrgFunc.h"

#include <stdlib.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
/*--*/


/*-defines-*/
#ifndef NULL
#define NULL 0
#endif
/*--*/




/*-local-vars-*/

char errorp[] = " Module  program output for function call with illegal null pointer";
char create_error[] = "Module Agentprogram function Create:memory allocation error ";
char insert_error[] = "Module Agentprogram function Insert : node1 is null pointer ";
char set_error[] = "Module Agentprogram function Set : node is null pointer";
char del_warning[] = "Module Agentprogram function del : Warning node is null pointer";
char get_child_error1[] = "Agentprogram function  getchild :position must be 0,1,2 ";
char get_child_error2[] = " Agentprogram function  getchild : null tree has no children returnig null";
char get_child_error3[] = " Agentprogram function  getchild : this node cannot hold so many children";
char get_command_error[] = " Agentprogram function getcommand :node is null pointer";
char get_node_error[] = "Agentprogram function getnode:count must be positive ";
char insert_error2[] = "Agentprogram function insert er2 : this node cannot hold so many children";
/*--*/


/*-ftn-********************************************************************
 *
 * Funktion  : freesubtree
 *
 * Beschreibung : Gibt den Speicherplatz ,der durch malloc Funktion erzeugt
 * wurde ,zurueck. (dealocation of memory)
 *           
 *
 *
 * Parameter : node : NodePtrType
 *
 * Return    : void 
 *
 *************************************************************************/

void 
freesubtree (node)
     NodeType *node;
{
  auto int ch;

  if (node == (NodeType *) NULL)
    return;
  for (ch = 0; ch < node->argc; ch++)
    {
      freesubtree (node->argv[ch]);
    }
  free (node);
  return;
}



/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_Create
 *
 * Beschreib.: Diese Funktion erzeugt einen leeren Baum .Liefert die Wurzel vom
 *            Baum zurueck.
 *
 *
 * Parameter :
 *
 * Return    : NodePtrType
 *
 *************************************************************************/
NodePtrType 
AgentPrgFunc_Create (maxarguments)
     int maxarguments;
/*--*/

{
  NodeType *pnode;
  register int node;

  pnode = (NodeType *) malloc (sizeof (NodeType) +
			       maxarguments * sizeof (NodePtrType));
  if (pnode == (NodeType *) NULL)
    {
      printf (" %s \n", create_error);
      exit (1);
    }
  pnode->command = (CommandType) NULL;
  pnode->argc = (byte) maxarguments;
  for (node = 0; node < maxarguments; node++)
    {
      pnode->argv[node] = (NodePtrType) NULL;
    }
  return pnode;
}
/*-end-ftn-*/



/* --------------------------------------------------------------------------

   Function: AgentPrgFunc_GetFatherOfNode

   This fuction traverses the tree whose root is RootNode and returns
   a pointer to the father node of node Node.

   It returns NULL if RootNode or Node is NULL or if the subtree doesn't
   contain Node. 

   !!! unused yet - but maybe it will become useful  (it did!)

   Author:   Thorsten Kohnhorst & erik

   ------------------------------------------------------------------------ */

NodePtrType 
AgentPrgFunc_GetParentOfNode (NodePtrType RootNode, NodePtrType Node)
{
  NodePtrType tmpnode = NULL;
  int i;

  if ((Node == NULL) || (RootNode == NULL))
    return NULL;

  if (Node == RootNode)
    {
      Error ("AgentPrgFunc_GetParentOfNode", "node in question is the root node.");
      exit (1);
    }

  for (i = 0; i < RootNode->argc; i++)
    if (RootNode->argv[i] == Node)
      return RootNode;

  for (i = 0; i < RootNode->argc; i++)
    {
      tmpnode = AgentPrgFunc_GetParentOfNode (RootNode->argv[i], Node);
      if (tmpnode != NULL)
	return tmpnode;
    }

  return NULL;
}


/* --------------------------------------------------------------------------

   Function: AgentPrgFunc_Replace

   This fuction replaces the subtree Node1 in the tree RootNode1 with the 
   subtree Node2. 

   Author:   Thorsten Kohnhorst & erik

   ------------------------------------------------------------------------ */

void 
AgentPrgFunc_Replace (NodePtrType RootNode1, NodePtrType Node1,
		      NodePtrType Node2)
{
  NodePtrType parentNode1;
  int i;

  parentNode1 = AgentPrgFunc_GetParentOfNode (RootNode1, Node1);
  if (parentNode1 == NULL)
    {
      Error ("AgentPrgFunc_Replace", "NULL node encountered");
      /* fprintf(stderr,"AgentPrgFunc_Replace : NULL node encountered"); */
      return;
    }

  for (i = 0; i < parentNode1->argc; i++)
    if (parentNode1->argv[i] == Node1)
      {
	freesubtree (parentNode1->argv[i]);
	parentNode1->argv[i] = AgentPrgFunc_CopyTree (Node2);
	break;
      }
  return;
}




void 
AgentPrgFunc_Prune (NodePtrType Node, int MaxDepth, bool withMutation)
{
  int i;

  if ((Node == NULL) || (MaxDepth < 1))
    return;

  if (MaxDepth > 1)
    {
      for (i = 0; i < Node->argc; i++)
	AgentPrgFunc_Prune (Node->argv[i], MaxDepth - 1, withMutation);
    }
  else if (Node->argc > 0)
    {
      Node->command = Interpreter_RandomCommand (withMutation ? IRFM_LEAF : 0);
      for (i = 0; i < Node->argc; i++)
	{
	  freesubtree (Node->argv[i]);
	  Node->argv[i] = NULL;
	}
      Node->argc = 0;
    }
/*      else    node is a terminal and doesn't need any changes */
}


/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_Insert
 *
 * Beschreib.: Der TeilBaum node2 wird in der entsprechende Position von node1
 *             eingefuegt.Der fruehere Teilbaum von node1 wird geloescht.
 * Achtung :Zur Zeit duerfen sie nicht ein node2 der zum Baum node1 gehoert
 * benutzen .Kein control dafuer wird durchgefuert. 
 * Parameter : (i)node1 : NodePtrType
 *             (i)pos : NodePositionType
 *             (i)node2 : NodePtrType
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
AgentPrgFunc_Insert (node1, pos, node2)
     NodePtrType node1, node2;
     NodePositionType pos;

/*--*/

{
  if (node1 == NULL)
    {
      printf (" %s \n", insert_error);
      exit (1);
    }
  if (pos < (int) node1->argc)
    {
      freesubtree (node1->argv[pos]);
      node1->argv[pos] = node2;
    }
}
/*-end-ftn-*/

/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_Set
 *
 * Beschreib.: Der Befehl com wird im Knoten node eingefuegt.
 *
 * Parameter : (i)node : NodePtrType
 *             (i)com : CommandType
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
AgentPrgFunc_Set (node, com)
     NodePtrType node;
     CommandType com;
/*--*/
{
  if (!node)
    {
      printf (" %s\n", set_error);
      return;
    }
  node->command = com;
}
/*-end-ftn-*/

/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_Del
 *
 * Beschreib.: Der Teilbaum in der Position pos wird geloescht.
 *
 *
 * Parameter : (i)node : NodePtrType
 *             (i)pos : NodePositionType
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
AgentPrgFunc_Del (node, pos)
     NodePtrType node;
     NodePositionType pos;

/*- uses the function freesubtree -*/
{
  if (!node)
    {
      printf (" %s \n", del_warning);
      return;
    }
  if (pos >= (int) node->argc)
    return;			/* ignore this case (erik) */
  freesubtree (node->argv[pos]);
  node->argv[pos] = (NodeType *) NULL;
  return;

}
/*-end-ftn-*/

#if 0
/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_Count
 *
 * Beschreib.: Liefert die  Numerierung des Knoten node2.
 *             Node1 ist die  Wurzel Teilbaumes zurueck.
 *             Nummer(root,left,middle,right)= root,Nummer(left),Nummer(middle),
 *             Nummer(right).
 *
 * Parameter : (i)node1 : NodePtrType
 *             (i)node2 : NodePtrType
 *
 * Return    : CountType
 *
 *************************************************************************/
CountType 
AgentPrgFunc_Count (node1, node2)
     NodePtrType node1, node2;

/*--*/
{
  static int count = 0, count_function_calls = 0;
  auto int tmpcount;
  char ch;
  static bool found = FALSE;
  count_function_calls++;
  if (node1 == (NodeType *) NULL)
    {
      count_function_calls--;
      return 0;			/* reduce depth */
    }
  if (node1 == node2)
    {
      found = TRUE;
      ++count;
      if (count_function_calls == 1)
	{
	  count = 0;
	  found = FALSE;
	  count_function_calls = 0;
	  return 1;
	}
      count_function_calls--;
      return 1;
    }
  ++count;
  for (ch = 0; ch < node1->argc; ch++)
    {
      if (found == FALSE)
	AgentPrgFunc_Count (node1->argv[ch], node2);
    }
  if (found == FALSE)
    {
      if (count_function_calls == 1)
	{
	  count_function_calls = 0;
	  return (-1);
	}
      else
	{
	  count_function_calls--;
	  return 1;
	}
    }
  else
    {
      if (count_function_calls == 1)
	{
	  found = FALSE;
	  count_function_calls = 0;
	  tmpcount = count;
	  count = 0;
	  return tmpcount;
	}
      else
	{
	  count_function_calls--;
	  return 1;
	}
    }

}
/*-end-ftn-*/
#endif

/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_GetNode
 *
 * Beschreib.: Liefert Zeiger ,der auf den count-ten Knoten des Teilbaumes mit Wurzel
 *             node zeigt,zurueck. (preorder)
 *
 * Parameter : (i)node : NodePtrType
 *             (i)count : CountType
 *
 * Return    : NodePtrType
 *
 *************************************************************************/

static NodePtrType 
GetNodeWithNumber (NodePtrType Node, CountType * countP)
{
  NodePtrType tNode;
  int i;

  if (Node == NULL)
    return NULL;

  if (*countP == 1)
    return Node;

  *countP = *countP - 1;
  tNode = NULL;
  for (i = 0; i < Node->argc; i++)
    {
      tNode = GetNodeWithNumber (Node->argv[i], countP);
      if (tNode != NULL)
	break;
    }

  return tNode;
}


NodePtrType 
AgentPrgFunc_GetNode (NodePtrType Node, CountType Count)
{
  if (Count <= 0)
    {
      fprintf (stderr, "%s \n", get_node_error);
      return (NodePtrType) NULL;
    }

  return GetNodeWithNumber (Node, &Count);
}


/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_NodeCount
 *
 * Beschreib.: Liefert die Anzahl von Knoten des Baumes node.
 *
 * Parameter : (i)node : NodePtrType
 *
 * Return    : CountType
 *
 *************************************************************************/

CountType 
AgentPrgFunc_NodeCount (NodePtrType Node)
{
  int i, count;

  count = 0;
  if (Node != NULL)
    {
      count++;
      for (i = 0; i < Node->argc; i++)
	count += AgentPrgFunc_NodeCount (Node->argv[i]);
    }
  return count;
}


/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_GetCommand
 *
 * Beschreib.: Liefert den Befehl des Knoten node zurueck.
 *
 * Parameter : (i)node : NodePtrType
 *
 * Return    : CommandType
 *
 *************************************************************************/

CommandType 
AgentPrgFunc_GetCommand (NodePtrType node)
{
  if (!node)
    {
      printf (" %s \n", get_command_error);
      return (CommandType) NULL;
    }
  return node->command;
}


/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_GetChild
 *
 * Beschreib.: Liefert den  Nachfolger an der Position pos des Knoten node zurueck.
 *
 * Parameter : (i)node : NodePtrType
 *             (i)pos : NodePositionType
 *
 * Return    : NodePtrType
 *
 *************************************************************************/

NodePtrType 
AgentPrgFunc_GetChild (NodePtrType node, NodePositionType pos)
{
  if (!node)
    {
      printf ("%s \n", get_child_error2);
      return (NodeType *) NULL;
    }
  return (pos < node->argc) ? node->argv[pos] : NULL;
}


/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_CopyTree
 *
 * Beschreib.: Erzeugt Kopie von Baum mit Wurzel node und liefert einen Zeiger darauf
 *             zurueck.
 *
 *
 * Parameter : (i)node : NodePtrType
 *
 * Return    : NodePtrType
 *
 *************************************************************************/
NodePtrType 
AgentPrgFunc_CopyTree (node)
     NodePtrType node;
/*--*/

{
  auto NodeType *tmpn;
  auto int ch;
  if (!node)
    {
      return ((NodeType *) NULL);
    }
  tmpn = AgentPrgFunc_Create ((int) node->argc);
  tmpn->argc = node->argc;	/*do i need this assinment */
  tmpn->command = node->command;
  for (ch = 0; ch < tmpn->argc; ch++)
    {
      tmpn->argv[ch] = AgentPrgFunc_CopyTree (node->argv[ch]);
    }
  return tmpn;
}
/*-end-ftn-*/



/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_NodePtr2Program
 *
 * Beschreib.: Liefert einen ProgramType Typ zurueck ,der die Wurzel des Programmsbaumes
 *             und die Laenge des Programms  enthaelt.
 *
 *
 * Parameter : (i)NodePtr : NodePtrType
 *           : (i)Lenght  : CountType
 *              
 * Return    : ProgramType
 *
 *************************************************************************/

ProgramType *
AgentPrgFunc_NodePtr2Program (NodePtrType NodePtr, CountType Length)
{
  ProgramPtrType pgr;

  pgr = (ProgramType *) malloc (sizeof (ProgramType));
  pgr->NodePtr = NodePtr;
  pgr->Length = Length;
  return pgr;
}


/*-ftn-********************************************************************
 *
 * Funktion  : AgentPrgFunc_Program2NodePtr
 *
 * Beschreib.: Liefert einen Zeiger auf die  Wurzel des Programmsbaumes 
 *             zurueck.
 *
 *
 * Parameter : (i)Program : pointer to ProgramType
 *
 * Return    : NodePtrType
 *
 *************************************************************************/

NodePtrType 
AgentPrgFunc_Program2NodePtr (ProgramType * Program)
{
  return Program ? Program->NodePtr : NULL;
}



CountType 
AgentPrgFunc_GetProgramLength (ProgramType * Program)
{
  if (!Program)
    {
      printf (" warning null pointer APF getprogramlenght \n");
      return 0;
    }
  return Program->Length;
}


void 
AgentPrgFunc_SetProgramLength (ProgramType * Program, CountType Length)
{
  if (!Program)
    {
      printf (" warning null pointer APF setprogramlenght \n");
      return;
    }
  Program->Length = Length;
}

/******************************************************************/
  /*   int    AgentPrgFunc_NumberOfChildren(NodeType * node)        */
  /*   return the number of children that can hold the node "node"  */
/******************************************************************/

int 
AgentPrgFunc_NumberOfChildren (NodeType * node)
{
  if (!node)
    {
      fprintf (stderr, "AgentPrgFunc_numberofchildren : argument is null pointer");
      return 0;
    }
  return (int) node->argc;
}
