/*-module-*****************************************************************
 *
 * Modul     : Field (FUNKTIONSMODUL)
 *
 * Beschreib.: Der Name is falsch. In diesem File stehen die
 *             Implementierungen fuer Hilfsfunktionen fuer
 *             World.c
 *
 * Autor     : Jens Ch. Lisner
 *
 * Datum     : 29. 1. 96
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/

/*-includes-*/
#include <stdio.h>
#include <stdlib.h>
#include "../misc/CDefines.h"
#include "../misc/TypeDefs.h"
#include "../misc/DynaString.h"
#include "../misc/Messages.h"
#include "../misc/Memory.h"
#include "../agents/AgentSet.h"
#include "FieldTypeDefs.h"
#include "Field.h"
#include "../dump/Dump.h"

extern DynaStringType *colors;
extern int numColors;

FieldType *
createWorld ()
{
  FieldType *f;
  unsigned i;

  worldsize = worldx * worldy;
  f = world = Callocate (worldsize, sizeof (FieldType));

  /* 2. Setze Default-Werte fuer die Felder ein. */
  for (i = 0; i < worldsize; i++)
    {
      f->AgentId = 0;
      f->FoodEnergy = 0;
      f->FType = NONE;
      f->CarrionType = 0;
      f->FoodProp = foodprop;
      f++;
    }

  if (!incmap)
    {
      incmap = DynaString_Alloc (INITIAL_MAP_SIZE);
      DynaString_Append (incmap, "MAP\n");
    }

  return world;
}

FieldType *
getfield (PositionType pos)
/* Gibt einen Zeiger auf das Feld mit den Koordinaten pos
   zurueck */
{
  int a = worldx * pos.y + pos.x;

  if (!world)
    createWorld ();

  if ((pos.x < 0) || (pos.y < 0) || (pos.x >= worldx) || (pos.y >= worldy))
    Error ("getfield", "Koordinaten zu gross/klein!!.X:%d Y:%d MaxX:%d MaxY:%d\n", pos.x, pos.y, worldx, worldy);

  return world + a;
}

void 
completedump (int agentsonly, int foodonly)
/* Baut einen dynamischen String map auf in den die Map ausgegeben wird.
   Uebergibt den String weiter nach Dump.
   Falls agentsonly == TRUE werden nur Agents gedumpt.
   Falls foodonly   == TRUE wird nur Food gedumpt.
   Falls beide Flags == FALSE wird eine komplette Map 
   erzeugt. */
{
  FieldType *f = world;
  DynaStringType *map;
  PositionType pos;

  map = DynaString_Alloc (INITIAL_MAP_SIZE);

  DynaString_Append (map, "MAP\n");

  for (pos.y = 0; pos.y < worldy; pos.y++)
    {
      for (pos.x = 0; pos.x < worldx; pos.x++)
	{
	  f = getfield (pos);

	  if (f->FoodEnergy && !agentsonly)
	    {
	      DynaString_AppendFormat (map, 64, "%d %d F %d %c %c %f\n",
				     pos.x, pos.y, f->FoodEnergy, f->FType, (f->FType == PLANT) ? '0' : f->CarrionType, f->FoodProp);
	      /*  fprintf (stderr, "%d %d F %d %c %c %f\n", pos.x, pos.y, f->FoodEnergy, f->FType, f->CarrionType, f->FoodProp); */
	    }
	  if ((f->AgentId && !f->FoodEnergy) && !foodonly)
	    {
	      DynaString_AppendFormat (map, 64, "%d %d A %d\n",
				       pos.x, pos.y, f->AgentId);
	      /*  fprintf (stderr, "%d %d A %d\n", pos.x, pos.y, f->AgentId); */
	    }
	}
    }

  DynaString_Append (map, "END_MAP\n");

  if (agentsonly)
    {
      Dump_DumpMapAgents (DynaString_CString (map));
      DynaString_Free (map);
    }
  else if (foodonly)
    {
      Dump_DumpMapFood (DynaString_CString (map));
      DynaString_Free (map);
    }
  else
    {
      Dump_DumpMapComplete (DynaString_CString (map));
      DynaString_Free (map);
    }
}

static void 
mapdump_common (int agentsonly, int foodonly, char *buffer)
{
  if (agentsonly)
    Dump_DumpMapAgentsXPM (buffer);
  else if (foodonly)
    Dump_DumpMapFoodXPM (buffer);
  else
    Dump_DumpMapCompleteXPM (buffer);
}

void 
xpmdump (int timestep, int agentsonly, int foodonly)
{
  char *line = alloca (worldx + 5);
  PositionType pos;
  FieldType *f;

  mapdump_common (agentsonly, foodonly, "MAP\n");
  mapdump_common (agentsonly, foodonly, "/* XPM */\n");
  mapdump_common (agentsonly, foodonly, "static char * timestep");
  sprintf (line, "%d_xpm[] = {\n", timestep);
  mapdump_common (agentsonly, foodonly, line);

  if (colors)
    {
      sprintf (line, "\"%d %d %d 1\",\n", worldy, worldx, numColors);
      mapdump_common (agentsonly, foodonly, line);

      mapdump_common (agentsonly, foodonly, DynaString_CString (colors));
    }
  else
    {
      sprintf (line, "\"%d %d 9 1\",\n", worldy, worldx);
      mapdump_common (agentsonly, foodonly, line);

      mapdump_common (agentsonly, foodonly, "\"  c black\",\n");	/* empty */
      mapdump_common (agentsonly, foodonly, "\". c green\",\n");	/* food */
      mapdump_common (agentsonly, foodonly, "\"+ c brown\",\n");	/* carrion */
      mapdump_common (agentsonly, foodonly, "\"@ c red\",\n");	/* agent Type 0 */
      mapdump_common (agentsonly, foodonly, "\"A c blue\",\n");		/* Type A */
      mapdump_common (agentsonly, foodonly, "\"B c yellow\",\n");	/* Type B */
      mapdump_common (agentsonly, foodonly, "\"C c grey\",\n");		/* Type C */
      mapdump_common (agentsonly, foodonly, "\"D c violet\",\n");	/* Type D */
      mapdump_common (agentsonly, foodonly, "\"E c white\",\n");	/* Type E */
    }

  line[0] = '"';
  line[worldx + 1] = '"';
  line[worldx + 2] = ',';
  line[worldx + 3] = '\n';
  line[worldx + 4] = '\0';
  for (pos.y = 0; pos.y < worldy; pos.y++)
    {
      for (pos.x = 0; pos.x < worldx; pos.x++)
	{
	  f = getfield (pos);
	  if ((f->FType != NONE) && !agentsonly)
	    {
	      if (f->FType == CARRION)
		line[pos.x + 1] = '+';
	      else
		line[pos.x + 1] = '.';
	    }
	  else if ((f->AgentId > 0) && !foodonly)
	    {
	      AgentIdType type = AgentSet_GetType (f->AgentId);

	      if (type)
		line[pos.x + 1] = (char) type;
	      else
		line[pos.x + 1] = '@';
	    }
	  else
	    {
	      line[pos.x + 1] = ' ';
	    }
	}

      mapdump_common (agentsonly, foodonly, line);
    }

  mapdump_common (agentsonly, foodonly, "};\n");
  mapdump_common (agentsonly, foodonly, "END_MAP\n");
}
