



/*-module-*****************************************************************
 *
 * Modul     : World (FUNKTIONSMODUL)
 *
 * Beschreib.: Implementierung der Funktionen aus World.h
 *             Hilfsfunktionen befinden sich in Field.c
 *             Globale Variablen in Field.h
 *
 * Autor     : Jens Ch. Lisner
 *
 * Datum     : 29. 1. 96
 *
 * Aender.   :
 *
 *   geaendert am              von             Bemerkungen
 *
 *--**********************************************************************/


/*-includes-*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <limits.h>
#include "FieldTypeDefs.h"

#include "../misc/CDefines.h"
#include "../misc/Messages.h"
#include "../misc/TypeDefs.h"
#include "../misc/DynaString.h"
#include "../random/Random.h"

#include "../dump/Dump.h"
#include "Field.h"
#include "../agents/AgentTypeDefs.h"
#include "../agents/AgentSet.h"
#include "World.h"

/*-defines-*/

#define FRANDOM()	((float) Random_Rand() / (float) ULONG_MAX)
#define LARGE_ENERGY	1000000

void xpmdump (int, int, int);

static int RottingCarrion = 0;
static int KeepFood = 0;
static int RottingPlants = 0;

DynaStringType *colors = 0;
int numColors = 0;

int incq;

inline FoodType 
World_GetFoodType (PositionType pos)
{
  return getfield (pos)->FType;
}

inline AgentIdType 
World_GetCarrionType (PositionType pos)
{
  return getfield (pos)->CarrionType;
}

AgentIdType
World_GetFoodCarrionType (PositionType pos)
{
  if (getfield (pos)->FType == 'C')
    return getfield (pos)->CarrionType;
  else
    return 'P';
}

/*-ftn-********************************************************************
 *
 * Funktion  : World_SetAgent
 *
 * Beschreib.: Setzt einen Agenten an die angegebene Stelle .
 *             Geht von korrekten Werten aus sonst exit(1).
 *
 *
 * Parameter : (i)id : AgentIdType
 *             (i)pos : PositionType
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
World_SetAgent (AgentIdType id, PositionType pos)
{
  FieldType *f = getfield (pos);

  if ((f->AgentId) || (f->FoodEnergy))
    Error ("World_SetAgent", "Platz schon besetzt!\n");

  f->AgentId = id;
  f->FType = NONE;
  f->CarrionType = 0;

  if (incq)
    DynaString_AppendFormat (incmap, 64, "%d %d A %d\n", pos.x, pos.y, id);
}


/*-ftn-********************************************************************
 *
 * Funktion  : World_AgentThere
 *
 * Beschreib.: Gibt eine Id zurueck falls dort ein Agent steht ,sonst 0.
 *
 *
 *
 * Parameter : (i)pos : PositionType
 *
 * Return    : AgentIdType
 *
 *************************************************************************/
AgentIdType 
World_AgentThere (PositionType pos)
{
  FieldType *f = getfield (pos);

  if (f->FoodEnergy == 0)
    return f->AgentId;
  else
    return 0;
}


/*-ftn-********************************************************************
 *
 * Funktion  : World_GetEnergy
 *
 * Beschreib.: Gibt den EnergieLevelk des gefunden Futteres zurueck,sonst 0.
 *
 *
 *
 * Parameter : (i)pos : PositionType
 *
 * Return    : EnergyType
 *
 *************************************************************************/
inline EnergyType 
World_GetEnergy (PositionType pos)
{
  return getfield (pos)->FoodEnergy;
}


/*-ftn-********************************************************************
 *
 * Funktion  : World_AgentMove
 *
 * Beschreib.: Geht von korrekten Werten aus sonst exit(1).
 *
 * Parameter : (i)id : AgentIdType
 *             (i)pos_old : PositionType
 *             (i)pos_new : PositionType
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
World_AgentMove (AgentIdType id, PositionType pos_old,
		 PositionType pos_new)
{
  FieldType *o = getfield (pos_old);

  if (o->AgentId != id)
    Error ("World_AgentMove", "Falsche AgentId %d!\n", id);

  o->AgentId = 0;
  if (incq)
    DynaString_AppendFormat (incmap, 64, "%d %d 0\n", pos_old.x, pos_old.y);
  World_SetAgent (id, pos_new);
}


/*-ftn-********************************************************************
 *
 * Funktion  : World_FoodRemove
 *
 * Beschreib.: Loescht das Futter an der anggegebenen Stelle.
 *
 * Parameter : (i)pos : PositionType
 *                amount : EnergyType (Beinhaltet die Energie, die *mglicher-
 *                                     weise* zurckbehalten wird, falls 
 *                                     KeepFood == 1)
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
World_FoodRemove (PositionType pos, EnergyType amount)
{
  FieldType *f = getfield (pos);

  f->AgentId = 0;
  if (KeepFood == 1)
    f->FoodEnergy = (amount<0) ? 0 : amount;
  else
    f->FoodEnergy = 0;
    
  if (f->FoodEnergy == 0)
    {
      f->FoodEnergy = 0;
      f->FType = NONE;
      f->CarrionType = 0;
    }
    
  if (incq)
    if (f->FoodEnergy == 0) 
      DynaString_AppendFormat (incmap, 64, "%d %d 0\n", pos.x, pos.y);
    else
      DynaString_AppendFormat (incmap, 64, "%d %d F %d %c %c %f\n",
	pos.x, pos.y, f->FoodEnergy, f->FType, f->CarrionType, f->FoodProp);
      
}


/*-ftn-********************************************************************
 *
 * Funktion  : World_AgentFoodConvert
 *
 * Beschreib.: Convertiert einen Agenten in Futter.
 *
 * Parameter : (i)id : AgentIdType
 *             (i)energie : EnergyType
 *             (i)pos : PositionType
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
World_AgentFoodConvert (AgentIdType id, EnergyType energie,
			PositionType pos)
{
  FieldType *f = getfield (pos);

  if (f->AgentId != id)
    Error ("World_AgentFoodConvert", "Falsche AgentId!\n");

  if (energie < 0)
    Error ("World_AgentFoodConvert", "Negative Energien!??\n");

  f->AgentId = 0;
  f->FoodEnergy = energie;
  if (!energie)
    f->FType = NONE;
  else
    {
      f->FType = CARRION;
      f->CarrionType = AgentSet_GetType (id);
    }

  if (incq)
    if (energie)
      DynaString_AppendFormat (incmap, 64, "%d %d F %d %c %c %f\n",
	pos.x, pos.y, f->FoodEnergy, f->FType, f->CarrionType, f->FoodProp);
    else
      DynaString_AppendFormat (incmap, 64, "%d %d 0\n", pos.x, pos.y);
}

/*-ftn-********************************************************************
 *
 * Funktion  : World_InitTimeStep
 *
 * Beschreib.: Zeitschrittfunktion zum Ausloesen verschiedener Funktionen.
 *
 * Parameter : (i)timestep : CountType
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
World_InitTimestep (CountType timestep)
{
  FieldType *f = world;
  int x, y, i;
  unsigned int foodCount = 0, carrionCount = 0;
  unsigned int completefoodenergy[3], completecarrionenergy[3];
  unsigned int completeagentsenergy[81];
  bool stat_wanted;
  AgentIdType atype;
  EnergyType aenergy;

  DebugInfo ("World_InitTimeStep - Start\n");
  DebugInfo ("World_InitTimeStep - Futter\n");

  incq = Dump_WantDumpMapIncremental ();
  stat_wanted = Dump_WantDumpStatistics ();

  if (stat_wanted)
    {
      completefoodenergy[0] = completecarrionenergy[0] = LARGE_ENERGY;
      /* Mglichst grer als der hchste zu erwartende Energiewert */

      completefoodenergy[1] = completecarrionenergy[1] = 0;
      completefoodenergy[2] = completecarrionenergy[2] = 0;

      for (i = 0; i <= 80; i += 3)
	{
	  completeagentsenergy[i] = LARGE_ENERGY;	/* s.o. */
	  completeagentsenergy[i + 1] = completeagentsenergy[i + 2] = 0;
	}
    }

  /* 1. Phase : Lasse neues Futter wachsen. */

  for (y = 0; y < worldy; y++)
    {
      for (x = 0; x < worldx; x++)
	{
	  if ((f->FoodEnergy == 0) && (f->AgentId == 0))
	    {
	      if (FRANDOM () <= f->FoodProp)
		{
		  f->FoodEnergy = fenergy;
		  f->FType = PLANT;

		  if (incq)
		    DynaString_AppendFormat (incmap, 64, "%d %d F %d %c %c %f\n",
		      x, y, fenergy, f->FType, f->CarrionType, f->FoodProp);
		}
	    }			/* if */
	  else if ((f->FoodEnergy > 0) && (f->FType == CARRION))

	    /*      Aas wird schlecht */
	    if (f->FoodEnergy <= RottingCarrion)
	      {
		f->FoodEnergy = 0;
		f->FType = NONE;	/* dust to dust */
		f->CarrionType = 0;

		if (incq)
		  DynaString_AppendFormat (incmap, 64, "%d %d 0\n", x, y);
	      }
	    else
	      {
		f->FoodEnergy -= RottingCarrion;
		if (incq)
		  DynaString_AppendFormat (incmap, 64, "%d %d F %d %c %c %f\n",
					   x, y, f->FoodEnergy, f->FType, f->CarrionType, f->FoodProp);
	      }
	  else if ((f->FoodEnergy > 0) && (f->FType == PLANT))
	    if (f->FoodEnergy <= RottingPlants)
	      {
		f->FoodEnergy = 0;
		f->FType = NONE;	/* dust to dust */
		f->CarrionType = 0;

		if (incq)
		  DynaString_AppendFormat (incmap, 64, "%d %d 0\n", x, y);
	      }
	    else
	      {
		f->FoodEnergy -= RottingPlants;
		if (incq)
		  DynaString_AppendFormat (incmap, 64, "%d %d F %d %c %c %f\n",
					   x, y, f->FoodEnergy, f->FType, f->CarrionType, f->FoodProp);
	    
               }
	  /* else if */
/* Infos fr Statistics sammeln */

	  if (stat_wanted)
	    if (f->FoodEnergy > 0)
	      {
		if (f->FType == PLANT)
		  {
		    foodCount++;
		    if (f->FoodEnergy < completefoodenergy[0])
		      completefoodenergy[0] = f->FoodEnergy;
		    completefoodenergy[1] += f->FoodEnergy;
		    if (f->FoodEnergy > completefoodenergy[2])
		      completefoodenergy[2] = f->FoodEnergy;
		  }
		else
		  {
		    carrionCount++;
		    if (f->FoodEnergy < completecarrionenergy[0])
		      completecarrionenergy[0] = f->FoodEnergy;
		    completecarrionenergy[1] += f->FoodEnergy;
		    if (f->FoodEnergy > completecarrionenergy[2])
		      completecarrionenergy[2] = f->FoodEnergy;
		  }
	      }
	    else if (f->AgentId > 0)
	      {
		atype = AgentSet_GetType (f->AgentId);
		if (atype == '0')
		  atype = 'Z' + 1;
		atype = (atype - 'A') * 3;
		aenergy = AgentSet_GetEnergy (f->AgentId);
		if (aenergy < completeagentsenergy[atype])
		  completeagentsenergy[atype] = aenergy;
		completeagentsenergy[atype + 1] += aenergy;
		if (aenergy > completeagentsenergy[atype + 2])
		  completeagentsenergy[atype + 2] = aenergy;
	      }
	  /* stat_wanted */


	  f++;
	}
    }
	
  if (stat_wanted)
    {
      char buf[80];

      sprintf (buf, "#! Food = %d\n", foodCount);
      Dump_DumpStatistics (buf);
      sprintf (buf, "#! Carrion = %d\n", carrionCount);
      Dump_DumpStatistics (buf);
      if (completeagentsenergy[78] != LARGE_ENERGY)
	{
	  sprintf (buf, "#! Type 0 Min = %d\n", completeagentsenergy[78]);
	  Dump_DumpStatistics (buf);
	  sprintf (buf, "#! Type 0 Total = %d\n", completeagentsenergy[79]);
	  Dump_DumpStatistics (buf);
	  sprintf (buf, "#! Type 0 Max = %d\n", completeagentsenergy[80]);
	  Dump_DumpStatistics (buf);
	}
      for (i = 0; i <= 77; i += 3)
	{
	  if (completeagentsenergy[i] != LARGE_ENERGY)
	    {
	      atype = 'A' + (i / 3);
	      sprintf (buf, "#! Type %c Min = %d\n", atype, completeagentsenergy[i]);
	      Dump_DumpStatistics (buf);
	      sprintf (buf, "#! Type %c Total = %d\n", atype, completeagentsenergy[i + 1]);
	      Dump_DumpStatistics (buf);
	      sprintf (buf, "#! Type %c Max = %d\n", atype, completeagentsenergy[i + 2]);
	      Dump_DumpStatistics (buf);
	    }
	}
      sprintf (buf, "#! Food Min = %d\n", completefoodenergy[0]);
      Dump_DumpStatistics (buf);
      sprintf (buf, "#! Food Total = %d\n", completefoodenergy[1]);
      Dump_DumpStatistics (buf);
      sprintf (buf, "#! Food Max = %d\n", completefoodenergy[2]);
      Dump_DumpStatistics (buf);

      if (completecarrionenergy[0] == LARGE_ENERGY)
	completecarrionenergy[0] = 0;
      sprintf (buf, "#! Carrion Min = %d\n", completecarrionenergy[0]);
      Dump_DumpStatistics (buf);
      sprintf (buf, "#! Carrion Total = %d\n", completecarrionenergy[1]);
      Dump_DumpStatistics (buf);
      sprintf (buf, "#! Carrion Max = %d\n", completecarrionenergy[2]);
      Dump_DumpStatistics (buf);
    }				/* stat_wanted */
}

/*-ftn-********************************************************************
 *
 * Funktion  : World_ConcludeTimestep
 *
 * Beschreib.: Zeitschrittfunktion zum Ausloesen des Dumpings.
 *
 * Parameter : (i)timestep : CountType
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
World_ConcludeTimestep (CountType timestep)
{
  /* Dumping */

  if (incq)
    {
      DynaString_Append (incmap, "END_MAP\n");
      Dump_DumpMapIncremental (DynaString_CString (incmap));
      DynaString_Reset (incmap);
      DynaString_Append (incmap, "MAP\n");
    }

  if (Dump_WantDumpMapAgents ())
    completedump (1, 0);

  if (Dump_WantDumpMapFood ())
    completedump (0, 1);

  if (Dump_WantDumpMapComplete ())
    completedump (0, 0);

  if (Dump_WantDumpMapAgentsXPM ())
    xpmdump (timestep, 1, 0);

  if (Dump_WantDumpMapFoodXPM ())
    xpmdump (timestep, 0, 1);

  if (Dump_WantDumpMapCompleteXPM ())
    xpmdump (timestep, 0, 0);

}


/*-ftn-********************************************************************
 *
 * Funktion  : World_EndSimulation
 *
 * Beschreib.: Loest das Letzte Dumping aus.
 *
 *
 *
 * Parameter :
 *
 * Return    : * keiner *
 *
 *************************************************************************/
void 
World_EndSimulation ()
{
  if (Dump_WantDumpMapComplete ())
    completedump (0, 0);

  free (world);
  free (DynaString_CStringAndFree (incmap));
}


/*-ftn-********************************************************************
 *
 * Funktion  : World_ParseSection
 *
 * Beschreib.:
 *
 * Parameter : (i)FilePointer : file
 *
 * Return    : bool
 *
 *************************************************************************/
bool 
World_ParseSection (file FilePointer)
{
  char buf[256] = "";
  string bufp = buf;
  FieldType *f;
  int n, id, energy;
  char c, c1, carriontype, ftype;
  PositionType pos;
  float prop;

  createWorld ();

  incq = Dump_WantDumpMapIncremental ();

  /* Parser */

  /* Stoppt bei Fileende oder ENDMAP */

  while ((fgets (buf, 255, FilePointer) != NULL) &&
	 (strncmp (buf, "END_MAP", 6)))
    {
      bufp = buf;

      while (isspace (*bufp) || (*bufp == '\n'))
	{
	  bufp++;
	}

      if (!*bufp || (strncmp (bufp, "MAP", 3) == 0))
	continue;

      if (sscanf (bufp, "%d %d %1s %n", &pos.x, &pos.y, &c, &n) == EOF)
	{
	  Error ("World_ParseSection", "Fehler in der Map!\n");
	}

      switch (c)
	{
	case 'A':
	  if (sscanf (bufp + n, " %1s", &c1) == EOF)
	    {
	      Error ("World_ParseSection", "Fehler in der Map: Keine Id fuer Agent!\n");
	    }
	  if (isdigit (c1))
	    {
	      sscanf (bufp + n, "%d", &id);
	      if (incq)
		DynaString_AppendFormat (incmap, 64, "%d %d A %d", pos.x, pos.y, id);
	    }
	  else
	    {
	      id = -(AgentIdType) c1;	/* Nicht in die incmap eintragen! Wird noch gesetzt */
	    }

	  World_SetAgent (id, pos);
	  break;
	case 'F':
	  if (sscanf (bufp + n, "%d %1s %1s %f", &energy, &ftype, &carriontype, &prop) == EOF)
	    {
	      Error ("World_ParseSection", "Fehler in der Map: Fehlende Angabe(n) bei Foodpatches!!\n");
	    }
	  f = getfield (pos);
	  f->FoodEnergy = energy;
	  f->FType = (int) ftype;
	  f->CarrionType = (AgentIdType) (carriontype == '0') ? 0 : carriontype;
	  f->FoodProp = prop;
	  if (incq)
	    DynaString_AppendFormat (incmap, 64, "%d %d F %d %c %c %f", pos.x, pos.y,
			    energy, (char) ftype, (char) carriontype, prop);
	  break;
	case '0':
	  break;		/* Dummy-Argument. Wird benutzt
				   um anzuzeigen, dass ein Feld
				   leer ist. */

	default:
	  Error ("World_ParseSection", "Unbekannter Kennbuchstabe in der Map!\n");
	}
    }
  return TRUE;			/* Falls nicht ist der eh schon ausgestiegen. */
}


/*-ftn-********************************************************************
 *
 * Funktion  : World_SetValueForKey
 *
 * Beschreib.:
 *
 * Parameter : (i)AttributeName : string
 *             (i)AttributeValue : string
 *
 * Return    : bool
 *
 *************************************************************************/
bool 
World_SetParameterValue (string AttributeName, string AttributeValue)

/* Setzt globale Attributwerte */
{
  static string AttribTable[] =
  {
    "FoodPropMap", "WorldWidth", "WorldHeight", "FoodEnergy",
    "Decay", "KeepFood", "RottingPlants",
    "XPMColorMap"
  };

  const int Attribnr = 7;
  int ecode = 0, i;
  bool found = TRUE;

  for (i = 0; i <= Attribnr; i++)
    if (strcmp (AttributeName, AttribTable[i]) == 0)
      break;

  switch (i)
    {
    case 0:			/* FoodPropMap */
      ecode = sscanf (AttributeValue, "%f", &foodprop);
      break;
    case 1:			/* WorldWidth */
      if ((ecode = sscanf (AttributeValue, "%d", &worldx)) == 1)
	worldsize = worldx * worldy;
      break;
    case 2:			/* WorldHeight */
      if ((ecode = sscanf (AttributeValue, "%d", &worldy)) == 1)
	worldsize = worldx * worldy;

      break;
    case 3:			/* FoodEnergy */
      ecode = sscanf (AttributeValue, "%d", &fenergy);
      break;
    case 4:			/* Decay */
      ecode = sscanf (AttributeValue, "%d", &RottingCarrion);
      break;
    case 5:                     /* KeepFood */
      ecode = sscanf (AttributeValue, "%d", &KeepFood);
      break;
    case 6:                     /* RottingPlants */
      ecode = sscanf (AttributeValue, "%d", &RottingPlants);
      break;
    case 7:			/* XPMColorMap */
      {
	char filename[257];
	FILE *colorfile;

	ecode = sscanf (AttributeValue, "%s", filename);
	if ((colorfile = fopen (filename, "r")))
	  {
	    char line[81];

	    if (!colors)
	      colors = DynaString_Alloc (256);

	    numColors = 0;
	    DynaString_Reset (colors);
	    while (!feof (colorfile))
	      {
		line[0] = '\0';
		fgets (line, 81, colorfile);
		if (*line)
		  {
		    DynaString_Append (colors, line);
		    numColors++;
		  }
		else
		  break;
	      }
	  }
	break;
      }
    default:
      found = FALSE;
      break;
    }

  if (ecode == EOF)
    Error ("World_SetParameterValue", "AttributeValue==Muell.\n");
  return found;
}
